import java.awt.*;
import java.awt.event.*; 
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.io.PrintStream;
import java.io.FileReader;
import java.io.File;
import java.io.BufferedReader;
import java.io.*;
import java.util.*;

import uk.ac.ox.cs.fdr.*;

public class InterpreterGUI 
{
    	//final static boolean DEBUG = false;
    	
    	final static boolean INITIAL_SHOW_VARS = true;
    	final static boolean INITIAL_SHOW_WRITE_STARTS = true;
    	final static boolean INITIAL_SHOW_ATOMICS = true;
    	
    	final static String CSPFILE = "output.csp";
    	final static String DEFAULTRESULTSFILE = "FDRoutput.output";
    	String resultsFile = null;
    	
    	//declerations of all the components to be used
    	protected JFrame frame = new JFrame("SVA");
    	protected TableData data = new TableData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>(),INITIAL_SHOW_VARS, INITIAL_SHOW_WRITE_STARTS,INITIAL_SHOW_ATOMICS);
    	protected JList assertsList = new JList();	
    	protected JLabel result = new JLabel("Result:");
		protected JScrollPane assertsPane = new JScrollPane(assertsList);
		protected JCheckBox vars = new JCheckBox("Show Variable Reads");
		protected JCheckBox writes = new JCheckBox("Show Starts of Dirty Variable Assignments");
		protected JCheckBox atomic = new JCheckBox("Show Atomic Sections");
		protected JButton runFDR = new JButton("Check All Assertions");
		protected JButton runOneAssert = new JButton("Check Assertion");
		//protected JButton interrupt = new JButton("Interrupt"); removed this button to halt progress of long processes since it did not actually stop fdr from running
		protected JTable counterexample = new JTable(data);
		protected JScrollPane examplePane = new JScrollPane(counterexample);
		protected JMenuBar menubar = new JMenuBar();
		protected JMenu fileMenu= new JMenu("File");
		protected JMenu helpMenu= new JMenu("Help");
		protected JMenu optionsMenu= new JMenu("Options");
		protected JMenuItem about = new JMenuItem("About");
		protected JMenuItem load = new JMenuItem("Load");
		protected JMenuItem reload = new JMenuItem("Reload");
		protected JMenuItem saveCSP = new JMenuItem("Save CSP");
		protected JMenuItem exit = new JMenuItem("Exit");
		protected JMenuItem showStatus = new JMenuItem("Show Status");
		protected JFileChooser fileChooser = new JFileChooser(".");
    	
		
		protected JFrame statusFrame = new JFrame("Status");
		protected JTextArea statusText = new JTextArea(25,50);
		protected JScrollPane statusPane = new JScrollPane(statusText);
    	
    	protected Map<Integer,TableDataInterpreter.AssertionOutput> results = new TreeMap<Integer,TableDataInterpreter.AssertionOutput>();
    	
    	protected FDRStatusListener statusListener = new FDRStatusListener(statusText);
    	
    	protected String fileName = "";
    	
    	protected Interpreter interpreter;
    	
    	public void addComponentsToFrame() 
    	{
    		Container pane = frame.getContentPane();

	//behaviour of components

    		//JList assertsList
		
		assertsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		assertsList.addListSelectionListener( 
			new ListSelectionListener()
			{
				int current;
				public void valueChanged(ListSelectionEvent e) 
				{
					updateResult();
				}
			});
		
		//JScrollPane assertsPane
		
		
		//JLabel result
		
		
		//JCheckBox vars
		
		vars.setSelected(INITIAL_SHOW_VARS);
		vars.addItemListener(
			new ItemListener()
			{
				public void itemStateChanged(ItemEvent e)
				{
					data.changeVarReads(e.getStateChange() == ItemEvent.SELECTED);
				}
			});
		
		//JCheckBox writes
		
		writes.setSelected(INITIAL_SHOW_WRITE_STARTS);
		writes.addItemListener(
			new ItemListener()
			{
				public void itemStateChanged(ItemEvent e)
				{
					data.changeWriteStarts(e.getStateChange() == ItemEvent.SELECTED);
				}
			});
		
		//JCheckBox atomic
		
		atomic.setSelected(INITIAL_SHOW_ATOMICS);
		atomic.addItemListener(
			new ItemListener()
			{
				public void itemStateChanged(ItemEvent e)
				{
					data.changeAtomics(e.getStateChange() == ItemEvent.SELECTED);
				}
			});
		
		
		runOneAssert.setEnabled(false);
		runOneAssert.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							runOneAssert(assertsList.getSelectedIndex());
						}
					});
		
		//JButton runFDR
		
		runFDR.setEnabled(false);
		runFDR.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							runFDR();
						}
					});
		
		//JButton interrupt
		//interrupt.setEnabled(false);
		
		//JTable counterexample
		
		counterexample.setColumnSelectionAllowed(false);
		counterexample.setRowSelectionAllowed(false);
		
		//JScrollPane examplePane

		//JMenuBar menubar
		
		//JMenu fileMenu
		
		//JMenuItem about
		
		about.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							showAbout();
						}
					});
		
		
		//JMenuItem load
		
		load.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							if (! fileName.equals(""))
								fileChooser.setSelectedFile(new File(fileName));
								
							int returnVal = fileChooser.showOpenDialog(load);	

							if (returnVal == JFileChooser.APPROVE_OPTION) 
								load(fileChooser.getSelectedFile().toString());
						}
					});
		
		//JMenuItem reload 
		
		reload.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							load(fileName);
						}
					});
		reload.setEnabled(false);
		
		//JMenuItem saveCSP
		
		saveCSP.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							if (! fileName.equals(""))
							{	
								int index = fileName.indexOf('.');
								String defaultName = ( index > 0 ? fileName.substring(0,index) : fileName) + ".csp"; 
								fileChooser.setSelectedFile(new File(defaultName));
							}
							int returnVal = fileChooser.showSaveDialog(saveCSP);	
							
							if (returnVal == JFileChooser.APPROVE_OPTION) 
								saveCSP(fileChooser.getSelectedFile().toString());
						}
					});
		
			
		saveCSP.setEnabled(false);
		
		//JMenuItem exit
		
		exit.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							System.exit(0);
						}
					});
		
			
		exit.setEnabled(true);
		
		
		//JMenuItem showStatus
		
		showStatus.addActionListener( new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							statusFrame.setVisible(true);
						}
					});
		
			
		showStatus.setEnabled(true);
		
		statusText.getDocument().addDocumentListener(
			new DocumentListener()
			{
				public void makeChange()
				{	
					
					SwingUtilities.invokeLater(
					new Runnable() {
						public void run() {
							statusPane.getVerticalScrollBar().setValue(statusPane.getVerticalScrollBar().getMaximum());
						}
					}
					);
				
				}
				
				public void changedUpdate(DocumentEvent e) {makeChange();}
				public void insertUpdate(DocumentEvent e) {makeChange();}
				public void removeUpdate(DocumentEvent e) {makeChange();}
			
			
			})
			
			;

		statusListener.activate();

		//JFileChooser fileChooser

	//Layout of components

		pane.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.anchor=GridBagConstraints.LINE_START;
		c.fill = GridBagConstraints.BOTH;

		
    		//JList assertsList
		
		//JScrollPane assertsPane
		
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = .5;
		c.gridheight = 5;
		pane.add(assertsPane, c);
		c.gridheight = 1;//reset
		c.weightx = 0;
		c.weighty = 0;
		
		//JLabel result
		
		result.setFont(result.getFont().deriveFont(20.0f));
		c.gridx = 0;
		c.gridy = 5;
		c.gridwidth= 3;
		pane.add(result, c);
		c.gridwidth = 1;
		
		//JCheckBox vars
		
		c.gridx = 1;
		c.gridy = 0;
		c.gridwidth= 2;
		c.weightx = 0;
		c.weighty = 0;
		pane.add(vars, c);
		
		//JCheckBox writes
		
		c.gridx = 1;
		c.gridy = 1;
		pane.add(writes, c);
		
		//JCheckBox atomic
		
		c.gridx = 1;
		c.gridy = 2;
		pane.add(atomic, c);
		//reset
		c.gridwidth= 1;
		
		//JButton runOneAssert
		
		c.gridx = 1;
		c.gridy = 3;
		pane.add(runOneAssert,c);
		
		//JButton runFDR
		
		c.gridx = 2;
		c.gridy = 3;
		pane.add(runFDR,c);
		
		//JButton interrupt
		/*c.gridx = 1;
		c.gridy = 4;
		pane.add(interrupt,c);*/
		
		//JTable counterexample
		
		counterexample.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		counterexample.setEnabled(false);
		//counterexample.setPreferredScrollableViewportSize(new Dimension(500, 700));
		
		//JScrollPane examplePane

		c.gridwidth = 3;
		c.gridx = 0;
		c.gridy = 6;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		pane.add(examplePane, c);
		
		//JMenuBar menubar
		
		frame.setJMenuBar(menubar);

		//JMenu fileMenu
		
		menubar.add(fileMenu);	
		menubar.add(optionsMenu);		
		menubar.add(helpMenu);
		
		helpMenu.add(about);
		
		//JMenuItem load
		
		fileMenu.add(load);
		
		//JMenuItem reload 
		
		fileMenu.add(reload);
		
		//JMenuItem saveCSP
		
		fileMenu.add(saveCSP);
				
		//JMenuItem exit
		
		fileMenu.add(exit);
		
		
		optionsMenu.add(showStatus);
		

		//protected JTextArea statusText= new JTextArea();
		statusText.setEditable(false);
		
		
		//protected JFrame statusFrame = new JFrame("Status");
		statusFrame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
		statusFrame.getContentPane().add(statusPane, BorderLayout.CENTER);
		statusFrame.pack();
		statusFrame.setVisible(false);
		
		
    	}

	protected void updateResult()
	{
	if (assertsList.getSelectedIndex() == -1) 
					{
						data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
						result.setText("Result:");
					} 
					else if (results.containsKey(assertsList.getSelectedIndex()))
					{
						TableDataInterpreter.AssertionOutput selected = results.get(assertsList.getSelectedIndex());
					
						data.setData(selected.procs,selected.counterexample);
						result.setText("Result: " + selected.result);
						runOneAssert.setEnabled(false);
					}
					else
					{
						data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
						result.setText("Result: unknown");
						runOneAssert.setEnabled(true);
					
					}
					setTableWidths(counterexample);
	}

	
	//modified from TableRenderDemo
	static protected void setTableWidths(JTable table) 
	{
		TableModel model = table.getModel();
		TableColumn column = null;
		Component comp = null;
		int headerWidth = 0;
		int cellWidth = 0;
		//Object[] longValues = model.longValues;
		
		TableCellRenderer headerRenderer =
		table.getTableHeader().getDefaultRenderer();
	
		for (int i = 0; i < table.getColumnCount(); i++) 
		{
				
			column = table.getColumnModel().getColumn(i);
		
			comp = headerRenderer.getTableCellRendererComponent(
						null, column.getHeaderValue(),
						false, false, 0, 0);
			headerWidth = comp.getPreferredSize().width;
			cellWidth = 90;//have a minimum width of 100
		
			for(int j = 0; j < table.getRowCount(); j++)
			{
				comp = table.getDefaultRenderer(model.getColumnClass(i)).
					getTableCellRendererComponent(
						table, table.getValueAt(j,i),
						false, false, 0, i);
				cellWidth = Math.max(cellWidth,comp.getPreferredSize().width);
			}
		
			Debug.trace("Initializing width of column " + i + ". " + "headerWidth = " + headerWidth + "; cellWidth = " + cellWidth);
		
		
			//XXX: Before Swing 1.1 Beta 2, use setMinWidth instead.
			column.setPreferredWidth(Math.max(headerWidth, cellWidth)+10);
			//column.setMinWidth(100);//for empty columns
			//column.setResizable(false);
		}
    }


	/**
	* Create the GUI and show it.  For thread safety,
	* this method should be invoked from the
	* event-dispatching thread.
	*/
	private void createAndShowGUI() 
	{
		//Create and set up the window.
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	
		//Set up the content pane.
		addComponentsToFrame();
	
		//Display the window.
		frame.pack();
		frame.setVisible(true);
		
		statusFrame.setLocation(frame.getX() + frame.getWidth(), frame.getY());
	}

	public void run()
	{
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
		public void run() {
			createAndShowGUI();
		}
		});
	}
	
	public void run(String source)
	{ run(source,CSPFILE,null,null);}
	
	public void run(final String source,final String CSPFile,final String readFDRFile,final String writeFDRFile) 
	{
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
		public void run() {
			createAndShowGUI();
			load(source);
		}
		});
	
	}
	
	/**
	* For thread safety,
	* this method should be invoked from the
	* event-dispatching thread.
	*/
	protected void load(final String source)
	{
		Debug.trace("load(" + source + ")"); 
	
		//show busy cursor
		frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			
		//change title
		frame.setTitle("SVA - Loading " + source);	
		fileName = source;
			
		assertsList.setEnabled(false);
		vars.setEnabled(false);
		writes.setEnabled(false);
		atomic.setEnabled(false);
		runFDR.setEnabled(false);
		runOneAssert.setEnabled(false);
		load.setEnabled(false);
		reload.setEnabled(false);
		saveCSP.setEnabled(false);
		results.clear();
	
		final SwingWorker worker = new SwingWorker<String[], Void>()
		{
			@Override
    			public String[] doInBackground()
    			{
     				try
				{
					resultsFile = null;	
					interpreter = new Interpreter(source);
					return new String[] {};
				}
				catch(java.io.FileNotFoundException e)
				{
					return new String[] {"File Not Found While Loading" + source, e.getMessage()};
				}
				catch (uk.ac.ox.cs.fdr.Error e)
				{
					return new String[] {"CSP Error in" + source, e.getMessage()};
				}
				/*
				catch(java.io.IOException e)
				{
					return new String[] {"Error", e.getMessage()};
				}
				catch(InterruptedException e)
				{
					return new String[] {"Interrupted", e.getMessage()};
				}
				catch(ParserException e)
				{
					return new String[] {"Error Parsing File", e.getMessage()};
				}*/
			}
			
			@Override
    			public void done()
    			{
    				String[] message;
    				try
    				{
    					//if (isCancelled())
    					//	message = new String[] {"Interrupted", "Loading File Halted"};
    					//else
    					//{
					message = get();
					if (message.length == 0)
					{	//success
						Vector<String> asserts = new Vector<String>();
						Iterator<Assertion> i = interpreter.compiler.asserts.iterator(); 
						while (i.hasNext())
							asserts.add(i.next().toString());
						assertsList.setListData(asserts);
						//assertsList.setSelectedIndex(-1);//DOES THIS WORK?? not needed anymore?
						
						//reenable options to get results or a new file
						
						runFDR.setEnabled(true);
						load.setEnabled(true);
						reload.setEnabled(true);
						saveCSP.setEnabled(true);
						
						
						assertsList.setEnabled(true);
						
						//interrupt.setEnabled(false);
						//for (ActionListener al : interrupt.getActionListeners())
						//	interrupt.removeActionListener(al);
							
						//set title
						frame.setTitle("SVA - " + fileName);
						
						//remove busy cursor
						frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						return;
					}
					//}
				}
				catch(InterruptedException e)
				{
					message = new String[] {"Interrupted While Loading" + source, e.getMessage()};
				}
				catch(java.util.concurrent.ExecutionException e)
				{
					message = new String[] {"Error While Loading" + source, e.getMessage()};
				}
    				//else error!
    				//{
    					JOptionPane.showMessageDialog(frame, message[1], message[0], JOptionPane.WARNING_MESSAGE);
					
					//clear everything
					data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
					assertsList.setListData(new String[] {});
					//assertsList.setEnabled(true);
					//vars.setEnabled(true);
					//exps.setEnabled(true);
					//atomic.setEnabled(true);
					load.setEnabled(true);
					reload.setEnabled(true);
					//interrupt.setEnabled(false);
					//for (ActionListener al : interrupt.getActionListeners())
					//	interrupt.removeActionListener(al);
					
					//set title
					frame.setTitle("SVA - Error Loading " + fileName);
					
					//remove busy cursor
					frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					return;
					
					//remove loading screen
    				//}
    			}
		
		};
		
		
		/*interrupt.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							worker.cancel(true);
						}
					});
		interrupt.setEnabled(true);*/
		worker.execute();
	}
	
	
	/**
	* For thread safety,
	* this method should be invoked from the
	* event-dispatching thread.
	*/
	protected void runFDR()
	{
		//show busy cursor
		frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			
		//change title
		frame.setTitle("SVA - Running FDR - " + fileName);	
			
		//disable everything
		assertsList.setEnabled(false);
		vars.setEnabled(false);
		writes.setEnabled(false);
		atomic.setEnabled(false);
		runFDR.setEnabled(false);
		runOneAssert.setEnabled(false);
		load.setEnabled(false);
		reload.setEnabled(false);
		saveCSP.setEnabled(false);
	
		final SwingWorker worker = new SwingWorker<String[], Void>()
		{
			protected Map<Integer,TableDataInterpreter.AssertionOutput> newResults;
			@Override
    			public String[] doInBackground()
    			{
     				try
				{
					//store results in RESULTSFILE so we can save them elsewhere if the
					//user requests it later on
					interpreter.runAllAsserts(CSPFILE);
					TableDataInterpreter output = new TableDataInterpreter(interpreter.compiler.context.globalConsts, interpreter.compiler.context.dirtyVars);
										
					interpreter.interpret(new TraceConstructor(), output);	
					
					newResults = new TreeMap<Integer,TableDataInterpreter.AssertionOutput>();
					int i = 0;
					for (TableDataInterpreter.AssertionOutput ao : output.getResults())
					{
						newResults.put(i,ao);
						i++;
					}
					resultsFile = DEFAULTRESULTSFILE;
					return new String[] {};
				}
				/*catch(java.io.FileNotFoundException e)
				{
					return new String[] {"File Not Found", e.getMessage()};
				}*/
				catch(java.io.IOException e)
				{
					return new String[] {"I/O Error While Running FDR", e.getMessage()};
				}
				catch(InterruptedException e)
				{
					return new String[] {"Interrupted While Running FDR", e.getMessage()};
				}/*
				catch(ParserException e)
				{
					return new String[] {"Error Parsing File", e.getMessage()};
				}
				catch(InterpreterException e)
				{
					return new String[] {"Error Interpreting Results", e.getMessage()};
				}
				catch(Throwable e)
				{
					return new String[] {"Error", e.getMessage()};
				}*/
			}
			
			@Override
    			public void done()
    			{
    				String[] message;
    				try
    				{
    					//if (isCancelled())
    					//	message = new String[] {"Interrupted", "FDR Halted"};
    					//else
    					//{ 
					message = get();
					if (message.length == 0)
					{//success
						results = newResults;
						
						//reenable everything
						assertsList.setEnabled(true);
						updateResult();
						vars.setEnabled(true);
						writes.setEnabled(true);
						atomic.setEnabled(true);
						//runFDR.setEnabled(false);
						load.setEnabled(true);
						reload.setEnabled(true);
						saveCSP.setEnabled(true);
						//interrupt.setEnabled(false);
						//for (ActionListener al : interrupt.getActionListeners())
						//	interrupt.removeActionListener(al);
						
						//set title
						frame.setTitle("SVA - " + fileName);
						
						//remove busy cursor
						frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						

						return;
						
						//remove loading screen
					}
					//}
				}
				catch(InterruptedException e)
				{
					message = new String[] {"Interrupted While Running FDR", e.getMessage()};
				}
				catch(java.util.concurrent.ExecutionException e)
				{
					message = new String[] {"Error While Running FDR", e.getMessage()};
				}
				/*catch(Throwable e)
				{
					message = new String[] {"Error", e.getMessage()};
				}*/
    				//else error!
    				//{
    					JOptionPane.showMessageDialog(frame, message[1], message[0], JOptionPane.WARNING_MESSAGE);
					
					//clear everything
					//data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
					//assertsList.setListData(new String[] {});
					//assertsList.setEnabled(true);
					//vars.setEnabled(true);
					//exps.setEnabled(true);
					//atomic.setEnabled(true);
					runFDR.setEnabled(true);
					load.setEnabled(true);
					reload.setEnabled(true);
					
					saveCSP.setEnabled(true);
					//interrupt.setEnabled(false);
					//for (ActionListener al : interrupt.getActionListeners())
					//		interrupt.removeActionListener(al);
					
					//set title
					frame.setTitle("SVA - Error Running " + fileName);
					
					resultsFile = null;
					
					//remove busy cursor
					frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					return;
					
					//remove loading screen
    				//}
    			}
		
		};
		
		/*interrupt.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							worker.cancel(true);
						}
					});
		interrupt.setEnabled(true);*/
		
		worker.execute();
		
	}
	
	
	/**
	* For thread safety,
	* this method should be invoked from the
	* event-dispatching thread.
	*/
	protected void runOneAssert(final int index)
	{
		//show busy cursor
		frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			
		//change title
		frame.setTitle("SVA - Running FDR - " + fileName);	
			
		//disable everything
		assertsList.setEnabled(false);
		vars.setEnabled(false);
		writes.setEnabled(false);
		atomic.setEnabled(false);
		runFDR.setEnabled(false);
		runOneAssert.setEnabled(false);
		load.setEnabled(false);
		reload.setEnabled(false);
		saveCSP.setEnabled(false);
	
		final SwingWorker worker = new SwingWorker<String[], Void>()
		{
			protected TableDataInterpreter.AssertionOutput newResults;
			@Override
    			public String[] doInBackground()
    			{
     				try
				{
					interpreter.runOneAssert(CSPFILE, index);
					TableDataInterpreter output = new TableDataInterpreter(interpreter.compiler.context.globalConsts, interpreter.compiler.context.dirtyVars);
					
					interpreter.interpretOneAssert(new TraceConstructor(), output, index);
					
					newResults = output.getResults().firstElement();
					
					return new String[] {};
				}
				/*catch(java.io.FileNotFoundException e)
				{
					return new String[] {"File Not Found", e.getMessage()};
				}*/
				catch(java.io.IOException e)
				{
					return new String[] {"I/O Error While Running FDR", e.getMessage()};
				}
				catch(InterruptedException e)
				{
					return new String[] {"Interrupted While Running FDR", e.getMessage()};
				}/*
				catch(ParserException e)
				{
					return new String[] {"Error Parsing File", e.getMessage()};
				}
				catch(InterpreterException e)
				{
					return new String[] {"Error Interpreting Results", e.getMessage()};
				}
				catch(Throwable e)
				{
					return new String[] {"Error", e.getMessage()};
				}*/
			}
			
			@Override
    			public void done()
    			{
    				String[] message;
    				try
    				{
    					//if (isCancelled())
    					//	message = new String[] {"Interrupted", "FDR Halted"};
    					//else
    					//{ 
					message = get();
					if (message.length == 0)
					{//success
						results.put(index,newResults);
						
						//reenable everything
						assertsList.setEnabled(true);
						vars.setEnabled(true);
						writes.setEnabled(true);
						atomic.setEnabled(true);
						load.setEnabled(true);
						reload.setEnabled(true);
						saveCSP.setEnabled(true);
						
						
						if (results.size() < assertsList.getModel().getSize())
						{
							runFDR.setEnabled(true);
						}
						
						//set data
						
						updateResult();
						//data.setData(newResults.procs,newResults.counterexample);
						//result.setText("Result: " + newResults.result);
						//setTableWidths(counterexample);
						
						//set title
						frame.setTitle("SVA - " + fileName);
						
						//remove busy cursor
						frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						
						
						
						return;
						
						//remove loading screen
					}
					//}
				}
				catch(InterruptedException e)
				{
					message = new String[] {"Interrupted While Running FDR", e.getMessage()};
				}
				catch(java.util.concurrent.ExecutionException e)
				{
					message = new String[] {"Error While Running FDR", e.getMessage()};
				}
				/*catch(Throwable e)
				{
					message = new String[] {"Error", e.getMessage()};
				}*/
    				//else error!
    				//{
    					JOptionPane.showMessageDialog(frame, message[1], message[0], JOptionPane.WARNING_MESSAGE);
					
					//clear everything
					//data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
					//assertsList.setListData(new String[] {});
					//assertsList.setEnabled(true);
					//vars.setEnabled(true);
					//exps.setEnabled(true);
					//atomic.setEnabled(true);
					runFDR.setEnabled(true);
					runOneAssert.setEnabled(true);
					load.setEnabled(true);
					reload.setEnabled(true);
					
					saveCSP.setEnabled(true);
					//interrupt.setEnabled(false);
					//for (ActionListener al : interrupt.getActionListeners())
					//		interrupt.removeActionListener(al);
					
					//set title
					frame.setTitle("SVA - Error Running " + fileName);
					
					resultsFile = null;
					
					//remove busy cursor
					frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					return;
					
					//remove loading screen
    				//}
    			}
			

		};
		
		/*interrupt.addActionListener(new ActionListener()
					{
						public void actionPerformed(ActionEvent e)
						{
							worker.cancel(true);
						}
					});
		interrupt.setEnabled(true);*/
		
		worker.execute();
		
	}
	
	protected void saveCSP(final String file)
	{
		SwingWorker worker = new SwingWorker<String[], Void>()
		{
			@Override
			public String[] doInBackground()
			{
				try
				{
					interpreter.compiler.output(new PrintStream(file));
					return new String[] {};
				}
				catch(java.io.FileNotFoundException e)
				{
					return new String[] {"File Not Found While Saving CSP file", e.getMessage()};
				}
				catch(java.io.IOException e)
				{
					return new String[] {"I/O Error While Saving CSP file", e.getMessage()};
				}/*
				catch(InterruptedException e)
				{
					return new String[] {"Interrupted", e.getMessage()};
				}
				catch(ParserException e)
				{
					return new String[] {"Error Parsing File", e.getMessage()};
				}
				catch(InterpreterException e)
				{
					return new String[] {"Error Interpreting Results", e.getMessage()};
				}*/
			}
			
			@Override
    			public void done()
    			{
    				String[] message;
    				try
    				{
					message = get();
					if (message.length == 0)
						return;
				}
				catch(InterruptedException e)
				{
					message = new String[] {"Interrupted While Saving CSP file", e.getMessage()};
				}
				catch (java.util.concurrent.ExecutionException e)
				{
					message = new String[] {"Error While Saving CSP file", e.getMessage()};	
				}
				JOptionPane.showMessageDialog(frame, message[1], message[0], JOptionPane.WARNING_MESSAGE);
    			}
		
		};
		
		worker.execute();
		
	}
	
	
	/**
	* For thread safety,
	* this method should be invoked from the
	* event-dispatching thread.
	*/
	protected void load(final String source,final String CSPFile,final String readFDRFile,final String writeFDRFile)
	{
		//show loading screen?
		
		//show busy cursor
		frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			
		//change title
		frame.setTitle("SVA - Loading " + source);	
		fileName = source;
			
		//clear data	
		/*data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
		assertsList.setListData(new Vector<String>());
		results = new Vector<TableDataInterpreter.AssertionOutput>();
		assertsList.setSelectedIndex(-1);//DOES THIS WORK??
		*/
		//disable everything
		assertsList.setEnabled(false);
		vars.setEnabled(false);
		writes.setEnabled(false);
		atomic.setEnabled(false);
		runOneAssert.setEnabled(false);
		runFDR.setEnabled(false);
		load.setEnabled(false);
		reload.setEnabled(false);
		saveCSP.setEnabled(false);
		results.clear();
		
		SwingWorker worker = new SwingWorker<String[], Void>()
		{
			protected Map<Integer,TableDataInterpreter.AssertionOutput> newResults;
			@Override
    			public String[] doInBackground()
    			{
     				try
				{
					Debug.trace("Doing in background");
					interpreter = new Interpreter(source, CSPFile, readFDRFile, writeFDRFile);		
					TableDataInterpreter output = new TableDataInterpreter(interpreter.compiler.context.globalConsts, interpreter.compiler.context.dirtyVars);
					Debug.trace("interpreting in background");
					interpreter.interpret(new TraceConstructor(), output);
					Debug.trace("getting results in background");		
					newResults = new TreeMap<Integer,TableDataInterpreter.AssertionOutput>();
					int i = 0;
					for (TableDataInterpreter.AssertionOutput ao : output.getResults())
					{
						newResults.put(i,ao);
						i++;
					}
					Debug.trace("got results in background");
					resultsFile = writeFDRFile;
					Debug.trace("Done in background");
					return new String[] {};
				}
				catch(java.io.FileNotFoundException e)
				{
					return new String[] {"File Not Found While Loading", e.getMessage()};
				}
				catch(java.io.IOException e)
				{
					return new String[] {"I/O Error While Loading", e.getMessage()};
				}
				catch(InterruptedException e)
				{
					return new String[] {"Interrupted While Loading", e.getMessage()};
				}
				catch(ParserException e)
				{
					return new String[] {"Error Parsing File", e.getMessage()};
				}
				catch(InterpreterException e)
				{
					return new String[] {"Error Interpreting Results", e.getMessage()};
				}
			}
			
			@Override
    			public void done()
    			{
    				String[] message;
    				try
    				{
    					message = get();
					if (message.length == 0)
					{//success
						results = newResults;
						
						Vector<String> asserts = new Vector<String>();
						for (Integer i : results.keySet())
						{
							asserts.add(results.get(i).assertion);
						}
						
// 						Iterator<TableDataInterpreter.AssertionOutput> i = results.iterator(); 
// 						while (i.hasNext())
// 							asserts.add(i.next().assertion);
							
						assertsList.setListData(asserts);
						//assertsList.setSelectedIndex(-1);//DOES THIS WORK?? not needed anymore?
						
						//reenable everything
						
						assertsList.setEnabled(true);
						vars.setEnabled(true);
						writes.setEnabled(true);
						atomic.setEnabled(true);
						runFDR.setEnabled(true);
						load.setEnabled(true);
						reload.setEnabled(true);
						saveCSP.setEnabled(true);
						//set title
						frame.setTitle("SVA - " + fileName);
						
						
						updateResult();
						//remove busy cursor
						frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						return;
						
						//remove loading screen
					}
				}
				catch(InterruptedException e)
				{
					message = new String[] {"Interrupted While Loading", e.getMessage()};
				}
				catch(java.util.concurrent.ExecutionException e)
				{
					message = new String[] {"Excecution Error While Loading", e.getMessage()};
					e.printStackTrace();
				}
    				//else error!
    				//{
    					JOptionPane.showMessageDialog(frame, message[1], message[0], JOptionPane.WARNING_MESSAGE);
					
					//clear everything
					data.setData(new Vector<String>(), new Vector<TableDataInterpreter.Entry>());
					assertsList.setListData(new String[] {});
					assertsList.setEnabled(true);
					vars.setEnabled(true);
					writes.setEnabled(true);
					atomic.setEnabled(true);
					load.setEnabled(true);
					reload.setEnabled(true);
					resultsFile = null;
					
					//set title
					frame.setTitle("SVA - Error Loading " + fileName);
					
					//remove busy cursor
					frame.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					return;
					
					//remove loading screen
    				//}
    			}
		
		};
		
		worker.execute();
		
	}
	
	void showAbout()
	{
		
		JOptionPane.showMessageDialog(frame,
		"\nSVA : Shared Variable Analyser\n\nBack-end by A.W. Roscoe\nFront-end by D. Hopkins\nUniversity of Oxford\n\n\n",
		"About SVA v1.2a",
		JOptionPane.INFORMATION_MESSAGE,
		new ImageIcon("oxon2.gif"));		

	}

	
	public static void main(String[] args)
	{	
		InterpreterGUI gui = new InterpreterGUI();
		
		if (args.length == 0) 
		{
		
			//System.out.println("Usage: java InterpreterGUI source -r=FDRoutputForReading -w=FDROutputToWrite -c=CSPFileToCreate");
			gui.run();
		}
		else
		{
			String source = args[0];
			String readFDRFile = null;
			String writeFDRFile = null;
	
			String CSPFile = CSPFILE;
			
			gui.run(source,CSPFile,readFDRFile,writeFDRFile);
		}

		fdr.libraryExit();
	}
}
