import java.io.*;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Map;
import java.util.Vector;

import uk.ac.ox.cs.fdr.*;

public class Interpreter
{
	Compiler compiler;	
	Session session;

	public Interpreter(String source, String CSPFile, String fdrReadOutputFile, String fdrWriteOutputFile) throws FileNotFoundException, InterruptedException, IOException
	{
		this(source);
		// if (fdrReadOutputFile != null)
		// {
		// 	loadOutput(fdrReadOutputFile);
		// }
		// else
		{
			runAllAsserts(CSPFile);
		}
	}


	public Interpreter(String source) throws FileNotFoundException
	{
		Debug.trace("Creating parser for " + source);
		Parser parser = new Parser(new FileReader(source));
		compiler = parser.getCompiler();
	}
	
	// public void loadOutput(String fdrReadOutputFile) throws FileNotFoundException
	// {
	// 	Debug.trace("Reading FDR output from " + fdrReadOutputFile);
	// 	// reader = new BufferedReader( new FileReader(fdrReadOutputFile));
	// }
	
	public void runAllAsserts(String CSPFile) throws FileNotFoundException, IOException, InterruptedException
	{
		
		Debug.trace("Creating " + CSPFile);
		compiler.output(new PrintStream(CSPFile));
		
		executeFDR(CSPFile);

	}
	
	public void runOneAssert(String CSPFile, int index) throws FileNotFoundException, IOException, InterruptedException
	{
		Debug.trace("Creating " + CSPFile);
		compiler.outputOneAssert(new PrintStream(CSPFile),index);
		
		executeFDR(CSPFile);
	}
	
	
	private void executeFDR(String CSPFile) throws FileNotFoundException, IOException, InterruptedException
	{
		session = new Session();
		session.loadFile(CSPFile);

		for (uk.ac.ox.cs.fdr.Assertion assertion : session.assertions())
		{
			assertion.execute(null);
		}	
	}
		
	public void interpret(TraceConstructor constructor, InterpreterOutput output) throws IOException
	{
		if (session == null)
			throw new InterpreterException("Trying to interpret file before results are known");
		
		for (int i = 0 ; i < compiler.asserts.size(); ++i)
		{
			Assertion assertion = compiler.asserts.get(i);
			uk.ac.ox.cs.fdr.Assertion fdrAssertion = session.assertions().get(i);
			interpretAssert(assertion, fdrAssertion, constructor, output);
		}
	}
	
	public void interpretOneAssert(TraceConstructor constructor, InterpreterOutput output, int index) throws IOException
	{
		if (session == null)
			throw new InterpreterException("Trying to interpret file before results are known");
		
		Assertion assertion = compiler.asserts.get(index);
		uk.ac.ox.cs.fdr.Assertion fdrAssertion = session.assertions().get(0);
		interpretAssert(assertion, fdrAssertion, constructor, output);
	}

	// boolean isStartOf(Assertion assertion, String line)
	// {
	// 	if ( !line.startsWith("Checking ") || assertion == null )
	// 		return false;
	// 	String s = line.substring(9).replaceAll(" ", "");//remove "Checking " and any spaces
	// 	String t = assertion.genCode(compiler.context).substring(7).replaceAll(" ", ""); //remove "assert " and any spaces
	// 	Debug.trace("Checking if " + s + " matches " + t);
	// 	return (s.equals(t));
	// }
	
	// TraceConstructor constructor, InterpreterOutput output)  throws IOException

	void interpretAssert(Assertion assertion, uk.ac.ox.cs.fdr.Assertion fdrAssertion,
		TraceConstructor constructor, InterpreterOutput output)
	{
		// if (!reader.ready()) throw new InterpreterException("interpreter.interpretAssert: end of stream reached");
		Debug.trace("Interpreting " + assertion);
		if (fdrAssertion.passed()) {
			output.startAssertion(assertion, true);
			return;
		}
		else {
			output.startAssertion(assertion, false);

			if (fdrAssertion.counterexamples().size() == 0)
				throw new InterpreterException("Generated no counterexamples");

			Vector<TraceEvent> trace = constructor.constructTrace(session,
				fdrAssertion.counterexamples().get(0), compiler.context);
			for (TraceEvent event : trace)
			{
				event.output(output);
			}
			
			output.endTrace();
		}
		// String s = reader.readLine();
		// if (s.equals("true"))
		// {
		// 	output.startAssertion(assertion, true);
		// 	return;
		// }
		// else if (s.equals("xfalse"))
		// {
		// 	output.startAssertion(assertion, false);
			
		// 	//Context con = compiler.context;
		// 	Vector<TraceEvent> trace = constructor.constructTrace(assertion, reader, compiler.context);

		// 	for (TraceEvent event : trace)
		// 	{
		// 		event.output(output);
		// 	}
			
		// 	output.endTrace();

		
		// }
	}

}

interface StatusListener
{
	public void write(String s);
}