import java.io.BufferedReader;
import java.util.Vector;

public interface TraceEvent
{
	public void output(InterpreterOutput io);
	public boolean equals(Object obj);
	public int hashCode();
	public String toString();
}

class Tau implements TraceEvent
{

	public void output(InterpreterOutput io){Debug.trace("Ignoring Taus");}
	public boolean equals(Object obj) {return obj instanceof Tau;}
	public int hashCode(){return 1;}
	public String toString(){return "tau";}
}

class IVarRead implements TraceEvent
{
	int process;
	String variable;
	int value;
	
	public IVarRead(int process, String variable, int value)
	{
		this.process = process;
		this.variable = variable;
		this.value = value;
	}
	
	public void output(InterpreterOutput io){io.iVarEval(variable,value,process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof IVarRead)
		{
			IVarRead o = (IVarRead) obj;
			return (process == o.process) && (variable.equals(o.variable)) && (value == o.value);
		}
		else return false;
	}
	public int hashCode(){ return 2 + process + variable.hashCode() + value;}
	public String toString(){return "IVarRead(" + String.valueOf(process) + "," + variable + "," + String.valueOf(value) + ")" ;}
}

class IAVarRead implements TraceEvent
{
	int process;
	String array;
	int index;
	int value;
	
	public IAVarRead(int process, String array, int index, int value)
	{
		this.process = process;
		this.array = array;
		this.index = index;
		this.value = value;
	}
	
	public void output(InterpreterOutput io)
	{
		io.iVarEval(array + "[" + index + "]",value,process);
	}

	public boolean equals(Object obj) 
	{
		if (obj instanceof IAVarRead)
		{
			IAVarRead o = (IAVarRead) obj;
			return (process == o.process) && (array.equals(o.array)) && (value == o.value) && (index == o.index);
		}
		else return false;
	}
	public int hashCode(){ return 3 + process + array.hashCode() + value + index;}
	public String toString(){return "IAVarRead(" + String.valueOf(process) + "," + array + "," + String.valueOf(index) + "," + String.valueOf(value) + ")" ;}
}
class BVarRead implements TraceEvent
{
	int process;
	String variable;
	boolean value;
	
	public BVarRead(int process, String variable, boolean value)
	{
		this.process = process;
		this.variable = variable;
		this.value = value;
	}
	public void output(InterpreterOutput io){io.bVarEval(variable,value,process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof BVarRead)
		{
			BVarRead o = (BVarRead) obj;
			return (process == o.process) && (variable.equals(o.variable)) && (value == o.value);
		}
		else return false;
	}
	public int hashCode(){ return 4 + process + variable.hashCode() + (value ? 2 : 1);}
	public String toString(){return "BVarRead(" + String.valueOf(process) + "," + variable + "," + String.valueOf(value) + ")" ;}
}

class BAVarRead implements TraceEvent
{
	int process;
	String array;
	int index;
	boolean value;
	
	public BAVarRead(int process, String array, int index, boolean value)
	{
		this.process = process;
		this.array = array;
		this.index = index;
		this.value = value;
	}

	public void output(InterpreterOutput io){io.bVarEval(array + "[" + index + "]",value,process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof BAVarRead)
		{
			BAVarRead o = (BAVarRead) obj;
			return (process == o.process) && (array.equals(o.array)) && (value == o.value) && (index == o.index);
		}
		else return false;
	}
	public int hashCode(){ return 5 + process + array.hashCode() + (value ? 2 : 1) + index;}
	public String toString(){return "BAVarRead(" + String.valueOf(process) + "," + array + "," + String.valueOf(index) + "," + String.valueOf(value) + ")" ;}
}

class IVarWrite implements TraceEvent
{
	int process;
	String variable;
	int value;
	
	public IVarWrite(int process, String variable, int value)
	{
		this.process = process;
		this.variable = variable;
		this.value = value;
	}
	public void output(InterpreterOutput io){io.iVarWrite(variable,value,process);}

	public boolean equals(Object obj) 
	{
		if (obj instanceof IVarWrite)
		{
			IVarWrite o = (IVarWrite) obj;
			return (process == o.process) && (variable.equals(o.variable)) && (value == o.value);
		}
		else return false;
	}
	public int hashCode(){ return 6 + process + variable.hashCode() + value;}
	public String toString(){return "IVarWrite(" + String.valueOf(process) + "," + variable + "," + String.valueOf(value) + ")" ;}

}

class IAVarWrite implements TraceEvent
{
	int process;
	String array;
	int index;
	int value;
	
	public IAVarWrite(int process, String array, int index, int value)
	{
		this.process = process;
		this.array = array;
		this.index = index;
		this.value = value;
	}
	
	public void output(InterpreterOutput io){io.iVarWrite(array + "[" + index + "]",value,process);}

	public boolean equals(Object obj) 
	{
		if (obj instanceof IAVarWrite)
		{
			IAVarWrite o = (IAVarWrite) obj;
			return (process == o.process) && (array.equals(o.array)) && (value == o.value) && (index == o.index);
		}
		else return false;
	}
	public int hashCode(){ return 7 + process + array.hashCode() + value + index;}
	public String toString(){return "IAVarWrite(" + String.valueOf(process) + "," + array + "," + String.valueOf(index) + "," + String.valueOf(value) + ")" ;}
}
class BVarWrite implements TraceEvent
{
	int process;
	String variable;
	boolean value;
	
	public BVarWrite(int process, String variable, boolean value)
	{
		this.process = process;
		this.variable = variable;
		this.value = value;
	}
	public void output(InterpreterOutput io){io.bVarWrite(variable,value,process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof BVarWrite)
		{
			BVarWrite o = (BVarWrite) obj;
			return (process == o.process) && (variable.equals(o.variable)) && (value == o.value);
		}
		else return false;
	}
	public int hashCode(){ return 8 + process + variable.hashCode() + (value ? 2 : 1);}
	public String toString(){return "BVarWrite(" + String.valueOf(process) + "," + variable + "," + String.valueOf(value) + ")" ;}
}

class BAVarWrite implements TraceEvent
{
	int process;
	String array;
	int index;
	boolean value;
	
	public BAVarWrite(int process, String array, int index, boolean value)
	{
		this.process = process;
		this.array = array;
		this.index = index;
		this.value = value;
	}
	public void output(InterpreterOutput io){io.bVarWrite(array + "[" + index + "]",value,process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof BAVarWrite)
		{
			BAVarWrite o = (BAVarWrite) obj;
			return (process == o.process) && (array.equals(o.array)) && (value == o.value) && (index == o.index);
		}
		else return false;
	}
	public int hashCode(){ return 9 + process + array.hashCode() + (value ? 2 : 1) + index;}
	public String toString(){return "BAVarWrite(" + String.valueOf(process) + "," + array + "," + String.valueOf(index) + "," + String.valueOf(value) + ")" ;}
}
class StartAt implements TraceEvent
{
	int process;
	
	public StartAt(int process)
	{
		this.process = process;
	}
	public void output(InterpreterOutput io){io.startAtom(process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof StartAt)
		{
			StartAt o = (StartAt) obj;
			return (process == o.process);
		}
		else return false;
	}
	public int hashCode(){ return 10 + process;}
	public String toString(){return "StartAt(" + String.valueOf(process) + ")" ;}
}
class EndAt implements TraceEvent
{
	int process;
	
	public EndAt(int process)
	{
		this.process = process;
	}
	public void output(InterpreterOutput io){io.endAtom(process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof EndAt)
		{
			EndAt o = (EndAt) obj;
			return (process == o.process);
		}
		else return false;
	}
	public int hashCode(){ return 11 + process;}
	public String toString(){return "EndAt(" + String.valueOf(process) + ")" ;}
}
class Signal implements TraceEvent
{
	String sig;
	
	public Signal(String sig)
	{
		this.sig = sig;
	}
	public void output(InterpreterOutput io){io.signal(sig);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof Signal)
		{
			Signal o = (Signal) obj;
			return (sig.equals(o.sig));
		}
		else return false;
	}	
	public int hashCode(){ return 12 + sig.hashCode();}
	public String toString(){return "Sig(" + sig + ")" ;}
}
class ErrorEvent implements TraceEvent
{
	int process;
	
	public ErrorEvent(int process)
	{
		this.process = process;
	}
	public void output(InterpreterOutput io){io.error(process);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof ErrorEvent)
		{
			ErrorEvent o = (ErrorEvent) obj;
			return (process == o.process);
		}
		else return false;
	}
	public int hashCode(){ return 13 + process;}
	public String toString(){return "ErrorEvent(" + String.valueOf(process) + ")" ;}
}
class VErrorEvent implements TraceEvent
{
	String var;
	
	public VErrorEvent(String var)
	{
		this.var = var;
	}
	public void output(InterpreterOutput io){io.verror(var);}
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof VErrorEvent)
		{
			VErrorEvent o = (VErrorEvent) obj;
			return (var.equals(o.var));
		}
		else return false;
	}	
	public int hashCode(){ return 14 + var.hashCode();}
	public String toString(){return "VErrorEvent(" + var + ")" ;}
}
class ISignal implements TraceEvent
{
	String sig;
	int i;
	
	public ISignal(String sig, int i)
	{
		this.sig = sig;
		this.i = i;
	}
	public void output(InterpreterOutput io){io.iSignal(sig,i);}	
	
	public boolean equals(Object obj) 
	{
		if (obj instanceof ISignal)
		{
			ISignal o = (ISignal) obj;
			return (sig.equals(o.sig)) && (i == o.i);
		}
		else return false;
	}	
	public int hashCode(){ return 15 + sig.hashCode() + i;}
	public String toString(){return "ISig(" + sig + "," + String.valueOf(i) + ")" ;}
}

class PStart implements TraceEvent
{
	public void output(InterpreterOutput io){io.pStart();}
	
	public boolean equals(Object obj) 
	{
		return obj instanceof PStart;
	}
	public int hashCode(){ return 16 ;}
	public String toString(){return "PStart" ;}
}

class PEnd implements TraceEvent
{
	public void output(InterpreterOutput io){io.pEnd();}
	
	public boolean equals(Object obj) 
	{
		return obj instanceof PEnd;
	}
	public int hashCode(){ return 17 ;}
	public String toString(){return "PEnd" ;}
}


class TraceEventFactory
{
	static public TraceEvent makeTraceEvent(String event, Context con)
	{
		String[] ss = event.split("\\.");
		Debug.trace("interpreting " + event);
			
		//if (s.startsWith("END TRACE"))
		//	{output.endTrace();break;}
		//else 
		if (event.startsWith("\u03C4"))
		{
			return new Tau();
		}
		else if (ss[0].equals("bveval"))
		{
			if (ss[2].equals("BV"))
				return new BVarRead(Integer.parseInt(ss[1]),con.getBName(Integer.parseInt(ss[3])),Boolean.parseBoolean(ss[4]));
			else if (ss[2].equals("BA"))
				return new BAVarRead(Integer.parseInt(ss[1]),con.getBAName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]),Boolean.parseBoolean(ss[5]));
			else
				throw new InterpreterException("Line in unexpected format:" + event);
		}
		else if (ss[0].equals("bvwrite"))
		{
			if (ss[2].equals("BV"))
				return new BVarWrite(Integer.parseInt(ss[1]),con.getBName(Integer.parseInt(ss[3])),Boolean.parseBoolean(ss[4]));
			else if (ss[2].equals("BA"))
				return new BAVarWrite(Integer.parseInt(ss[1]),con.getBAName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]),Boolean.parseBoolean(ss[5]));
			else
				throw new InterpreterException("Line in unexpected format:" + event);
		}
		else if (ss[0].equals("iveval"))
		{
			if (ss[2].equals("IV"))
				return new IVarRead(Integer.parseInt(ss[1]),con.getIName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]));
			else if (ss[2].equals("IA"))
				return new IAVarRead(Integer.parseInt(ss[1]),con.getIAName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]),Integer.parseInt(ss[5]));
			else
				throw new InterpreterException("Line in unexpected format:" + event);
		}
		else if (ss[0].equals("ivwrite"))
		{
			if (ss[2].equals("IV"))
				return new IVarWrite(Integer.parseInt(ss[1]),con.getIName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]));
			else if (ss[2].equals("IA"))
				return new IAVarWrite(Integer.parseInt(ss[1]),con.getIAName(Integer.parseInt(ss[3])),Integer.parseInt(ss[4]),Integer.parseInt(ss[5]));
			else
				throw new InterpreterException("Line in unexpected format:" + event);
		}
// 		else if (ss[0].equals("breq"))
// 		{
// 			output.bExprStart(Integer.parseInt(ss[2]));
// 		}
// 		else if (ss[0].equals("ireq"))
// 		{
// 			output.iExprStart(Integer.parseInt(ss[2]));
// 		}
// 		else if (ss[0].equals("beval"))
// 		{
// 			output.bExprEnd(Integer.parseInt(ss[2]),Boolean.parseBoolean(ss[3]));
// 		}
// 		else if (ss[0].equals("ieval"))
// 		{
// 			output.iExprEnd(Integer.parseInt(ss[2]),Integer.parseInt(ss[3]));
// 		}
		else if (event.equals("PStart"))
		{
			return new PStart();
		}
		else if (event.equals("PEnd"))
		{
			return new PEnd();
		}
		else if (ss[0].equals("start_at"))
		{
			return new StartAt(Integer.parseInt(ss[1]));
		}
		else if (ss[0].equals("end_at"))
		{
			return new EndAt(Integer.parseInt(ss[1]));
		}
		else if (ss[0].equals("error"))
		{
			return new ErrorEvent(Integer.parseInt(ss[1]));
		}
		else if (ss[0].equals("verror"))
		{
			if (ss[1].equals("IV"))
			{
				return new VErrorEvent(con.getIName(Integer.parseInt(ss[2])));
			}
			else if (ss[1].equals("IA"))
			{
				return new VErrorEvent(con.getIAName(Integer.parseInt(ss[2])) + "[" + ss[3] + "]");
			}
			if (ss[1].equals("BV"))
			{
				return new VErrorEvent(con.getBName(Integer.parseInt(ss[2])));
			}
			else if (ss[1].equals("BA"))
			{
				return new VErrorEvent(con.getBAName(Integer.parseInt(ss[2])) + "[" + ss[3] + "]");
			}
			else
				throw new InterpreterException("Line in unexpected format:" + event);
		}
		else if (con.isSig(ss[0]) /*|| event.equals("outofrange")*/ || event.equals("assertionfailed") /*|| ss[0].equals("error") || ss[0].equals("verror")*/ )
		{
			return new Signal(event);
		}
		else if (ss.length > 1 && con.isISig(ss[0]))
		{
			return new ISignal(event.substring(0, event.length() - (ss[ss.length-1].length() + 1)), Integer.parseInt(ss[ss.length -1]));
		}
		else
		{
			throw new InterpreterException("Unexpected event \"" + event + "\"");
		}
	}
	
	//Constructs a sequence of traceevents, corresponding to the input
	//read from the reader up until the end of the current trace.
	//Errors will occur if any line before the "END TRACE" does
	//not correcspond to an event
	public static Vector<TraceEvent> readTrace(BufferedReader reader, Context con) throws java.io.IOException
	{
		Vector<TraceEvent> trace = new Vector<TraceEvent>();
		
		while (reader.ready())
		{
			String line = reader.readLine();
			if (line.startsWith("END TRACE") || line.startsWith("END ALLOWS"))
				return trace;
			else if (!line.equals(""))
				trace.add(makeTraceEvent(line,con));
		}
		throw new InterpreterException("TraceEventFactory.readTrace: Reached end of stream before reaching end of trace");
	}
	private TraceEventFactory(){}
}