import java.util.*;

public interface Assertion  extends ProcessContainer
{
//All programs can invlove minus one so below procedure is no longer required
// 	boolean usesMinusOne(); //whether this assertion uses the process index -1 (refinement assertions do, others don't)
}

class BoolAssertion implements Assertion
{
	Program prog;
	String progName, condition;
	
	
	public BoolAssertion (Program p, BExpr b)
	{
		progName = p.getName();
		condition= b.toString();
		
		//this next line doesnt work. it creates a new vector, but the inner vectors are the same ones.
		//Vector<Vector<CompilableProc>> procs = new Vector<Vector<CompilableProc>>(p.procs);
		
		/*Vector<Vector<CompilableProc>> procs = new Vector<Vector<CompilableProc>>();
		Iterator<Vector<CompilableProc>> i = p.procs.iterator();
		while (i.hasNext())
			procs.add(new Vector<CompilableProc>(i.next()));*/
		
		prog = p.unnamedCopy();
		
		//put the assertion in the top priority level. 
		//for efficiency we do not create a new priority level specially for it
		//(unless the list of processes is empty
		
		//if (prog.procs.size() > 1) //prog.procs./*firstElement().*/add(new PriorityAssert( b ));
		//else if (prog.procs.size() == 1) prog.procs./*firstElement().*/add(new SimpleAssert(b));
		//	prog.procs.add(new SimpleAssert(b));
		//else throw new ParserException("BoolAssertion - Process List Empty");	
		
		prog.addProcess(new SimpleAssert(b));	
		//prog = new UnnamedProgram(procs);
	}
	
	
	//The Assert process must go in the highest priority level as we always want it to be able to make its check
	//If there are lower levels though it must generate noop events or it will stop them from running
	//otherwise we can just b once and since fdr explores every possible execution path it will pick up on 
	//any 1 path that fails the check.
// 	class PriorityAssert extends Iter
// 	{
// 		public PriorityAssert(BExpr b) {super(new Seq(new Atomic(new If(new Not(b), new Sig("assertionfailed"))),new NoOp()));}
// 		
// 		public String toString() {return "assertion";}
// 	}
	class SimpleAssert implements CompilableProc
	{
		Cmd assertion;
		public SimpleAssert(BExpr b) {assertion = new Atomic(new If(b, new Skip(), new Sig("assertionfailed", new Vector<IConst>())));}
		
		public String toString() {return "assertion";}

		public String genCode(Context c) 
		{ 
			return "(" + assertion.genCode(c) + ",  ({},{}))";
		}
		public boolean usesSig(String sig,Map<String,Integer> consts){return assertion.usesSig(sig,consts);}
		
		public Vector<LocalVarUsage> getLocalIVars(){return new Vector<LocalVarUsage>();}
		public Vector<LocalVarUsage> getLocalBVars(){return new Vector<LocalVarUsage>();}
		public Vector<LocalVarUsage> getLocalIAVars(){return new Vector<LocalVarUsage>();}
		public Vector<LocalVarUsage> getLocalBAVars(){return new Vector<LocalVarUsage>();}
	}
	
	
	
	public String genCode(Context con)
	{
		return "assert CHAOS(diff(Events, {assertionfailed})) [T= " + prog.genCode(con);
	}
	
	
	
	
// 	public int numIEs(Map<String,Integer> consts){return prog.numIEs(consts);}
// 	public int numBEs(Map<String,Integer> consts){return prog.numBEs(consts);}
	public int numProcs() {return prog.numProcs();}

// 	public String getIExpr(int i, Map<String, Integer> ivals) {return prog.getIExpr(i,ivals);}
// 	public String getBExpr(int i, Map<String, Integer> ivals,Map<String,Boolean> bvals) {return prog.getBExpr(i,ivals,bvals);}
	
	public String toString()
	{ return "assert always " + condition + " in " + progName;}


	
	public String process(int i) {return prog.process(i);}
// 	public String ownerOfIExpr(int i,Map<String,Integer> consts) { return prog.ownerOfIExpr(i,consts);}
// 
// 	public String ownerOfBExpr(int i,Map<String,Integer> consts) {return prog.ownerOfBExpr(i,consts);}
// 
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts){return prog.indexOfBExprOwner(i,consts);}
// 	
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts){return prog.indexOfIExprOwner(i,consts);}

	public String ownerOfSig(String sig,Map<String,Integer> consts){return prog.ownerOfSig(sig,consts);}
	public int indexOfOwnerOfSig(String sig,Map<String,Integer> consts){return prog.indexOfOwnerOfSig(sig,consts);}
	public boolean usesSig(String sig,Map<String,Integer> consts){return prog.usesSig(sig,consts);}
	
	public Iterator<CompilableProc> iterator(){return prog.iterator();}

// 	public boolean usesMinusOne() {return false;}
}

class SignalAssertion implements Assertion
{
	Program prog;
	String signals;
	String renamedSignals;
	
	public SignalAssertion(Program prog, Vector<String> signals)
	{
		if (prog == null) throw new ParserException("prog argument to SignalAssertion null");
		if (signals == null) throw new ParserException("signals argument to SignalAssertion null");
		
		this.prog = prog;
		
		this.signals = "{|";
		//renamedSignals = "{|";
		Iterator<String> i = signals.iterator();
		while (i.hasNext())
		{
			String sig = i.next();
			this.signals = this.signals + sig;
			//String renamedSig = sig.contains(".") ? sig.replaceFirst("\\.","T.") : sig + "T";
			
			
			//renamedSignals = renamedSignals + renamedSig;
			
			if (i.hasNext())
				{this.signals = this.signals + ","; /*renamedSignals = renamedSignals + ",";*/}
		}
		this.signals = this.signals + "|}";
		//renamedSignals += "|}";
	}
	
	public String genCode(Context con)
	{
		return "assert CHAOS(diff(Events," + signals /*renamedSignals*/ + ")) [T= " + prog.getName();
	}

	
	
// 	public int numIEs(Map<String,Integer> consts){return prog.numIEs(consts);}
// 	public int numBEs(Map<String,Integer> consts){return prog.numBEs(consts);}
	public int numProcs() {return prog.numProcs();}


// 	public String getIExpr(int i, Map<String, Integer> ivals) {return prog.getIExpr(i,ivals);}
// 	public String getBExpr(int i, Map<String, Integer> ivals,Map<String,Boolean> bvals) {return prog.getBExpr(i,ivals,bvals);}
	
	
	public String toString()
	{
		return "assert nosignal " + signals + " in " + prog.getName();
	}
	
	public String process(int i) {return prog.process(i);}

	public String ownerOfSig(String sig,Map<String,Integer> consts){return prog.ownerOfSig(sig,consts);}
	public int indexOfOwnerOfSig(String sig,Map<String,Integer> consts){return prog.indexOfOwnerOfSig(sig,consts);}
	public boolean usesSig(String sig,Map<String,Integer> consts){return prog.usesSig(sig,consts);}
// 	public String ownerOfIExpr(int i, Map<String,Integer> consts) { return prog.ownerOfIExpr(i,consts);}
// 
// 	public String ownerOfBExpr(int i, Map<String,Integer> consts) {return prog.ownerOfBExpr(i, consts);}
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts){return prog.indexOfBExprOwner(i,consts);}
// 	
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts){return prog.indexOfIExprOwner(i,consts);}
	public Iterator<CompilableProc> iterator(){return prog.iterator();}
	
// 	public boolean usesMinusOne() {return false;}
}
	
class CSPAssertion implements Assertion
{
	String csp;
	Program prog;

	public CSPAssertion(String csp, Program prog){this.csp = csp; this.prog = prog;}


	public int numProcs(){return prog.numProcs();}

	public String process(int i){return prog.process(i);}//returns a string representing the ith process

	public String ownerOfSig(String sig,Map<String,Integer> consts){return prog.ownerOfSig(sig,consts);}
	public int indexOfOwnerOfSig(String sig, Map<String,Integer> consts){return prog.indexOfOwnerOfSig(sig,consts);}
// 	public String ownerOfIExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith IExpr
// 
// 	public String ownerOfBExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith BExpr
// 
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts);
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts);

	//it should hold that process( indexOfIExprOwner(i,consts)) == ownerOfIExpr(i, consts)
	//and similarly for BExprs

	public Iterator<CompilableProc> iterator(){return prog.iterator();}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return prog.usesSig(sig,consts);}
	
	public String genCode(Context con){return "assert " + csp;}


	public String toString()
	{
		return "assert " + csp;
	}
	
// 	public boolean usesMinusOne() {return false;}//may need to change this if it turns out people do want to use CSP to reason about the refinement processes
}	


class SeqRefinementAssertion implements Assertion
{
	CompilableProc spec; 
	CompilableProc impl;

	public SeqRefinementAssertion(CompilableProc spec, CompilableProc impl){this.spec = spec; this.impl = impl;}


	public int numProcs(){return 1;}

	public String process(int i)
	{
		if (i == 0) 
			return impl.toString(); 
// 		else if (i == -1) 
// 			return "Environment";
		else 
			throw new InterpreterException("SeqRefinementAssertion asked process " + i);
	}//returns a string representing the ith process

	public String ownerOfSig(String sig,Map<String,Integer> consts)
	{
		if (impl.usesSig(sig,consts))
			return impl.toString(); 
		else 
			throw new InterpreterException("SeqRefinementAssertion asked for owner of unused sig " + sig);
	}
	
	public int indexOfOwnerOfSig(String sig, Map<String,Integer> consts)
	{
		if (impl.usesSig(sig,consts))
			return 0;
		else 
			throw new InterpreterException("SeqRefinementAssertion asked for owner of unused sig " + sig);
	} 
// 	public String ownerOfIExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith IExpr
// 
// 	public String ownerOfBExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith BExpr
// 
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts);
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts);

	//it should hold that process( indexOfIExprOwner(i,consts)) == ownerOfIExpr(i, consts)
	//and similarly for BExprs

	public Iterator<CompilableProc> iterator()
	{
		return new Iterator<CompilableProc>()
		{
			public boolean hasNext = true;
			public boolean hasNext(){return hasNext;}
			public CompilableProc next()
			{ 
				if (hasNext)
				{
					hasNext = false;
					return impl;
				}
				else throw new NoSuchElementException();
			}
			public void remove() {throw new UnsupportedOperationException();}
		};
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return impl.usesSig(sig,consts);}
	
	public String genCode(Context con){return "assert compress(SeqRefTest(fst("+spec.genCode(con)+"))) [T= SeqRefTest(fst("+impl.genCode(con)+"))";}


	public String toString()
	{
		return "assert " + spec.toString() + " [S= " + impl.toString();
	}

	
// 	public boolean usesMinusOne() {return true;}
}
	
	

class GenRefinementAssertion implements Assertion
{
	CompilableProc spec; 
	CompilableProc impl;

	public GenRefinementAssertion(CompilableProc spec, CompilableProc impl){this.spec = spec; this.impl = impl;}


	public int numProcs(){return 1;}

	public String process(int i)
	{
		if (i == 0) 
			return impl.toString(); 
// 		else if (i == -1) 
// 			return "Environment"; 
		else 
			throw new InterpreterException("GenRefinementAssertion asked process " + i);
	}//returns a string representing the ith process

	public String ownerOfSig(String sig,Map<String,Integer> consts)
	{
		if (impl.usesSig(sig,consts))
			return impl.toString(); 
		else 
			throw new InterpreterException("GenRefinementAssertion asked for owner of unused sig " + sig);
	}
	public int indexOfOwnerOfSig(String sig, Map<String,Integer> consts)
	{
		if (impl.usesSig(sig,consts))
			return 0;
		else 
			throw new InterpreterException("GenRefinementAssertion asked for owner of unused sig " + sig);
	} 
// 	public String ownerOfIExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith IExpr
// 
// 	public String ownerOfBExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith BExpr
// 
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts);
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts);

	//it should hold that process( indexOfIExprOwner(i,consts)) == ownerOfIExpr(i, consts)
	//and similarly for BExprs

	public Iterator<CompilableProc> iterator()
	{
		return new Iterator<CompilableProc>()
		{
			public boolean hasNext = true;
			public boolean hasNext(){return hasNext;}
			public CompilableProc next()
			{ 
				if (hasNext)
				{
					hasNext = false;
					return impl;
				}
				else throw new NoSuchElementException();
			}
			public void remove() {throw new UnsupportedOperationException();}
		};
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return impl.usesSig(sig,consts);}
	
	public String genCode(Context con){return "assert compress(GenRefTest(fst("+spec.genCode(con)+"))) [T= GenRefTest(fst("+impl.genCode(con)+"))";}


	public String toString()
	{
		return "assert " + spec.toString() + " [G= " + impl.toString();
	}


	
// 	public boolean usesMinusOne() {return true;}
}
	
class ParRefinementAssertion implements Assertion
{
	//We really just want the CmdStruct part, but for toString we want them to have a name.
	//Hence it was simpler to store them as CmdStructPrograms although it is a little ugly and hackish.
	CmdStructProgram spec; 
	CmdStructProgram impl;

	public ParRefinementAssertion(CmdStructProgram spec, CmdStructProgram impl)
	{
		if (spec == null) throw new ParserException("ParRefinementAssertion - spec == null");
		if (impl == null) throw new ParserException("ParRefinementAssertion - impl == null");
		this.spec = spec; 
		this.impl = impl;
	}


	public int numProcs(){return impl.numProcs();}

	public String process(int i)
	{
// 		if (i == -1) 
// 			return "Environment"; 
// 		else 
			return impl.process(i);
	}//returns a string representing the ith process

	public String ownerOfSig(String sig,Map<String,Integer> consts)
	{
		return impl.ownerOfSig(sig,consts);
	}
	public int indexOfOwnerOfSig(String sig, Map<String,Integer> consts)
	{
		return impl.indexOfOwnerOfSig(sig,consts);
	} 
// 	public String ownerOfIExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith IExpr
// 
// 	public String ownerOfBExpr(int i,Map<String,Integer> consts);//returns a string representing the process containing the ith BExpr
// 
// 	public int indexOfBExprOwner(int i,Map<String,Integer> consts);
// 	public int indexOfIExprOwner(int i,Map<String,Integer> consts);

	//it should hold that process( indexOfIExprOwner(i,consts)) == ownerOfIExpr(i, consts)
	//and similarly for BExprs

	public Iterator<CompilableProc> iterator()
	{
		return impl.iterator();
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return impl.usesSig(sig,consts);}
	
	public String genCode(Context con){return "assert ParRefTest(" + spec.CS.genCode(con)+") [T= ParRefTest(" + impl.CS.genCode(con)+")";}


	public String toString()
	{
		return "assert " + spec.toString() + " [P= " + impl.toString();
	}


	
// 	public boolean usesMinusOne() {return true;}
}
	
	
	
	
	
	
	
	
