import java.util.Map;
import java.util.Vector;

public interface BExpr extends Genable 
{ 	
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals);
}

class Not implements BExpr
{
	BExpr b;
	public Not(BExpr b) 
	{if (b==null) throw new ParserException("Argument to Not null");
	this.b=b;}
	
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new Not(b.substituteConst(vars,vals));
	}
	public String genCode(Context con)
	{
		return "Not." + b.genCode(con);
	}

	public String toString()
	{
		return "!(" + b + ")";
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Not) ) return false;
		
		Not that = (Not)aThat;

		return ( this.b.equals(that.b));
	}
	@Override public int hashCode() { 
		return b.hashCode() + 1;
	}
	
}

/*abstract*/ class BinBExpr implements BExpr
{
	String op, print;
	BExpr b1,b2;
	BinBExpr(String op, String print, BExpr b1, BExpr b2)
	{
		if (op==null) throw new ParserException("op Argument to BinBExp null");
		if (print==null) throw new ParserException("print Argument to BinBExpr(" + op +") null");
		if (b1==null) throw new ParserException("b1 Argument to BinBExp (" + op +") null");
		if (b2==null) throw new ParserException("b2 Argument to BinBExp (" + op +")  null");
		this.op=op;
		this.print = print;
		this.b1=b1;
		this.b2=b2;
		
	}
	
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new BinBExpr(op,print, b1.substituteConst(vars,vals),b2.substituteConst(vars,vals));
	}
	public String genCode(Context con) {return "BBOp." + op + "." + b1.genCode(con) + "." + b2.genCode(con); }

	
	public String toString()
	{ return b1 + print + b2;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BinBExpr) ) return false;
		
		BinBExpr that = (BinBExpr)aThat;

		return ( this.b1.equals(that.b1)
			&& this.b2.equals(that.b2));
	}
	@Override public int hashCode() { 
		return 1 + b1.hashCode()
			+ b2.hashCode();
	}
}


/*abstract*/ class CompExpr implements BExpr
{
	String op, print;
	IExpr i1,i2;
	CompExpr(String op, String print, IExpr i1, IExpr i2)
	{
		if (op==null) throw new ParserException("op Argument to CompExp null");
		if (print==null) throw new ParserException("print Argument to CompExpr(" + op +") null");
		if (i1==null) throw new ParserException("i1 Argument to CompExp (" + op +") null");
		if (i2==null) throw new ParserException("i2 Argument to CompIExp (" + op +")  null");
		this.op=op;
		this.print = print;
		this.i1=i1;
		this.i2=i2;
		
	}
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new CompExpr(op,print, i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));
	}
	
	public String genCode(Context con) {return "CompOp." + op + "." + i1.genCode(con) + "." + i2.genCode(con); }

	
	public String toString()
	{ return i1 + print + i2;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof CompExpr) ) return false;
		
		CompExpr that = (CompExpr)aThat;

		return ( this.i1.equals(that.i1)
			&& this.i2.equals(that.i2)
			&& this.print.equals(that.print)
			&& this.op.equals(that.op));
	}
	@Override public int hashCode() { 
		return 1 + i1.hashCode()
			+ i2.hashCode()
			+ print.hashCode()
			+ op.hashCode();
	}

}

class And extends BinBExpr
{
	public And(BExpr b1, BExpr b2)
	{
		super("And","&&",b1,b2);
	}
	
	
}

class Or extends BinBExpr
{
	public Or(BExpr b1, BExpr b2)
	{
		super("Or","||",b1,b2);
	}
	
	
}

class XOr extends BinBExpr
{
	public XOr(BExpr b1, BExpr b2)
	{
		super("Xor","^",b1,b2);
	}
	
	
}


class Eq extends CompExpr
{
	public Eq(IExpr i1, IExpr i2)
	{
		super("Eq","=",i1,i2);
	}
}
class Neq extends CompExpr
{
	public Neq(IExpr i1, IExpr i2)
	{
		super("Neq","!=",i1,i2);
	}
}
class Lt extends CompExpr
{
	public Lt(IExpr i1, IExpr i2)
	{
		super("Lt","<",i1,i2);
	}
}
class Gt extends CompExpr
{
	public Gt(IExpr i1, IExpr i2)
	{
		super("Gt",">",i1,i2);
	}
}
class Le extends CompExpr
{
	public Le(IExpr i1, IExpr i2)
	{
		super("Le","<=",i1,i2);
	}
}
class Ge extends CompExpr
{
	public Ge(IExpr i1, IExpr i2)
	{
		super("Ge",">=",i1,i2);
	}
}
	
class True implements BExpr
{
	public final int ies = 0;
	public final int bes = 0;
	public String genCode(Context con) {return "True";}	

	public String toString()
	{ return "true";}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof True) ) return false;
		
		True that = (True)aThat;

		return true;
	}
	@Override public int hashCode() { 
		return 1;
	}
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
	
}
class False implements BExpr
{
	public final int ies = 0;
	public final int bes = 0;
	public String genCode(Context con) {return "False";}

	
	public String toString()
	{ return "false";}
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof False) ) return false;
		
		False that = (False)aThat;

		return true;
	}
	@Override public int hashCode() { 
		return 2;
	}
	
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
}

class BVar implements BExpr
{
	String name;
	public BVar(String name)
	{
		if (name==null) throw new ParserException("Argument to Bvar null");
		this.name = name;
	}
	
	public String genCode(Context con) 
	{ 
		return "BVar." + con.getBEncoding(name); //con should throw an exception if name is not in con
	}

	public String toString() {return name;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BVar) ) return false;
		
		BVar that = (BVar)aThat;

		return ( this.name.equals(that.name));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode();
	}
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
}

class BArray implements BExpr
{
	String name;
	IExpr i;
	public BArray(String name, IExpr i)
	{
		if (name==null) throw new ParserException("name argument to BArray null"); 
		if (i==null) throw new ParserException("index argument to BArray null"); 
		
		this.name = name; this.i = i;
	}

	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new BArray(name, i.substituteConst(vars,vals));
	}
	
	public String genCode(Context con) 
	{ 
		return "BArc.(" + con.getBAEncoding(name) + "," + i.genCode(con) + ")";
	}

	
	public String toString()
	{ return name + "[" + i + "]";}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BArray) ) return false;
		
		BArray that = (BArray)aThat;

		return ( this.name.equals(that.name)
			&& this.i.equals(that.i));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ i.hashCode();
	}
	
}


class LocalBVar implements BExpr
{
	String name,process;
	public LocalBVar(String name,String process)
	{
		if (name==null) throw new ParserException("Argument to LocalBVar null");
		if (process==null) throw new ParserException("Argument to LocalBVar null");
		this.name = name;
		this.process = process;
	}
	
	public String genCode(Context con) 
	{ 
		String result = "BVar.LocalBVarMapping_" + process + "_" + name + "(";
		for (String s : con.consts)
			result += s + ",";
			
		if (result.endsWith(",")) 
			result = result.substring(0, result.length() -1);
			
		return result + ")";
	}

	public String toString() {return name;}

	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof LocalBVar) ) return false;
		
		LocalBVar that = (LocalBVar)aThat;

		return ( this.name.equals(that.name)
			&& this.process.equals(that.process));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ process.hashCode();
	}
	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
}

class LocalBArray implements BExpr
{
	String name,process;
	IExpr i;
	public LocalBArray(String name, String process, IExpr i)
	{
		if (name==null) throw new ParserException("name argument to LocalBArray null"); 
		if (process==null) throw new ParserException("process argument to LocalBArray null"); 
		if (i==null) throw new ParserException("index argument to LocalBArray null"); 
		
		this.name = name; this.i = i; this.process = process;
	}

	public BExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new LocalBArray(name, process, i.substituteConst(vars,vals));
	}
	public String genCode(Context con) 
	{ 
		String genName = "LocalBAVarMapping_" + process + "_" + name + "(";
		for (String s : con.consts)
			genName += s + ",";
			
		if (genName.endsWith(",")) 
			genName = genName.substring(0, genName.length() -1);

	
	
		return "BArc.(" + genName + ")," + i.genCode(con) + ")";
	}

	public String toString() {return name + "[" + i + "]";}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof LocalBArray) ) return false;
		
		LocalBArray that = (LocalBArray)aThat;

		return ( this.name.equals(that.name)
			&& this.i.equals(that.i)
			&& this.process.equals(that.process));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ i.hashCode()
			+ process.hashCode();
	}
}