import java.io.*;
import java.util.*;


public class Compiler
{
	Context context; //keeps track of variable names for entire program
	Map<String, Process> procs;
	Vector<String> csp;//needs to be a vector so its in order, so asserts occur in the correct order
	Map<String, Program> progs;
	Vector<Assertion> asserts;
	
	public Compiler()
	{
		context = new Context();
		procs = new TreeMap<String, Process> ();
		csp = new Vector<String>();
		progs = new TreeMap<String, Program>();
		asserts = new Vector<Assertion>();
	}


	public static void main(String args[]) throws IOException
	//reads args[0] outputs to args[1] or std output
	{
		if ( args.length > 0 ) 
		{
			Parser parser;
			// parse a file
			parser = new Parser(new FileReader(args[0]));
			PrintStream stream = (args.length > 1) ? new PrintStream( args[1]) : System.out;
			Compiler compiler = parser.getCompiler();
			compiler.output(stream);
			stream.close();
		}

	}
   
	public void generateLocalVarNames()
	{
		Debug.trace("compiler.generateLocalVarNames - IVars");
					
		Vector<LocalVarUsage> vars = new Vector<LocalVarUsage>();
		for (Assertion a : asserts)
		{
			Debug.trace("\tIVars - Assertion " + a);
					
			for (CompilableProc p : a)
			{
				Debug.trace("\t\tCompilableProc " + p );
					
				for (LocalVarUsage lvu : p.getLocalIVars())
				{
					Debug.trace("\t\t\tLocalVarUsage " + lvu);
					
					if (! vars.contains(lvu))
					{
						vars.add(lvu);
					}
				}
			}
			
		}
		context.addFinalLocalIVars(vars);
		
		vars = new Vector<LocalVarUsage>();
		for (Assertion a : asserts)
		{
			for (CompilableProc p : a)
			{
				for (LocalVarUsage lvu : p.getLocalBVars())
				{
					Debug.trace("compiler.generateLocalVarNames\n\tBVars - Assertion " + a + " CompilableProc " + p +" LocalVarUsage " + lvu);
					if (! vars.contains(lvu))
					{
						vars.add(lvu);
					}
				}
			}
			
		}
		context.addFinalLocalBVars(vars);
		
		vars = new Vector<LocalVarUsage>();
		for (Assertion a : asserts)
		{
			for (CompilableProc p : a)
			{
				for (LocalVarUsage lvu : p.getLocalIAVars())
				{
					Debug.trace("compiler.generateLocalVarNames\n\tIAVars - Assertion " + a + " CompilableProc " + p + " LocalVarUsage " + lvu);				
					if (! vars.contains(lvu))
					{
						vars.add(lvu);
					}
				}
			}
			
		}
		context.addFinalLocalIAVars(vars);
		
		vars = new Vector<LocalVarUsage>();
		for (Assertion a : asserts)
		{
			for (CompilableProc p : a)
			{
				for (LocalVarUsage lvu : p.getLocalBAVars())
				{
					Debug.trace("compiler.generateLocalVarNames\n\tBAVars - Assertion " + a + " CompilableProc " + p +" LocalVarUsage " + lvu);				
					if (! vars.contains(lvu))
					{
						vars.add(lvu);
					}
				}
			}
			
		}
		context.addFinalLocalBAVars(vars);
		
	}

   
   public void output(PrintStream stream) throws IOException
   {

		//stream.println("include \"share2priorities.csp\"\n");
		stream.println("include \"refsva.csp\"\n");
		stream.println(context.genCode());
		//stream.println("nie = " + numIEs());
		//stream.println("nbe = " + numBEs());
		stream.println("mostprocs = " + numProcs());
		for (Iterator<String> i = csp.iterator(); i.hasNext(); stream.println(i.next())) {}
		for (Iterator<String> i = procs.keySet().iterator(); i.hasNext(); stream.println(procs.get(i.next()).genCode(context)) ) {}
		for (Iterator<String> i = progs.keySet().iterator(); i.hasNext(); stream.println(progs.get(i.next()).genCode(context)) ) {}
		for (Iterator<Assertion> i = asserts.iterator(); i.hasNext(); stream.println(i.next().genCode(context)) ) {}
		stream.flush();   	
   }   
   
   public void outputOneAssert(PrintStream stream, int index) throws IOException
   {

		//stream.println("include \"share2priorities.csp\"\n");
		stream.println("include \"refsva.csp\"\n");
		stream.println(context.genCode());
		//stream.println("nie = " + numIEs());
		//stream.println("nbe = " + numBEs());
		stream.println("mostprocs = " + numProcs());
		for (Iterator<String> i = csp.iterator(); i.hasNext(); stream.println(i.next())) {}
		for (Iterator<String> i = procs.keySet().iterator(); i.hasNext(); stream.println(procs.get(i.next()).genCode(context)) ) {}
		for (Iterator<String> i = progs.keySet().iterator(); i.hasNext(); stream.println(progs.get(i.next()).genCode(context)) ) {}
		//for (Iterator<Assertion> i = asserts.iterator(); i.hasNext(); stream.println(i.next().genCode(context)) ) {}
		stream.println(asserts.get(index).genCode(context));
		stream.flush();   	
   }


	 
   public int numProcs()
   {
   	int num = 0;
   	Iterator<Assertion> i = asserts.iterator();
   	while (i.hasNext())
   	{
   		int j = i.next().numProcs();
   		num = (j > num)? j : num;
   	}
   	Iterator<String> k = progs.keySet().iterator();
   	while (k.hasNext())
   	{
   		int j = progs.get(k.next()).numProcs();
   		num = (j > num)? j : num;
   	}
   	return num;
   }

}