import java.util.Vector;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.Set;
import java.util.TreeSet;

//A Context keeps track of the names and types of everything
public class Context implements Genable
{
	//final boolean DEBUG = false;
	Vector<String> is,bs,ias,bas;//names of integer/boolean variables/arrays in order
	Collection<String> consts;//currently registered constants.  Will be replaced each time context is used by a different process (since constants are local to a process)
	Vector<String> localIs;//currently registered local variables.
	Vector<String> localBs;
	Vector<String> localIAs;
	Vector<String> localBAs;
	
	Vector<LocalVarUsage> finalLocalIs; // the collection of the different local integers that are used in the whole file	
	Vector<LocalVarUsage> finalLocalBs; // the collection of the different local integers that are used in the whole file	
	Vector<LocalVarUsage> finalLocalIAs; // the collection of the different local integers that are used in the whole file	
	Vector<LocalVarUsage> finalLocalBAs; // the collection of the different local integers that are used in the whole file
	
	Map<String, Integer> globalConsts;//the global constants

	Map<String, Integer> contentSizes,arrayContentSizes,arrayIndexSizes;

	Map<String,Integer> sigs; // associtate with each signal the number of constant indices it uses
	Map<String,Integer> isigs;
	
	Map<String, Integer> iinits;
	Map<String, Boolean> binits;
	Map<String, Vector<Integer>> iainits;
	Map<String, Vector<Boolean>> bainits;
	//Map<String, Integer> procs;// a mapping from the names of processes to the number of arguments they take
	Collection<String>procs;
	Collection<String>progs;//the names of programs

	Collection<String>dirtyVars;

	Collection<String>parReads,parWrites,seqReads,seqWrites;

	public Context()
	{
		is = new Vector<String>();
		bas = new Vector<String>();
		ias = new Vector<String>();
		bs = new Vector<String>(); 
		consts = new Vector<String>();
		localIs = new Vector<String>();
		localBs = new Vector<String>();
		localIAs = new Vector<String>();
		localBAs = new Vector<String>();
		globalConsts = new TreeMap<String, Integer>();
		contentSizes = new TreeMap<String, Integer>();
		arrayContentSizes = new TreeMap<String, Integer>();
		arrayIndexSizes = new TreeMap<String, Integer>();
		sigs = new TreeMap<String,Integer>(); // associtate with each signal the number of constant indices it uses
		isigs = new TreeMap<String,Integer>();
		
		iinits = new TreeMap<String, Integer>();
		binits = new TreeMap<String, Boolean>();
		iainits = new TreeMap<String, Vector<Integer>>();
		bainits = new TreeMap<String, Vector<Boolean>>();
		//procs = new TreeMap<String, Integer>();
		procs = new Vector<String>();
		progs = new Vector<String>();
		
		dirtyVars = new TreeSet<String>();
		
		parReads = new Vector<String>();
		parWrites = new Vector<String>();
		seqReads = new Vector<String>();
		seqWrites = new Vector<String>();
	}

	//methods for registering name as the appropriate type, with or without an initialization
	void newI(String name) {checkName(name);is.add(name);Debug.trace(name + " added to is");}
	void newB(String name) {checkName(name);bs.add(name);Debug.trace(name + " added to bs");}
	void newIA(String name) {checkName(name);ias.add(name);Debug.trace(name + " added to ias");}
	void newBA(String name) {checkName(name);bas.add(name);Debug.trace(name + " added to bas");}
	void newI(String name, int init) {checkName(name);is.add(name);iinits.put(name,init);Debug.trace(name + " added to is");}
	void newB(String name, boolean init) {checkName(name);bs.add(name);binits.put(name,init); Debug.trace(name + " added to bs");}
	void newIA(String name, Vector<Integer> init) {checkName(name);ias.add(name);iainits.put(name, init); Debug.trace(name + " added to ias");}
	void newBA(String name, Vector<Boolean>init) {checkName(name);bas.add(name);bainits.put(name, init); Debug.trace(name + " added to bas");}

	void newLocalI(String name){checkName(name);localIs.add(name);}
	void newLocalB(String name){checkName(name);localBs.add(name);}
	void newLocalIA(String name){checkName(name);localIAs.add(name);}
	void newLocalBA(String name){checkName(name);localBAs.add(name);}

	//Methods for giving the content size and index size of arrays and variables
	void setContentSize(String name, int i) 
	{ 
		assert is.contains(name) : "setContentSize(" + name + ") called for variable not yet declared";
		assert ! contentSizes.containsKey(name) : "setContentSize(" + name + ") called for variable with size already declared";

		contentSizes.put(name,i);
	}	
	void setArrayContentSize(String name, int i) 
	{ 
		assert ias.contains(name) : "setArrayContentSize(" + name + ") called for variable not yet declared";
		assert ! arrayContentSizes.containsKey(name) : "setArrayContentSize(" + name + ") called for variable with size already declared";

		arrayContentSizes.put(name,i);
	}	
	void setArrayIndexSize(String name, int i) 
	{ 
		assert (ias.contains(name)||bas.contains(name)) : "setArrayIndexSize(" + name + ") called for variable not yet declared";
		assert ! arrayIndexSizes.containsKey(name) : "setArrayIndexSize(" + name + ") called for variable with size already declared";

		arrayIndexSizes.put(name,i);
	}

	void declareDirty(String name)
	{
		dirtyVars.add(name);
	}

	void newGlobConst(String name, int value){checkName(name); globalConsts.put(name, value); Debug.trace(name + " = " + value + " added to globalConsts");	}

	void newSig(String name,int arity) {checkName(name);sigs.put(name,arity);Debug.trace(name + " added to sigs");}
	void newISig(String name,int arity) {checkName(name);isigs.put(name,arity);Debug.trace(name + " added to isigs");}

	void newProcess(String name/*, int args*/) {checkName(name);procs.add(name);/*procs.put(name,args);*/}	
	void newProgram(String name){checkName(name);progs.add(name);}	

	void registerConstants(Collection<String> newConsts)  //replaces current list of integer constants with those passed in
	{
		deregisterConstsAndLocals(); //ensure this.consts is empty before checking for clashes
		for (Iterator<String> i = newConsts.iterator(); i.hasNext(); )
		{
			String s = i.next();
			checkName(s);
			consts.add(s);
		}
	}
	
	
	Vector<String> getLocalIVars(){return localIs;}
	Vector<String> getLocalBVars(){return localBs;}
	Vector<String> getLocalIAVars(){return localIAs;}
	Vector<String> getLocalBAVars(){return localBAs;}
	
	void addFinalLocalIVars(Vector<LocalVarUsage> vars){finalLocalIs = vars;}
	void addFinalLocalBVars(Vector<LocalVarUsage> vars){finalLocalBs = vars;}
	void addFinalLocalIAVars(Vector<LocalVarUsage> vars){finalLocalIAs = vars;}
	void addFinalLocalBAVars(Vector<LocalVarUsage> vars){finalLocalBAs = vars;}
	
	void deregisterConstsAndLocals() 
	{
		consts.clear();
		localIs.clear();
		localBs.clear();
		localIAs.clear();
		localBAs.clear();
		Debug.trace("Constants deregistered");
	}

	void checkName(String name)
	{
		//Don't expect this check to ever fail - should be caught as a parse error.  Hence use assert.
		//Actually I think it could fail through having two parameters with the same name, so don't use assert.
		if
		(
			(is.contains(name)) || (bs.contains(name)) || (ias.contains(name)) || (bas.contains(name))
			|| (consts.contains(name)) || (globalConsts.containsKey(name)) || (sigs.containsKey(name))
			|| (isigs.containsKey(name)) || (procs.contains/*Key*/(name)) || (progs.contains(name)) || (name.equals("assertionfailed"))
			|| (localIs.contains(name)) || (localBs.contains(name)) || (localIAs.contains(name)) || (localBAs.contains(name))
		)
			throw new ParserException( "Name " + name + " is not valid or is already being used");

		//checks to see if name is already registered in the context
		//this shouldnt really be needed since the lexer declares names it has already seen as the appropriate type
// 		Debug.trace("Checking if " + name + " has already been declared");
// 	
// 		if (is.contains(name)) {throw new ParserException(name + " declared but is already declared as an integer");}
// 		else if (bs.contains(name)) {throw new ParserException(name + " declared but is already declared as a boolean");}
// 		else if (ias.contains(name)) {throw new ParserException(name + " declared but is already declared as an integer array");}
// 		else if (bas.contains(name)) {throw new ParserException(name + " declared but is already declared as a boolean array");}
// 		else if (consts.contains(name)) {throw new ParserException(name + " declared but is already declared as a constant");}
// 		else if (globalConsts.containsKey(name)) {throw new ParserException(name + " declared but is already declared as a  global constant");}
// 		else if (sigs.contains(name)) {throw new ParserException(name + " declared but is already declared as a signal");}
// 		else if (isigs.contains(name)) {throw new ParserException(name + " declared but is already declared as an integer signal");}
// 		else if (procs.contains/*Key*/(name)) {throw new ParserException(name + " declared but is already declared as a process");}
// 		else if (progs.contains(name)) {throw new ParserException(name + " declared but is already declared as an program");}
// 		else if (name.equals("assertionfailed")) {throw new ParserException(name + " declared but assertionfailed is reserved for use by assert statements");}
		
	}

	//methods for retrieving the encoding of name. Throws an exception if name has not been declared for the appropriate type
	String getIEncoding(String name)
	{
		if (is.contains(name))
			return "IV." + (is.indexOf(name) + 1);
		else
			throw new ParserException(name + " has not been declared as an int");
	}
	
	String getBEncoding(String name)
	{
		if (bs.contains(name))
			return "BV." + (bs.indexOf(name) + 1);
		else
			throw new ParserException(name + " has not been declared as a boolean");
	}
	String getIAEncoding(String name)
	{
		if (ias.contains(name))
			return "IA." + (ias.indexOf(name) + 1);
		else
			throw new ParserException(name + " has not been declared as an int array");
	}
	String getBAEncoding(String name)
	{
		if (bas.contains(name))
			return "BA." + (bas.indexOf(name) + 1);
		else
			throw new ParserException(name + " has not been declared as an boolean array");
	}

	String getEncoding(String name)
	{
		if (isI(name))
			return getIEncoding(name);
		else if (isB(name))
			return getBEncoding(name);
		else if (isIA(name))
			return getIAEncoding(name);
		else if (isBA(name))
			return getBAEncoding(name);
		else throw new ParserException("getEncoding: " + name + " not recognised");
	}

	//int numArgs(String proc) {return procs.get(proc);}
	
	boolean isConstant(String name) {return (consts.contains(name) || globalConsts.containsKey(name));}
	boolean isLocalI(String name) {return localIs.contains(name);}
	boolean isLocalB(String name) {return localBs.contains(name);}
	boolean isLocalIA(String name) {return localIAs.contains(name);}
	boolean isLocalBA(String name) {return localBAs.contains(name);}
	boolean isI(String name) { return is.contains(name);}
	boolean isB(String name) { return bs.contains(name);}
	boolean isIA(String name) { return ias.contains(name);}
	boolean isBA(String name) { return bas.contains(name);}
	boolean isSig(String name) { return sigs.containsKey(name);}
	boolean isISig(String name) { return isigs.containsKey(name);}
	boolean isProcess(String name){return procs.contains/*Key*/(name);}
	boolean isProgram(String name) { return progs.contains(name);}

	boolean isDirty(String name) {return dirtyVars.contains(name);}

	//private void Debug.trace (String s)
	//{if (DEBUG) System.out.println("Context: " + s);}

	public String genCode(Context con){return genCode();}



	public String genCode()
	{
	
		String s = genSignals();
		//ensure there is at least 1 variable of each type
		s = s + "ivnums = {1.." + (java.lang.Math.max(1,is.size() + finalLocalIs.size())) + "}" + "\n";
		s = s + "bvnums = {1.." + (java.lang.Math.max(1,bs.size() + finalLocalBs.size())) + "}" + "\n";
		s = s + "ianums = {1.." + (ias.size() + finalLocalIAs.size()) + "}" + "\n";
		s = s + "banums = {1.." + (bas.size() + finalLocalBAs.size()) + "}" + "\n";

		s += "DirtyVars = Union({";
		for (String v : dirtyVars)
		{
			if ( isI(v) || isB(v))
			{	s += "{" + getEncoding(v) + "},";}
			else if (isIA(v) || isBA(v))
			{
				s += "{" + getEncoding(v) + ".x | x <- " 
					+ ( arrayIndexSizes.containsKey(v) ? "{0.." + (arrayIndexSizes.get(v)-1) + "}" : "ditype ") + "},"  ;
				
			}
			else throw new ParserException("Variable " + v + " is dirty but has not been declared");
		}
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s += "})\n";

		//s = s + "sigRenameAll(p) =  p" + renaming + " | i <-{MinI..MaxI}, j<-{0..nps-1} ]]\n";
// 		s = s + "sigRenameJ(p,j) =  p" + renaming + " | i <-{MinI..MaxI}]]\n";
// 		s += "atomicSigs(p) = p" + atomic + "| i <-{MinI..MaxI}]]\n";
// 		s += "unatomicSigs(p) = p" + unatomic + " | i <-{MinI..MaxI}, j<-{0..nps-1} ]]\n";
		
		s = s + genSizes();

		s = s + genInit();
		
		s += genLocalVarMapping();
// 		s = s + "niv=" + (( is.size() > 0 )? is.size() : 1) + "\n";
// 		s = s + "nbv=" +(( bs.size() > 0 )? bs.size() : 1) + "\n";
// 		s = s + "nia=" + ias.size() + "\n";
// 		s = s + "nba=" + bas.size() + "\n\n";


		
		for (Iterator<String> i = globalConsts.keySet().iterator(); i.hasNext();)
		{
			String name = i.next();
			s = s + name + " = " + globalConsts.get(name) + "\n";
		}

		s += genRefInfo();

		return s;
	}
	
	protected String genLocalVarMapping()
	{
		//FDR (sensibly) requires all definitions for the same function to occur together
		//Hence sort the output (lexicographically) using a TreeSet
	
		Debug.trace("Generating local variable mappings");
		Debug.trace("finalLocalIs.size = " + finalLocalIs.size());
		
		TreeSet<String> set = new TreeSet<String>();
		int i = is.size() + 1;
		for (LocalVarUsage lvu : finalLocalIs)
		{
			String def = "LocalIVarMapping_" + lvu.proc + "_" + lvu.var + "(";
			
			for (Integer j : lvu.args)
				def += j + ",";
				
			if (def.endsWith(",")) 
				def = def.substring(0, def.length() -1);
				
			def += ") = IV." + i + "\n";
			set.add(def);
			i++;
		}
		
		i = bs.size() + 1;
		for (LocalVarUsage lvu : finalLocalBs)
		{
			String def = "LocalBVarMapping_" + lvu.proc + "_" + lvu.var + "(";
			
			for (Integer j : lvu.args)
				def += j + ",";
				
			if (def.endsWith(",")) 
				def = def.substring(0, def.length() -1);
				
			def += ") = BV." + i + "\n";
			set.add(def);
			i++;
		}
		
		i = ias.size() + 1;
		for (LocalVarUsage lvu : finalLocalIAs)
		{
			String def = "LocalIAVarMapping_" + lvu.proc + "_" + lvu.var + "(";
			
			for (Integer j : lvu.args)
				def += j + ",";
				
			if (def.endsWith(",")) 
				def = def.substring(0, def.length() -1);
				
			def += ") = IA." + i + "\n";
			set.add(def);
			i++;
		}
		
		i = bas.size() + 1;
		for (LocalVarUsage lvu : finalLocalBAs)
		{
			String def = "LocalBAVarMapping_" + lvu.proc + "_" + lvu.var + "(";
			
			for (Integer j : lvu.args)
				def += j + ",";
				
			if (def.endsWith(",")) 
				def = def.substring(0, def.length() -1);
				
			def += ") = BA." + i + "\n";
			set.add(def);
			i++;
		}
		
		String result = "";
		for (String s : set)
			result += s;
		return result;
	}
			
	protected String genSizes()
	{
		String s = "ctypes = <";
		for (String name : contentSizes.keySet())
		{
			s += "(" + getIEncoding(name) + ", {0.." + (contentSizes.get(name) -1) + "}),";
		}
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s += ">\ncatypes = <";
		for (String name : arrayContentSizes.keySet())
		{
			if (ias.contains(name))
				s += "(" + getIAEncoding(name) + ", {0.." + (arrayContentSizes.get(name)-1) + "}),";
			else
				s += "(" + getBAEncoding(name) + ", {0.." + (arrayContentSizes.get(name)-1) + "}),";
		}
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s += ">\nitypes = <";
		for (String name : arrayIndexSizes.keySet())
		{
			if (ias.contains(name))
				s += "(" + getIAEncoding(name) + ", {0.." + (arrayIndexSizes.get(name)-1) + "}),";
			else
				s += "(" + getBAEncoding(name) + ", {0.." + (arrayIndexSizes.get(name)-1) + "}),";
		}
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s += ">\n";
		return s;
	}
	
	protected String genSignals()
	{
		String result = "channel assertionfailed\n";
		String signals = "Signals = {| assertionfailed";
		String isignals = "ISignals :: {Int=>Event}\nISignals = Union ({";
	
	
		for (Map.Entry<String,Integer> e: sigs.entrySet() )
		{
			result += "channel " + e.getKey();
			signals += ", " + e.getKey();
			
			if (e.getValue() > 0)
			{
				result += " : ";
				for (int i = 0; i < e.getValue(); i++)
				{
					if (i > 0)
						result += ".";
					result += "{MinI..MaxI}";
				}
			}
			result += "\n";
		}
	
		for (Map.Entry<String,Integer> e: isigs.entrySet() )
		{
			result += "channel " + e.getKey() + " : {MinI..MaxI}";
			isignals += "{" + e.getKey() ;
			
			String tail = "| ";
			
			for (int i = 0; i < e.getValue(); i++)
			{
				result += ".{MinI..MaxI}";
				isignals += ".x" + i;
				tail += " x" + i + " <- {MinI..MaxI},";
			}
			
			result += "\n";
			
			if (tail.endsWith(",")) 
			{
				tail = tail.substring(0, tail.length() -1);
				isignals += tail + "}," ;
			}
			else //tail = "| " still
			{
				assert tail.equals("| ");
				
				isignals += "},";
			}
			
			
		}

		signals += "|}\n";
		if (isignals.endsWith(",")) 
			isignals = isignals.substring(0, isignals.length() -1);
		isignals += "})";
		
		result += signals + "\n" + isignals + "\n";
		
		return result;

// 		String s = new String();
// 		
// 		
// 		//generate code for channels, but also for renaming them to channels tagged with the process number
// 		//name the tagged channels the same but with a t on the end.
// 		
// 		String cs = "assertionfailed"; //a comma seperated list of the signals
// 		//String tcs = "assertionfailedT";//signals tagged with process number
// 		//String acs = "at_assertionfailed";//signals tagged with at_
// 		//String atcs = "at_assertionfailedT";//signals tagged with at_ and process number
// 		//String renaming = "[[assertionfailed <- assertionfailedT.j, at_assertionfailed <- at_assertionfailedT.j";
// 		//String atomic = "[[ assertionfailed <- at_assertionfailed";
// 		//String unatomic = "[[ at_assertionfailedT.j <- assertionfailedT.j";
// 		
// 		{
// 		Iterator<String> i = sigs.iterator();
// 		while (i.hasNext())
// 		{
// 			String channel = i.next();
// 			cs = cs + ", " + channel;
// // 			tcs = tcs + ", " + channel + "T";
// // 			renaming = renaming + ", " + channel + " <- " + channel + "T.j, at_" + channel + " <- at_" + channel + "T.j";
// // 			acs +=  ", at_" + channel;
// // 			atcs +=  ", at_" + channel + "T";
// // 			atomic += ", " + channel + " <- at_" + channel;
// // 			unatomic += ", at_" + channel + "T.j <- " + channel + "T.j";
// 		
// 		}		
// // 		s = s + "channel " + cs + ", " + acs + "\nSignals = {" + cs + "}\nchannel " + tcs + ", " + atcs + ":{0..nps-1}\ntaggedSignals = {" + tcs + "}\n" ;		
// 		s = s + "channel " + cs + "\nSignals = {" + cs + "}\n" ;
// 		}
// 		if (isigs.size() > 0)
// 		{
// 			cs = new String(); //a comma seperated list of the signals
// // 			tcs = new String();
// // 			acs = new String();
// // 			atcs = new String();
// 			
// 			Iterator<String> i = isigs.iterator();
// 			cs = i.next();
// // 			tcs = cs + "T";
// // 			acs = "at_" + cs;
// // 			atcs = "at_" + cs + "T";
// // 			renaming = renaming + ", " + cs + ".i <- " + cs + "T.i.j, at_" + cs + ".i <- at_" + cs + "T.i.j";
// // 			atomic += ", " + cs + ".i <- at_" + cs + ".i";
// // 			unatomic += ", at_" + cs + "T.i.j <- " + cs + "T.i.j";
// 			
// 			while (i.hasNext())
// 			{
// 				String channel = i.next();
// 				cs = cs + ", " + channel;
// // 				tcs = tcs + ", " + channel + "T";
// // 				renaming = renaming + ", " + channel + ".i <- " + channel + "T.i.j, at_" + channel + ".i <- at_" + channel + "T.i.j";
// // 				acs +=  ", at_" + channel;
// // 				atcs +=  ", at_" + channel + "T";
// // 				atomic += ", " + channel + ".i <- at_" + channel + ".i";
// // 				unatomic += ", at_" + channel + "T.i.j <- " + channel + "T.i.j";
// 			}
// // 			s = s + "channel " + cs + ", " + acs + ":{MinI..MaxI}\nISignals = {" + cs + "}\nchannel " + tcs + ", " + atcs + ":{MinI..MaxI}.{0..nps-1}\ntaggedISignals = {" + tcs + "}\n";
// 			s = s + "channel " + cs + ":{MinI..MaxI}\nISignals = {" + cs + "}\n";
// 		}
// 		else s = s + "ISignals = {}\n"/*taggedISignals={}\n"*/;
// 		
// 		return s;
	}

	private String genInit()
	{
		String s = "init = (<";
		
		Iterator<String> i = iinits.keySet().iterator();
		while (i.hasNext())
		{
			String name = i.next();
			//if (iinits.containsKey(name))
			s = s + "(" + getIEncoding(name) + "," + iinits.get(name) + "),";
		}
		i = iainits.keySet().iterator(); 
		while (i.hasNext())
		{
			String name = i.next();
			Vector<Integer> inits = iainits.get(name);
			
			int j = 0;
			while (j< inits.size())
			{
				s = s + "(" + getIAEncoding(name) + "." + j + "," + inits.elementAt(j) + "),";
				j++;
			}
		}
		
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s = s + ">,<";
		
		i = bs.iterator();
		while (i.hasNext())
		{
			String name = i.next();
			if (binits.containsKey(name))
				s = s + "(" + getBEncoding(name) + "," + binits.get(name) + "),";
		}
		i = bainits.keySet().iterator(); 
		while (i.hasNext())
		{
			String name = i.next();
			Vector<Boolean> inits = bainits.get(name);
			
			int j = 0;
			while (j< inits.size())
			{
				
				s = s + "(" + getBAEncoding(name) + "." + j + "," + inits.elementAt(j) + "),";
				j++;
			}
		}

		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s = s + ">)\n";
		return s;


		
	}
	
	public String getIName(int i) 
	{
		if (i <= is.size())
			return is.elementAt(i-1);
		else if (i - is.size() <= finalLocalIs.size())
		{
			return finalLocalIs.elementAt(i - is.size() - 1).var;
		}
		else
			throw new InterpreterException("Index for IVar out of range");
	}
	public String getBName(int i)
	{
		if (i <= bs.size())
			return bs.elementAt(i-1);
		else if (i - bs.size() <= finalLocalBs.size())
		{
			return finalLocalBs.elementAt(i - bs.size() - 1).var;
		}
		else
			throw new InterpreterException("Index for BVar out of range");
	}
	public String getIAName(int i) 
	{
		if (i <= ias.size())
			return ias.elementAt(i-1);
		else if (i - ias.size() <= finalLocalIAs.size())
		{
			return finalLocalIAs.elementAt(i - ias.size() - 1).var;
		}
		else
			throw new InterpreterException("Index " + i + " for IAVar out of range");
	}
	public String getBAName(int i)
	{
		if (i <= bas.size())
			return bas.elementAt(i-1);
		else if (i - bas.size() <= finalLocalBAs.size())
		{
			return finalLocalBAs.elementAt(i - bas.size() - 1).var;
		}
		else
			throw new InterpreterException("Index for BAVar out of range");
	}

	public Map<String, Integer> getConsts() {return globalConsts;}
	
	
	public void setParReads(Vector<String> v)
	{
		if (v == null)
			throw new ParserException("setParReads given null v");
		if (parReads.isEmpty())
			parReads = v;
		else
			throw new ParserException("ParReads declared twice");
	}
	public void setParWrites(Vector<String> v)
	{
		if (v == null)
			throw new ParserException("setParWrites given null v");
		if (parWrites.isEmpty())
			parWrites = v;
		else
			throw new ParserException("ParWrites declared twice");
	}
	public void setSeqReads(Vector<String> v)
	{
		if (v == null)
			throw new ParserException("setSeqReads given null v");
		if (seqReads.isEmpty())
			seqReads = v;
		else
			throw new ParserException("SeqReads declared twice");
	}
	public void setSeqWrites(Vector<String> v)
	{
		if (v == null)
			throw new ParserException("setSeqWrites given null v");
		if (seqWrites.isEmpty())
			seqWrites = v;
		else
			throw new ParserException("SeqWrites declared twice");
	}

	protected String genRefInfo()
	{
		 return "ParReads = " + genRefInfo(parReads) + "ParWrites = " + genRefInfo(parWrites) + "SeqReads = " + genRefInfo(seqReads) + "SeqWrites = " + genRefInfo(seqWrites);
		
	}
	
	//used by genRefInfo()
	private String genRefInfo(Collection<String> v)
	{
		String result = "Union({";
		for (String s : v)
		{	
			if (s.contains("["))
			{
				String name = s.substring(0,s.indexOf("["));
				String index = s.substring(s.indexOf("[") + 1, s.indexOf("]"));
				
				result += "{" + getEncoding(name) + "." + index + "},";
			}
			else if (isIA(s) || isBA(s))
			{
				result += "{" +  getEncoding(s) + ".v | v <- itype("+ getEncoding(s) + ")},";
			}
			else
				result +=  "{" + getEncoding(s) + "},";
		
		}
		if (result.endsWith(",")) result = result.substring(0, result.length() -1);
		return result + "})\n";
	}

}

