import java.util.Map;
import java.util.Vector;

public interface IConst extends Genable 
{
	public int constVal(Map<String, Integer> consts);//returns the value of the IConst
	
	public String printWithVals(Map<String, Integer> vals);

	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals);
	
}

abstract class BinIConst implements IConst
{
	String op;
	IConst i1,i2;
	BinIConst(String op, IConst i1, IConst i2)
	{
		if (op==null) throw new ParserException("op Argument to BinIConst null");
		if (i1==null) throw new ParserException("i1 Argument to BinIConst (" + op +") null");
		if (i2==null) throw new ParserException("i2 Argument to BinIConst (" + op +")  null");
		this.op=op;
		this.i1=i1;
		this.i2=i2;
		
	}
	
	public String genCode(Context con) { return toString(); }

	public String toString()
	{ return "(" + i1 + op + i2 + ")";}
	
	public String printWithVals(Map<String, Integer> vals)
	{
		return i1.printWithVals(vals) + op + i2.printWithVals(vals);
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BinIConst) ) return false;
		
		BinIConst that = (BinIConst)aThat;

		return ( this.op.equals(that.op)
			&& this.i1.equals(that.i1)
			&& this.i2.equals(that.i2));
	}
	@Override public int hashCode() { 
		return 1 + op.hashCode()
			+ i1.hashCode()
			+ i2.hashCode();
	}
}

class AddConst extends BinIConst
{
	public AddConst(IConst i1, IConst i2){super("+",i1,i2);}
	public int constVal(Map<String, Integer> consts){return i1.constVal(consts) + i2.constVal(consts);}
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new AddConst(i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));}
	
}
class SubConst extends BinIConst
{
	public SubConst(IConst i1, IConst i2){super("-",i1,i2);}
	public int constVal(Map<String, Integer> consts){return i1.constVal(consts) - i2.constVal(consts);}
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new SubConst(i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));}
}
class MultConst extends BinIConst
{
	public MultConst(IConst i1, IConst i2){super("*",i1,i2);}
	public int constVal(Map<String, Integer> consts){return i1.constVal(consts) * i2.constVal(consts);}
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new MultConst(i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));}
}
class DivConst extends BinIConst
{
	public DivConst(IConst i1, IConst i2){super("/",i1,i2);}
	public int constVal(Map<String, Integer> consts){return i1.constVal(consts) / i2.constVal(consts);}
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new DivConst(i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));}
}
class ModConst extends BinIConst
{
	public ModConst(IConst i1, IConst i2){super("%",i1,i2);}
	public int constVal(Map<String, Integer> consts){return i1.constVal(consts) % i2.constVal(consts);}
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new ModConst(i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));}
}


class NegIConst implements IConst
{
	IConst i;
	NegIConst(IConst i)
	{
		if (i==null) throw new ParserException("i Argument to NegIConst null");
		this.i=i;
	}
	
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return new NegIConst(i.substituteConst(vars,vals));}
	public String genCode(Context con) { return toString(); }

	public String toString()
	{ return "(-(" + i + "))";}
	
	public int constVal(Map<String, Integer> consts){return -i.constVal(consts);}
	public String printWithVals(Map<String, Integer> vals){return "-(" + i.printWithVals(vals) + ")";} 
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof NegIConst) ) return false;
		
		NegIConst that = (NegIConst)aThat;

		return ( this.i.equals(that.i));
	}
	@Override public int hashCode() { 
		return 1 + i.hashCode();
	}
}

class NumIConst implements IConst
{
	int i;
	NumIConst(int i)
	{
		this.i=i;
	}
	
	public String genCode(Context con) { return toString(); }

	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ return this;}
	public String toString()
	{ return String.valueOf(i);}
	
	public int constVal(Map<String, Integer> consts){return i;}
	public String printWithVals(Map<String, Integer> vals){return String.valueOf(i);} 
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof NumIConst) ) return false;
		
		NumIConst that = (NumIConst)aThat;

		return ( this.i == that.i );
	}
	@Override public int hashCode() { 
		return 1 + i;
	}
}

class VarIConst implements IConst
{
	String name;
	VarIConst(String name)
	{
		if (name == null) throw new ParserException("name argument to VarIConst null");
		this.name=name;
	}
	
	public IConst substituteConst(Vector<String> vars, Vector<IConst> vals)
	{ 
		int index = vars.indexOf(name);
		if (index < 0)
			return this;
		else
			return vals.get(index);
	}
	public String genCode(Context con) { return toString(); }

	public String toString()
	{ return name;}
	
	
	public int constVal(Map<String, Integer> consts)
	{
		if (consts.containsKey(name)) return consts.get(name);
		else throw new InterpreterException("VarIConst.constVal(consts) - " + name + " not in consts");
	
	}
	
	public String printWithVals(Map<String, Integer> vals)
	{	
		if (vals.containsKey(name)) return name + "(" + vals.get(name) + ")";
		else throw new InterpreterException("VarIConst.printWithVal(consts) - " + name + " not in consts");
	} 
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof VarIConst) ) return false;
		
		VarIConst that = (VarIConst)aThat;

		return ( this.name.equals(that.name));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode();
	}
}