import java.util.Map;
import java.util.Vector;

public interface IExpr extends Genable 
{ 
	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals);
}

/*abstract*/ class BinIExpr implements IExpr
{
	String op, print;
	IExpr i1,i2;
	BinIExpr(String op, String print, IExpr i1, IExpr i2)
	{
		if (op==null) throw new ParserException("op Argument to BinIExp null");
		if (print==null) throw new ParserException("print Argument to BinIExpr(" + op +") null");
		if (i1==null) throw new ParserException("i1 Argument to BinIExp (" + op +") null");
		if (i2==null) throw new ParserException("i2 Argument to BinIExp (" + op +")  null");
		this.op=op;
		this.print = print;
		this.i1=i1;
		this.i2=i2;
		
	}
	
	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new BinIExpr(op,print,i1.substituteConst(vars,vals),i2.substituteConst(vars,vals));
	}
	public String genCode(Context con) {return "BIOp." + op + "." + i1.genCode(con) + "." + i2.genCode(con); }

	
	public String toString()
	{ return i1 + print + i2;}


	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BinIExpr) ) return false;
		
		BinIExpr that = (BinIExpr)aThat;

		return ( this.op.equals(that.op)
			&& this.print.equals(that.print)
			&& this.i1.equals(that.i1)
			&& this.i2.equals(that.i2));
	}
	@Override public int hashCode() { 
		return 1 + op.hashCode()
			+ print.hashCode()
			+ i1.hashCode()
			+ i2.hashCode();
	}
}

class Plus extends BinIExpr
{
	public Plus(IExpr i1, IExpr i2){super("Plus","+",i1,i2);}
}
class Minus extends BinIExpr
{
	public Minus(IExpr i1, IExpr i2){super("Minus","-",i1,i2);}
}
class Times extends BinIExpr
{
	public Times(IExpr i1, IExpr i2){super("Times","*",i1,i2);}
}
class Div extends BinIExpr
{
	public Div(IExpr i1, IExpr i2){super("Div","/",i1,i2);}
}
class Mod extends BinIExpr
{
	public Mod(IExpr i1, IExpr i2){super("Mod","%",i1,i2);}
}
class Max extends BinIExpr
{
	public Max(IExpr i1, IExpr i2){super("Max","(max)",i1,i2);}
}
class Min extends BinIExpr
{
	public Min(IExpr i1, IExpr i2){super("Min","(min)",i1,i2);}
}

class Neg implements IExpr
{
	IExpr i;
	public Neg(IExpr i)
	{if (i==null) throw new ParserException("Argument to Neg null");this.i =i;}
	
	public String genCode(Context con) { return "UIOp.Uminus." + i.genCode(con);}

	
	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new Neg(i.substituteConst(vars,vals));
	}
	
	public String toString()
	{ return "-(" + i + ")";}

	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Neg) ) return false;
		
		Neg that = (Neg)aThat;

		return ( this.i.equals(that.i));
	}
	@Override public int hashCode() { 
		return 1 + i.hashCode();
	}
}

class IVar implements IExpr
{
	String name;
	public IVar(String name){if (name==null) throw new ParserException("Argument to IVar null");this.name = name;}
	
	public String genCode(Context con) 
	{ 
		/*if (con.isConstant(name) 	//was possible for this to be a constant.  No longer the case
			return name;
		else*/
			return "IVar." + con.getIEncoding(name);
	}

	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
	public String toString() {return name;}

	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof IVar) ) return false;
		
		IVar that = (IVar)aThat;

		return ( this.name.equals(that.name));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode();
	}

}

class IArray implements IExpr
{
	String name;
	IExpr i;
	public IArray(String name, IExpr i)
	{
		if (name==null) throw new ParserException("name argument to IArray null"); 
		if (i==null) throw new ParserException("index argument to IArray null"); 
		
		this.name = name; this.i = i;
	}

	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new IArray(name,i.substituteConst(vars,vals));
	}
	public String genCode(Context con) 
	{ 
		return "IArc.(" + con.getIAEncoding(name) + "," + i.genCode(con) + ")";
	}

	public String toString() {return name + "[" + i + "]";}

	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof IArray) ) return false;
		
		IArray that = (IArray)aThat;

		return ( this.name.equals(that.name)
			&& this.i.equals(that.i));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ i.hashCode();
	}
	
}

class IConstExpr implements IExpr
{
	IConst i;
	public IConstExpr(IConst i){this.i = i;}
	
	public String genCode(Context con) {return "Const." + i.genCode(con) ;}

	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new IConstExpr(i.substituteConst(vars, vals));
	}
	public String toString() {return i.toString();}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof IConstExpr) ) return false;
		
		IConstExpr that = (IConstExpr)aThat;

		return ( this.i.equals(that.i));
	}
	@Override public int hashCode() { 
		return 1 + i.hashCode();
	}
}


class LocalIVar implements IExpr
{
	String name,process;
	public LocalIVar(String name,String process)
	{
		if (name==null) throw new ParserException("Argument to LocalIVar null");
		if (process==null) throw new ParserException("Argument to LocalIVar null");
		this.name = name;
		this.process = process;
		
	}
	
	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return this;
	}
	public String genCode(Context con) 
	{ 
	
		String result = "IVar.LocalIVarMapping_" + process + "_" + name + "(";
		for (String s : con.consts)
			result += s + ",";
			
		if (result.endsWith(",")) 
			result = result.substring(0, result.length() -1);
			
		return result + ")";
	}

	public String toString() {return name;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof LocalIVar) ) return false;
		
		LocalIVar that = (LocalIVar)aThat;

		return ( this.name.equals(that.name)
			&& this.process.equals(that.process));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ process.hashCode();
	}
}

class LocalIArray implements IExpr
{
	String name,process;
	IExpr i;
	public LocalIArray(String name, String process, IExpr i)
	{
		if (name==null) throw new ParserException("name argument to LocalIArray null"); 
		if (process==null) throw new ParserException("process argument to LocalIArray null"); 
		if (i==null) throw new ParserException("index argument to LocalIArray null"); 
		
		this.name = name; this.i = i; this.process = process;
	}

	public IExpr substituteConst(Vector<String> vars, Vector<IConst> vals)
	{
		return new LocalIArray(name,process,i.substituteConst(vars,vals));
	}
	public String genCode(Context con) 
	{ 
		String genName = "LocalIAVarMapping_" + process + "_" + name + "(";
		for (String s : con.consts)
			genName += s + ",";
			
		if (genName.endsWith(",")) 
			genName = genName.substring(0, genName.length() -1);

	
	
		return "IArc.(" + genName + ")," + i.genCode(con) + ")";
	}

	public String toString() {return name + "[" + i + "]";}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof LocalIArray) ) return false;
		
		LocalIArray that = (LocalIArray)aThat;

		return ( this.name.equals(that.name)
			&& this.i.equals(that.i)
			&& this.process.equals(that.process));
	}
	@Override public int hashCode() { 
		return 1 + name.hashCode()
			+ i.hashCode()
			+ process.hashCode();
	}
}
