import java.util.*;


class Access
{
	public String array;
	public IConst index;
	public Access(String array, IConst index){this.array = array; this.index = index;}

}

class VariableAssignment
{
	public Vector<String> is = new Vector<String>();
	public Vector<String> bs = new Vector<String>();
	public Vector<String> ias = new Vector<String>();
	public Vector<String> bas = new Vector<String>();
	public Vector<Access> iaAccesses = new Vector<Access>();
	public Vector<Access> baAccesses = new Vector<Access>();

	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof VariableAssignment) ) return false;
		
		VariableAssignment that = (VariableAssignment)aThat;

		return 
			(this.is.equals(that.is) && 
			this.bs.equals(that.bs) && 
			this.ias.equals(that.ias) && 
			this.bas.equals(that.bas) && 
			this.iaAccesses.equals(that.iaAccesses) && 
			this.baAccesses.equals(that.baAccesses));
	}
	@Override public int hashCode() { 
		return 1 + 
			is.hashCode() 
			+ bs.hashCode() 
			+ ias.hashCode() 
			+ bas.hashCode() 
			+ iaAccesses.hashCode() 
			+ baAccesses.hashCode();
	}

}


public class Process implements Genable
{
	//final boolean DEBUG = false;
	//void trace(String s)
	//{ if (DEBUG) System.out.println("Process: " + s);}
	
	//represents the process name(consts) = prog
	
	String name;
	
	Vector<String> consts;
	private Vector<String> localIVars;
	private Vector<String> localBVars;
	private Vector<String> localIAVars;
	private Vector<String> localBAVars;
	private VariableAssignment va;

	Cmd prog;


	public Vector<String> getLocalIVars(){return localIVars;}
	public Vector<String> getLocalBVars(){return localBVars;}
	public Vector<String> getLocalIAVars(){return localIAVars;}
	public Vector<String> getLocalBAVars(){return localBAVars;}
	
	
	
	public Process(String name, Vector<String> consts, Cmd prog, VariableAssignment va) {this.name = name; this.consts = consts; this.prog=prog; this.va = va;}
	
	public String genCode(Context c) 
	{ 
		Debug.trace("Process " + name + ": Generating Code");
		String code = name + "(";
		Enumeration<String> e = consts.elements();
		if (e.hasMoreElements())
			code = code + e.nextElement();
		for (; e.hasMoreElements();code = code + "," + e.nextElement() ){}
		Debug.trace("Process " + name + ": Registering Constants");
		c.registerConstants(consts);
		Debug.trace("Process " + name + ": Generating program code");
		code = code + ") = (" + prog.genCode(c) + "\n, (Union({{";
		
		
		Debug.trace("Process " + name + ": Generating local IVar assignments");
		for (String var : localIVars)
		{
			
			code += "IVar.LocalIVarMapping_" + name + "_" + var + "(";
			for (String s : consts)
				code += s + ",";
				
			if (code.endsWith(",")) 
				code = code.substring(0, code.length() -1);
				
			code += "),";
		}
		
		Debug.trace("Process " + name + ": Generating assigned IVar assignments");
		for (String var : va.is)
			code += "IVar." + c.getIEncoding(var) + ",";
		for (Access acc : va.iaAccesses)
			code += "IVar." + c.getIAEncoding(acc.array) + "." + acc.index.genCode(c) + ",";
		
		
		
		if (code.endsWith(",")) 
			code = code.substring(0, code.length() -1);
		code += "},";
		
		Debug.trace("Process " + name + ": Generating local IAVar assignments");		
		for (String var : localIAVars)
		{
			
			String encoding = "LocalIAVarMapping_" + name + "_" + var + "(";
			for (String s : consts)
				encoding += s + ",";
				
			if (encoding.endsWith(",")) 
				encoding = encoding.substring(0, encoding.length() -1);
				
			encoding += ")";
			
			code += "{IVar." + encoding  + ".v | v <- itype("+ encoding + ")},";
		}
		
		Debug.trace("Process " + name + ": Generating assigned IAVar assignments");
		for (String array : va.ias)
			code += "{IVar." + c.getIAEncoding(array) + ".v | v <- itype("+ c.getIAEncoding(array) + ")},";
		
		
		if (code.endsWith(",")) 
			code = code.substring(0, code.length() -1);
		
		code += "}),Union({{";
		
		Debug.trace("Process " + name + ": Generating local BVar assignments");
		for (String var : localBVars)
		{
			
			code += "BVar.LocalBVarMapping_" + name + "_" + var + "(";
			for (String s : consts)
				code += s + ",";
				
			if (code.endsWith(",")) 
				code = code.substring(0, code.length() -1);
				
			code += "),";
		}
		
		Debug.trace("Process " + name + ": Generating assigned BVar assignments");
		for (String var : va.bs)
			code += "BVar." + c.getBEncoding(var) + ",";
		for (Access acc : va.baAccesses)
			code += "BVar." + c.getBAEncoding(acc.array) + "." + acc.index.genCode(c) + ",";
		
		if (code.endsWith(",")) 
			code = code.substring(0, code.length() -1);
		code += "},";
		
		Debug.trace("Process " + name + ": Generating local BAVar assignments");		
		for (String var : localBAVars)
		{
			
			String encoding = "LocalBAVarMapping_" + name + "_" + var + "(";
			for (String s : consts)
				encoding += s + ",";
				
			if (encoding.endsWith(",")) 
				encoding = encoding.substring(0, encoding.length() -1);
				
			encoding += ")";
			
			code += "{BVar." + encoding  + ".v | v <- itype("+ encoding + ")},";
		}
		
		Debug.trace("Process " + name + ": Generating assigned BAVar assignments");
		for (String array : va.bas)
			code += "{BVar." + c.getBAEncoding(array) + ".v | v <- itype("+ c.getBAEncoding(array) + ")},";
		
		
		if (code.endsWith(",")) 
			code = code.substring(0, code.length() -1);
		
		code += "})))";
		
		Debug.trace("Process " + name + ": Returning code");
		return code;
	}
	
	public void addLocalIVars(Vector<String> vars)
	{
		localIVars = (Vector<String>) vars.clone();
	}
	public void addLocalBVars(Vector<String> vars)
	{
		localBVars = (Vector<String>) vars.clone();
	}
	public void addLocalIAVars(Vector<String> vars)
	{
		localIAVars = (Vector<String>) vars.clone();
	}
	public void addLocalBAVars(Vector<String> vars)
	{
		localBAVars = (Vector<String>) vars.clone();
	}
	
	public int numArgs() {return consts.size();}
	
	public boolean usesSig(String sig,  Map<String, Integer> consts) {return prog.usesSig(sig,consts);}
// 	public int numIEs(Map<String,Integer> consts) {return prog.numIEs(consts);}	//these functions return the number of integer and boolean expressions within this Process.  This is the number of seperate expression processes the compiled Expr will generate
// 	public int numBEs(Map<String,Integer> consts) {return prog.numBEs(consts);}
// 
// 	public String getIExpr(int i, Map<String, Integer> ivals){return prog.getIExpr(i,ivals);} //returns a string representing the ith integer expression in this
// 	public String getBExpr(int i, Map<String, Integer> ivals,Map<String,Boolean> bvals){return prog.getBExpr(i,ivals,bvals);}
// 	
	public String toString()
	{
		String s = name + "(";
		Enumeration<String> e = consts.elements();
		if (e.hasMoreElements())
			s = s + e.nextElement();
		for (; e.hasMoreElements();s = s + "," + e.nextElement() ){}
		return s + ")";
	
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Process) ) return false;
		
		Process that = (Process)aThat;

		return 
			(this.name.equals(that.name) && 
			this.consts.equals(that.consts) && 
// 			this.localIVars.equals(that.localIVars) && 
// 			this.localBVars.equals(that.localBVars) && 
// 			this.localIAVars.equals(that.localIAVars) && 
// 			this.localBAVars.equals(that.localBAVars) && 
			this.va.equals(that.va) && 
			this.prog.equals(that.prog));
	}
	@Override public int hashCode() { 
		return 1 + 
			name.hashCode() 
			+ consts.hashCode() 
// 			+ localIVars.hashCode() 
// 			+ localBVars.hashCode() 
// 			+ localIAVars.hashCode() 
// 			+ localBAVars.hashCode() 
			+ va.hashCode() 
			+ prog.hashCode();
	}
}