%{
  import java.io.*;
  import java.util.*;


%}
      
%token CSP
%token NUM  /* a number */
%token NAME /* a name for a process or signal or variable/constant decleration*/
%token TRUE FALSE
%left ';'
%token SKIP ITER WHILE DO IF THEN ASSIGN SIG ISIG ATOMIC MIN MAX NOOP GLOBCONST FROM TO
%token PROCESS PROGRAM SIGNAME ISIGNAME IVAR BVAR IAVAR BAVAR CONST INT BOOL WITH WATCHING LEAFCOMPRESSED HIERARCHCOMPRESSED DIRTY LOCALIVAR LOCALBVAR LOCALIAVAR LOCALBAVAR
%token ASSERT IN ALWAYS NEVER OUTOFRANGE NOSIGNAL ERROR VERROR CSPASSERT PARREADS PARWRITES SEQREADS SEQWRITES GENREFINEMENT SEQREFINEMENT PARREFINEMENT OVERSEER
%nonassoc IFX
%nonassoc ELSE

%left OR
%left AND
%nonassoc '<' '>' GE LE '=' NEQ
%left '!'

%left '-' '+'
%left '*' '/' '%'
%left NEG        /* negation--unary minus */
%%

Input:   /* empty string */ 	{Debug.trace("Parser: Input - empty string matched");}
       | Stmt Input		{Debug.trace("Parser: Input - Stmt Input matched");}
       ;

Stmt: 	  Process 	{Debug.trace("Parser: Stmt - Process matched\n");}
	| Declaration MaybeSemiColon {Debug.trace("Parser: Stmt - Declaration matched\n");}
	| Program {Debug.trace("Parser: Stmt - Program matched\n");}
	| CSPStmt	{Debug.trace("Parser: Stmt - CSPStmt Matched\n");}
	| Assertion	{Debug.trace("Parser: Stmt - Assertion matched\n");}
	| RefDeclaration MaybeSemiColon {Debug.trace("Parser: Stmt - RefDeclaration matched\n");}
	;

MaybeSemiColon:
	';'
	|
;

CSPStmt:	CSP	{Debug.trace("Parser: CSPStmt - CSP(" + $1.sval + ") matched"); compiler.csp.add($1.sval);};

Declaration: 
	DirtyDeclaration INT SizeDeclaration INameList			
	{
		Debug.trace("Parser: Declaration - INT INameList matched");
		if ($1.ival > 0)
			declareDirty( (Vector<String>) $4.obj);
		if ($3.ival >= 0)
			declareContentSize( (Vector<String>) $4.obj, $3.ival);
		//else < 0 means no size declared
	}
	| DirtyDeclaration BOOL BNameList		
	{
		Debug.trace("Parser: Declaration - BOOL BNameList matched");
		if ($1.ival > 0)
			declareDirty( (Vector<String>) $3.obj);
	}
	| DirtyDeclaration INT SizeDeclaration '[' ArraySizeDeclaration ']' IANameList	
	{
		Debug.trace("Parser: Declaration - INT SizeDeclaration [ArraySizeDeclaration] IANameList matched");
		if ($1.ival > 0)
			declareDirty( (Vector<String>) $7.obj);
		if ($3.ival >= 0)
			declareArrayContentSize( (Vector<String>) $7.obj, $3.ival);
		if ($5.ival >= 0)
			declareArrayIndexSize( (Vector<String>) $7.obj, $5.ival);
	}
	| DirtyDeclaration BOOL '[' ArraySizeDeclaration ']' BANameList	
	{
		Debug.trace("Parser: Declaration - BOOL[ArraySizeDeclaration] BANameList matched");
		if ($1.ival > 0)
			declareDirty( (Vector<String>) $6.obj);
		if ($4.ival >= 0)
			declareArrayIndexSize( (Vector<String>) $6.obj, $4.ival);
	} 
	| SIG SigNameList		{Debug.trace("Parser: Declaration - SIG[] SigNameList matched");}
	| ISIG ISigNameList		{Debug.trace("Parser: Declaration - ISIG[] ISigNameList matched");}
	| GLOBCONST GlobConstList 	{Debug.trace("Parser: Declaration - GLOBCONST GlobConstList matched");}
	;

SizeDeclaration:
				{Debug.trace("SizeDeclaration - empty matched"); $$.ival = -1;}
	| '%' NUM 	{Debug.trace("SizeDeclaration - % NUM matched"); $$.ival = $2.ival;}

;

ArraySizeDeclaration:
				{Debug.trace("ArraySizeDeclaration - empty matched"); $$.ival = -1;}
	| NUM 	{Debug.trace("ArraySizeDeclaration - NUM matched"); $$.ival = $1.ival;}

;

DirtyDeclaration:
			{Debug.trace("DirtyDeclaration - empty matched"); $$.ival = -1;}
	| DIRTY 	{Debug.trace("DirtyDeclaration - DIRTY matched"); $$.ival = 1;}

;

INameList:	IName					{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: INameList - IName matched");}
		| IName ',' INameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: INameList - IName , INameList matched");}
		;

IName:		NAME	{compiler.context.newI($1.sval);$$.sval = $1.sval;Debug.trace("Parser: IName - NAME (" + $1.sval + ") matched");}
		| NAME '=' NUM {compiler.context.newI($1.sval, $3.ival);$$.sval = $1.sval;Debug.trace("Parser: IName - NAME (" + $1.sval + ") = NUM ("+ $3.ival + ") matched");}
		;
		

BNameList:	BName			{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: BNameList - BName matched");}
		| BName ',' BNameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: BNameList - BName), BNameList matched");}
		;

BName:		NAME			{compiler.context.newB($1.sval);$$.sval = $1.sval;Debug.trace("Parser: BName - NAME (" + $1.sval + ") matched");}
		|NAME '=' TRUE	{compiler.context.newB($1.sval, true);$$.sval = $1.sval;Debug.trace("Parser: BNameList - NAME (" + $1.sval + ") = TRUE matched");}
		|NAME '=' FALSE	{compiler.context.newB($1.sval, false);$$.sval = $1.sval;Debug.trace("Parser: BNameList - NAME (" + $1.sval + ") = FALSE matched");}
		;		
		
		

IANameList:	IAName			{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: IANameList - IAName matched");}
		| IAName ',' IANameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: IANameList - IAName , IANameList matched");}
		;
		
IAName:		NAME	{compiler.context.newIA($1.sval);$$.sval = $1.sval;Debug.trace("Parser: IAName - NAME (" + $1.sval + ") matched");}
		| NAME '=' '<' NonEmptyNumList '>' {compiler.context.newIA($1.sval, (Vector<Integer>)$4.obj);$$.sval = $1.sval;Debug.trace("Parser: IAName - NAME (" + $1.sval + ") = < NonEmptyNumList > matched");}
		;


BANameList:	BAName			{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v; Debug.trace("Parser: BANameList - BAName matched");}
		| BAName ',' BANameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v; Debug.trace("Parser: BANameList - BAName (" + $1.sval + "), BANameList matched");}
		;

GlobConstList:	NAME '=' NUM					{compiler.context.newGlobConst($1.sval, $3.ival); Debug.trace("Parser: GlobConstList - NAME(" + $1.sval + " = " + "NUM(" + $3.ival + ") - matched");}
		|NAME '=' NUM ',' GlobConstList {compiler.context.newGlobConst($1.sval, $3.ival); Debug.trace("Parser: GlobConstList - NAME(" + $1.sval + " = " + "NUM(" + $3.ival + ") , GlobConstList - matched");}
		;
	
BAName:		NAME	{compiler.context.newBA($1.sval);$$.sval = $1.sval;Debug.trace("Parser: BAName - NAME (" + $1.sval + ") matched");}
		|NAME '=' '<' NonEmptyBoolList '>' {compiler.context.newBA($1.sval, (Vector<Boolean>)$4.obj);$$.sval = $1.sval; Debug.trace("Parser: BAName - NAME(" + $1.sval + ") = < NonEmptyBoolList >");}
		;
		
NonEmptyBoolList:	TRUE	{Vector<Boolean> vector = new Vector<Boolean>(); vector.add(true); $$.obj = vector; Debug.trace("Parser: NonEmptyBoolList - TRUE matched");}
			| NonEmptyBoolList ',' TRUE	{((Vector<Boolean>)$1.obj).add(true); $$.obj = $1.obj; Debug.trace("Parser: NonEmptyBoolList - NonEmptyBoolList, TRUE matched");}
			| FALSE	{Vector<Boolean> vector = new Vector<Boolean>(); vector.add(false); $$.obj = vector; Debug.trace("Parser: NonEmptyBoolList - FALSE matched");}
			| NonEmptyBoolList ',' FALSE	{((Vector<Boolean>)$1.obj).add(false); $$.obj = $1.obj; Debug.trace("Parser: NonEmptyBoolList - NonEmptyBoolList, FALSE matched");}
			;	

SigNameList:	NAME SigType			{compiler.context.newSig($1.sval,$2.ival);Debug.trace("Parser: SigNameList - NAME (" + $1.sval + ") matched");}
		| NAME SigType ',' SigNameList	{compiler.context.newSig($1.sval,$2.ival);Debug.trace("Parser: SigNameList - NAME (" + $1.sval + "), SigNameList matched");}
		;

ISigNameList:	NAME SigType			{compiler.context.newISig($1.sval,$2.ival);Debug.trace("Parser: ISigNameList - NAME (" + $1.sval + ") matched");}
		| NAME SigType',' ISigNameList	{compiler.context.newISig($1.sval,$2.ival);Debug.trace("Parser: ISigNameList - NAME (" + $1.sval + "), ISigNameList matched");}
		;

SigType: {$$.ival = 0;}
	| ':' SigTypeAux {$$.ival = $2.ival;}
	;
	
SigTypeAux: INT {$$.ival = 1;}
		| INT '.' SigTypeAux {$$.ival = $3.ival + 1;}
		;

CompilationCommand:
	 {Debug.trace("Parser: CompilationCommand empty matched"); $$.sval = "StructuredCompile";}
	| LEAFCOMPRESSED {Debug.trace("Parser: CompilationCommand LEAFCOMPRESSED matched"); $$.sval = "LeafCompressedStructuredCompile";}
	| HIERARCHCOMPRESSED {Debug.trace("Parser: CompilationCommand HIERARCHCOMPRESSED matched"); $$.sval = "HierarchCompressedCompile";};
	

Program: 
		 NAME '=' CompilationCommand CmdStruct {
						Debug.trace("Parser: Program - NAME(" + $1.sval + ") = CompilationCommand CmdStruct matched");
						CmdStruct CS = ((PreCmdStruct) $4.obj).generate(compiler, new TreeMap<String,Integer>());
						compiler.progs.put($1.sval,new CmdStructProgram( $1.sval,CS, $3.sval));
						compiler.context.newProgram($1.sval);
				 	}/*
		| NAME '=' LEAFCOMPRESSED CmdStruct {
						Debug.trace("Parser: Program -  NAME(" + $1.sval + ") = LEAFCOMPRESSED CmdStruct matched");
						CmdStruct CS = (CmdStruct) $4.obj;
						compiler.progs.put($1.sval,new CmdStructProgram( $1.sval,CS, "LeafCompressedStructuredCompile"));
						compiler.context.newProgram($1.sval);
				 	}
		| NAME '=' HIERARCHCOMPRESSED CmdStruct {
						Debug.trace("Parser: Program - NAME(" + $1.sval + ") = HIERARCHCOMPRESSED CmdStruct matched");
						CmdStruct CS = (CmdStruct) $4.obj;
						compiler.progs.put($1.sval,new CmdStructProgram( $1.sval,CS, "HierarchCompressedCompile"));
						compiler.context.newProgram($1.sval);
				 	}*/;



CmdStruct: PROCESS '(' ParameterList ')'
		{
			$$.obj = new PreCSLeaf($1.sval, (Vector<ProcessParameter>) $3.obj);
			Debug.trace("Parser: CmdStruct - PROCESS '(' ParameterList ')' matched");
		}
		| OVERSEER PROCESS '(' ParameterList ')'
		{
			$$.obj = new PreCSOverseer($2.sval, (Vector<ProcessParameter>) $4.obj);
			Debug.trace("Parser: CmdStruct - OVERSEER PROCESS '(' ParameterList ')' matched");
		}
		| '<' CmdStructList '>'
		{
			Vector<PreCmdStruct> vec = (Vector<PreCmdStruct>) $2.obj;
			$$.obj = new PreCSNode(vec);
			Debug.trace("Parser: CmdStruct - < CmdStructList > matched");
		}
		| '<' CmdStructList '|' NAME FROM Parameter TO Parameter '>' 
		{
			$$.obj = new PreCSParamaterisedNode((Vector<PreCmdStruct>)$2.obj, $4.sval, (ProcessParameter)$6.obj, (ProcessParameter)$8.obj);
		
			Debug.trace("Parser: CmdStruct -  '<' CmdStructList '|' NAME FROM Parameter TO Parameter '>'  matched");
		}
		
;

CmdStructList: 	CmdStruct 
				{ 
					Vector<PreCmdStruct> vec = new Vector<PreCmdStruct>();
					vec.add( (PreCmdStruct) $1.obj);
					$$.obj = vec;
					Debug.trace("Parser: CmdStructList - CmdStruct matched");
				}
				| CmdStructList ',' CmdStruct
				{
					Vector<PreCmdStruct> vec = (Vector<PreCmdStruct>)$1.obj;
					vec.add( (PreCmdStruct) $3.obj);
					$$.obj = vec;
					Debug.trace("Parser: CmdStructList - CmdStructList , CmdStruct matched");
				}
;

ParameterList:
	{ Debug.trace("Parser: ParameterList - empty matched"); $$.obj = new Vector<ProcessParameter>(); }
	| NonEmptyParameterList { Debug.trace("Parser: ParameterList - NonEmptyParameterList matched"); $$.obj = $1.obj; } 
;

NonEmptyParameterList:
	Parameter 
		{ 	Debug.trace("Parser: NonEmptyParameterList - Parameter matched"); 
			Vector<ProcessParameter> vec = new Vector<ProcessParameter>(); 
			vec.add((ProcessParameter) $1.obj);
			$$.obj = vec;
		}
	|  NonEmptyParameterList ',' Parameter
		{ 	Debug.trace("Parser: NonEmptyParameterList - Parameter ',' NonEmptyParameterList matched"); 
			Vector<ProcessParameter> vec = (Vector<ProcessParameter>) $1.obj; 
			vec.add((ProcessParameter) $3.obj);
			$$.obj = vec;
		}
	;
	
Parameter:
	CONST			{ $$.obj = new ConstParam($1.sval); Debug.trace("Parser: Parameter - CONST (" + $1.sval + ") matched");}
	|NAME			{ $$.obj = new NameParam($1.sval); Debug.trace("Parser: Parameter - NAME (" + $1.sval + ") matched");}
	| NUM			{ $$.obj = new NumParam($1.ival); Debug.trace("Parser: Parameter -  NUM (" + $1.ival + ") matched");}
	| Parameter '+' Parameter	
		{ 	$$.obj = new AddParam( (ProcessParameter)$1.obj, (ProcessParameter) $3.obj); 
			Debug.trace("Parser: Parameter - Parameter + Parameter matched");}
	| Parameter '-' Parameter	
		{ 	$$.obj = new SubParam( (ProcessParameter)$1.obj, (ProcessParameter) $3.obj); 
			Debug.trace("Parser: Parameter - Parameter - Parameter matched");}
	| Parameter '*' Parameter	
		{ 	$$.obj = new MultParam((ProcessParameter)$1.obj, (ProcessParameter) $3.obj); 
			Debug.trace("Parser: Parameter - Parameter * Parameter matched");}
	| Parameter '/' Parameter	
		{ 	$$.obj = new DivParam( (ProcessParameter)$1.obj, (ProcessParameter) $3.obj);
			Debug.trace("Parser: Parameter - Parameter / Parameter matched");}
	| Parameter '%' Parameter	
		{ 	$$.obj = new ModParam( (ProcessParameter)$1.obj, (ProcessParameter) $3.obj); 
			Debug.trace("Parser: Parameter - Parameter % Parameter matched");}
	| '-' Parameter %prec NEG	
		{ $$.obj = new NegParam((ProcessParameter)$2.obj); Debug.trace("Parser: Parameter - -Parameter matched");}
	| '(' Parameter ')'	{ $$.obj = $2.obj; Debug.trace ("Parameter - (Parameter) matched");}
	;

Proc: 	PROCESS '(' NonEmptyNumList ')'
	{
		assert  $3.obj instanceof Vector :  "Proc - PROCESS(NonEmptyNumList) $3.obj incorrect type";
		
		Process process = compiler.procs.get($1.sval);
		Vector<Integer> args = (Vector<Integer>)$3.obj;
		if (process.numArgs() != args.size()) throw new ParserException($1.sval + " given " + args.size() + " arguments but expects " + process.numArgs());
		
		$$.obj = new InitializedProc(process, args);
		Debug.trace("Parser: Proc - PROCESS ( NonEmptyNumList ) matched");
	}		
	|PROCESS '('')'
	{
		Process process = compiler.procs.get($1.sval);
		//trace("process = " + process);
		
		if (process.numArgs() != 0) throw new ParserException($1.sval + " given no arguments but expects " + process.numArgs());
		
		$$.obj = new InitializedProc(process, new Vector<Integer>());
		Debug.trace("Parser: Proc - PROCESS () matched");
		//trace("Proc - returning " + $$.obj);
	}
	;

NonEmptyNumList:	NUM	
			{
				Vector<Integer> vector = new Vector<Integer>();
				vector.add($1.ival);
				$$.obj = vector;
				Debug.trace("Parser: NonEmptyNumList - NUM(" + $1.ival + ") matched");
				//trace("NonEmptyNumList returning " + $$.obj);
			}
			| NonEmptyNumList ',' NUM	
			{	
				assert  $1.obj instanceof Vector :  "NonEmptyNumList - NonEmptyNumList, Num $1.obj incorrect type";
				((Vector<Integer>)$1.obj).add($3.ival); 
				$$.obj = $1.obj; 
				Debug.trace("Parser: NonEmptyNumList - NonEmptyNumList, NUM(" + $3.ival + ") mathced");
				//trace("NonEmptyNumList returning " + $$.obj);
			}
			;

Process: NAME '(' ArgList ')' {compiler.context.registerConstants((Vector<String>)$3.obj); currentProcess = $1.sval; Debug.trace("Parser: Registering Constants for " + $1.sval);} '=' Prog AssignedVariables	{ 	
				Process proc =  new Process($1.sval, (Vector<String>)$3.obj, (Cmd)$7.obj, (VariableAssignment)$8.obj) ;
				proc.addLocalIVars(compiler.context.getLocalIVars());
				proc.addLocalBVars(compiler.context.getLocalBVars());
				proc.addLocalIAVars(compiler.context.getLocalIAVars());
				proc.addLocalBAVars(compiler.context.getLocalBAVars());
				compiler.procs.put($1.sval,proc);
				compiler.context.newProcess($1.sval);
				Debug.trace("Parser: Process - NAME(" + $1.sval + ") (ArgList) = Prog matched");
				compiler.context.deregisterConstsAndLocals();//needs doing before next arglist is found
							    //as lexer should flag names in arglist as names not consts	
			}
	 ;

ArgList:				{$$.obj = new Vector<String>(); Debug.trace("Parser: ArgList - Empty String matched\n");}
		| NonEmptyArgList 	{$$.obj = $1.obj;Debug.trace("Parser: ArgList - NonEmptyArgList matched\n");}
		;

NonEmptyArgList: NAME				{Vector<String> v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: NonEmptyArgList - Name(" + $1.sval + ") matched");}
		|NonEmptyArgList ',' NAME 	{ ((Vector<String>)$1.obj).add($3.sval); $$.obj = $1.obj;Debug.trace("Parser: NonEmptyArgList - Name(" + $1.sval + "), NonEmptyArgList matched");}
		;


Prog:	Cmd		{$$.obj = $1.obj;Debug.trace("Parser: Prog - Cmd matched");}
	| '{' Cmds '}'	{$$.obj = $2.obj;Debug.trace("Parser: Prog - { Cmds } matched");}
	;

Cmds:	Cmd		{$$.obj = $1.obj; Debug.trace("Parser: Cmds - Cmd matched");}
	| Cmd ';'	{$$.obj = $1.obj; Debug.trace("Parser: Cmds - Cmd ; matched");}
	| Cmd ';' Cmds	{$$.obj = new Seq( (Cmd)$1.obj, (Cmd)$3.obj); Debug.trace("Parser: Cmds - Cmd;Cmd matched\n");}
	| INT LocalINameList ';' Cmds {$$.obj = $4.obj; Debug.trace("Parser: Cmds -  INT LocalINameList ; matched\n");}
	| INT '[' ']' LocalIANameList ';' Cmds {$$.obj = $6.obj; Debug.trace("Parser: Cmds -  INT [ ] LocalIANameList ; matched\n");}
	| BOOL LocalBNameList ';' Cmds {$$.obj = $4.obj; Debug.trace("Parser: Cmds -  BOOL LocalBNameList ; matched\n");}
	| BOOL '[' ']' LocalBANameList ';' Cmds {$$.obj = $6.obj; Debug.trace("Parser: Cmds - BOOL [ ] LocalBANameList ; matched\n");}
	;

Cmd:	SKIP				{$$.obj = new Skip(); Debug.trace("Parser: Cmd - Skip matched\n");}
	| ITER Prog			{$$.obj = new Iter( (Cmd) $2.obj); Debug.trace("Parser: Cmd - ITER Prog matched\n");}
	| WHILE BExpr DO Prog		{$$.obj = new While( (BExpr) $2.obj, (Cmd) $4.obj); Debug.trace("Parser: Cmd - WHILE BExpr DO Prog matched\n");}
	//| DO Prog WHILE BExpr		{$$.obj = new DoWhile( (Cmd) $2.obj, (BExpr) $4.obj); Debug.trace("Parser: Cmd - DO Prog WHILE BExpr matched\n");}
	| IF BExpr THEN Prog %prec IFX 	{$$.obj = new If( (BExpr) $2.obj, (Cmd) $4.obj); Debug.trace("Parser: Cmd - IF BExpr THEN Prog matched\n");} 
	| IF BExpr THEN Prog ELSE Prog	{$$.obj = new If( (BExpr) $2.obj, (Cmd) $4.obj, (Cmd) $6.obj); Debug.trace("Parser: Cmd - IF BExpr THEN Prog Else Prog matched\n");}
	| IExpr ASSIGN IExpr		{$$.obj = new IAssign( (IExpr) $1.obj, (IExpr) $3.obj); Debug.trace("Parser: Cmd - IVName ASSIGN IExpr matched\n");}
	| BExpr ASSIGN BExpr		{$$.obj = new BAssign( (BExpr) $1.obj, (BExpr) $3.obj); Debug.trace("Parser: Cmd - BVName ASSIGN BExpr matched\n");}
	| SIG '(' SIGNAME Indices ')'		{$$.obj = new Sig($3.sval, (Vector<IConst>)$4.obj); Debug.trace("Parser: Cmd - Sig(" + $3.sval + " Indices)  matched\n");}
	| ISIG '(' ISIGNAME Indices ',' IExpr ')'{$$.obj = new ISig($3.sval, (Vector<IConst>)$4.obj, (IExpr) $6.obj); Debug.trace("Parser: Cmd - ISig(" + $3.sval + " Indices, IEXpr)  matched\n");}
	| ATOMIC Prog			{$$.obj = new Atomic((Cmd)$2.obj); Debug.trace("Parser: Cmd - Atomic Prog matched\n");}
	| PROCESS '(' IConstList ')'
		{ Debug.trace("Parser: Cmd - PROCESS '(' IConstList ')' matched\n");
			Process macro = compiler.procs.get($1.sval);
			Vector<IConst> args = (Vector<IConst>) $3.obj;
			if (macro.numArgs() != args.size()) throw new ParserException($1.sval + " given " + args.size() + " arguments but expects " + macro.numArgs());
			$$.obj = macro.prog.substituteConst(macro.consts, args) ;
			for (String name : macro.getLocalIVars())
				compiler.context.newLocalI(name);
			for (String name : macro.getLocalBVars())
				compiler.context.newLocalB(name);
			for (String name : macro.getLocalIAVars())
				compiler.context.newLocalIA(name);
			for (String name : macro.getLocalBAVars())
				compiler.context.newLocalBA(name);
		}			
//	| NOOP				{$$.obj = new NoOp(); Debug.trace("Parser: Cmd - NOOP matched\n");}
	;
	
IConstList: {Debug.trace("Parser: IConstList - Empty matched"); $$.obj = new Vector<IConst>();}
			| NonEmptyIConstList {Debug.trace("Parser: IConstList - NonEmptyIConstList matched"); $$.obj = $1.obj;}
			;
			
NonEmptyIConstList: IConst {	Debug.trace("Parser: NonEmptyIConstList - IConst matched"); 
								Vector<IConst> vec = new Vector<IConst>();
								vec.add((IConst) $1.obj);
								$$.obj = vec;}
					| NonEmptyIConstList ',' IConst {	Debug.trace("Parser: NonEmptyIConstList - NonEmptyIConstList ',' IConst matched"); 
								Vector<IConst> vec = (Vector<IConst>)$1.obj;
								vec.add((IConst) $3.obj);
								$$.obj = vec;}
			;

LocalINameList:	LocalIName					{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalINameList - LocalIName matched");}
		| LocalIName ',' LocalINameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalINameList - LocalIName , LocalINameList matched");}
		;
		
LocalBNameList:	LocalBName					{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalBNameList - LocalBName matched");}
		| LocalBName ',' LocalBNameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalBNameList - LocalBName , LocalBNameList matched");}
		;


LocalIANameList:	LocalIAName					{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalIANameList - LocalIAName matched");}
		| LocalIAName ',' LocalIANameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalIANameList - LocalIAName , LocalIANameList matched");}
		;
		
LocalBANameList:	LocalBAName					{Vector<String>v = new Vector<String>(); v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalBANameList - LocalBAName matched");}
		| LocalBAName ',' LocalBANameList	{Vector<String> v = (Vector<String>)$3.obj; v.add($1.sval); $$.obj = v;Debug.trace("Parser: LocalBANameList - LocalBAName , LocalBANameList matched");}
		;

LocalIName:		NAME	{compiler.context.newLocalI($1.sval);$$.sval = $1.sval;Debug.trace("Parser: LocalIName - NAME (" + $1.sval + ") matched");}
		;
LocalBName:		NAME	{compiler.context.newLocalB($1.sval);$$.sval = $1.sval;Debug.trace("Parser: LocalBName - NAME (" + $1.sval + ") matched");}
		;
LocalIAName:		NAME	{compiler.context.newLocalIA($1.sval);$$.sval = $1.sval;Debug.trace("Parser: LocalIAName - NAME (" + $1.sval + ") matched");}
		;
LocalBAName:		NAME	{compiler.context.newLocalBA($1.sval);$$.sval = $1.sval;Debug.trace("Parser: LocalBAName - NAME (" + $1.sval + ") matched");}
		;

Indices:  { $$.obj = new Vector<IConst>(); Debug.trace("Parser: Indices - Empty matched");}
		|  Indices '.' IConst { Vector<IConst> vec = (Vector<IConst>) $1.obj; vec.add((IConst) $3.obj); $$.obj = vec;}
;

/*IVName:	IVAR			{ $$.obj = new IVarName($1.sval); Debug.trace("Parser: IVName - IVAR (" + $1.sval + ") matched");}
	| IAVAR '[' IExpr ']'	{ $$.obj = new IArrayName($1.sval, (IConst) $3.obj); Debug.trace("Parser: IVName - IAVAR (" + $1.sval + ") [IConst] matched");}
	;

BVName:	BVAR			{ $$.obj = new BVarName($1.sval); Debug.trace("Parser: BVName - BVAR (" + $1.sval + ") matched");}
	| BAVAR '[' IExpr ']'	{ $$.obj = new BArrayName($1.sval, (IConst) $3.obj); Debug.trace("Parser: BVName - BAVAR (" + $1.sval + ") [IConst] matched");}
	;*/

AssignedVariables:
	{$$.obj = new VariableAssignment(); Debug.trace("Parser: AssignedVariables - Empty matched");}
	| WITH '{' AssignedVariableList '}' { $$.obj = $3.obj; Debug.trace("Parser: AssignedVariables - WITH '{' AssignedVariableList '}' matched");}
;

AssignedVariableList:
	IVAR {VariableAssignment va = new VariableAssignment(); va.is.add($1.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - IVAR matched");}
	|BVAR {VariableAssignment va = new VariableAssignment(); va.bs.add($1.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - BVAR matched");}
	|IAVAR {VariableAssignment va = new VariableAssignment(); va.ias.add($1.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - IAVAR matched");}
	|BAVAR {VariableAssignment va = new VariableAssignment(); va.bas.add($1.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - BAVAR matched");}
	|IAVAR '[' IConst ']' {VariableAssignment va = new VariableAssignment(); va.iaAccesses.add(new Access($1.sval,(IConst)$3.obj)); $$.obj = va; Debug.trace("Parser: AssignedVariableList - IAVAR '[' IConst ']' matched");}
	|BAVAR '[' IConst ']' {VariableAssignment va = new VariableAssignment(); va.baAccesses.add(new Access($1.sval,(IConst)$3.obj)); $$.obj = va; Debug.trace("Parser: AssignedVariableList - BAVAR '[' IConst ']' matched");}
	| AssignedVariableList ',' IVAR {VariableAssignment va = (VariableAssignment)$1.obj; va.is.add($3.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' IVAR  matched");}
	| AssignedVariableList ',' BVAR {VariableAssignment va = (VariableAssignment)$1.obj; va.bs.add($3.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' BVAR  matched");}
	| AssignedVariableList ',' IAVAR {VariableAssignment va = (VariableAssignment)$1.obj; va.ias.add($3.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' IAVAR  matched");}
	| AssignedVariableList ',' BAVAR {VariableAssignment va = (VariableAssignment)$1.obj; va.bas.add($3.sval); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' BAVAR  matched");}
	| AssignedVariableList ',' IAVAR '[' IConst ']' {VariableAssignment va = (VariableAssignment)$1.obj; va.iaAccesses.add(new Access($3.sval,(IConst)$5.obj)); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' IAVAR '[' IConst ']' matched");}
	| AssignedVariableList ',' BAVAR '[' IConst ']' {VariableAssignment va = (VariableAssignment)$1.obj; va.baAccesses.add(new Access($3.sval,(IConst)$5.obj)); $$.obj = va; Debug.trace("Parser: AssignedVariableList - AssignedVariableList ',' BAVAR '[' IConst ']' matched");}
;

IConst: CONST			{ $$.obj = new VarIConst($1.sval); Debug.trace("Parser: IConst - CONST (" + $1.sval + ") matched");}
	| NUM			{ $$.obj = new NumIConst($1.ival); Debug.trace("Parser: IConst -  NUM (" + $1.ival + ") matched");}
	| IConst '+' IConst	{ $$.obj = new AddConst( (IConst)$1.obj, (IConst) $3.obj); Debug.trace("Parser: IConst - IConst + IConst matched");}
	| IConst '-' IConst	{ $$.obj = new SubConst( (IConst)$1.obj, (IConst) $3.obj); Debug.trace("Parser: IConst - IConst - IConst matched");}
	| IConst '*' IConst	{ $$.obj = new MultConst((IConst)$1.obj, (IConst) $3.obj); Debug.trace("Parser: IConst - IConst * IConst matched");}
	| IConst '/' IConst	{ $$.obj = new DivConst( (IConst)$1.obj, (IConst) $3.obj); Debug.trace("Parser: IConst - IConst / IConst matched");}
	| IConst '%' IConst	{ $$.obj = new ModConst( (IConst)$1.obj, (IConst) $3.obj); Debug.trace("Parser: IConst - IConst % IConst matched");}
	| '-' IConst %prec NEG	{ $$.obj = new NegIConst((IConst)$2.obj); Debug.trace("Parser: IConst - -IConst matched");}
	| '(' IConst ')'	{ $$.obj = $2.obj; Debug.trace ("IConst - (IConst) matched");}
	;

BExpr: '!' BExpr		{$$.obj = new Not( (BExpr) $2.obj); Debug.trace("Parser: BExpr -  !BExpr matched");}
	| BExpr AND BExpr	{$$.obj = new And( (BExpr) $1.obj, (BExpr) $3.obj);Debug.trace("Parser: BExpr - BExpr AND BExpr matched");}
	| BExpr OR BExpr	{$$.obj = new Or( (BExpr) $1.obj, (BExpr) $3.obj);Debug.trace("Parser: BExpr - BExpr OR BExpr matched");}
	| IExpr '=' IExpr	{$$.obj = new Eq( (IExpr) $1.obj, (IExpr) $3.obj); Debug.trace("Parser: BExpr - IExpr = IExpr matched");}
	| IExpr NEQ IExpr	{$$.obj = new Neq( (IExpr) $1.obj, (IExpr) $3.obj); Debug.trace("Parser: BExpr - IExpr NEQ IExpr matched");}
	| IExpr GE IExpr	{$$.obj = new Ge( (IExpr) $1.obj, (IExpr) $3.obj); Debug.trace("Parser: BExpr - IExpr GE IExpr matched");}
	| IExpr '>' IExpr	{$$.obj = new Gt( (IExpr) $1.obj, (IExpr) $3.obj); Debug.trace("Parser: BExpr - IExpr > IExpr matched");}
	| IExpr LE IExpr	{$$.obj = new Le( (IExpr) $1.obj, (IExpr) $3.obj);  Debug.trace("Parser: BExpr - IExpr LE IExpr matched");}
	| IExpr '<' IExpr	{$$.obj = new Lt( (IExpr) $1.obj, (IExpr) $3.obj);  Debug.trace("Parser: BExpr - IExpr < IExpr matched");}
	| '(' BExpr ')'		{$$.obj = $2.obj;  Debug.trace("Parser: BExpr - (BExpr) matched");}
	| TRUE			{$$.obj = new True();  Debug.trace("Parser: BExpr - TRUE matched");}
	| FALSE			{$$.obj = new False(); Debug.trace("Parser: BExpr - FALSE matched");}
	| BVar			{$$.obj = $1.obj; Debug.trace("Parser: BExpr - BVar matched");}
	;

BVar:	BVAR			{ $$.obj = new BVar($1.sval);  Debug.trace("Parser: BVar - BVAR("+$1.sval+") matched");}
	| BAVAR '[' IExpr ']'	{ $$.obj = new BArray($1.sval, (IExpr) $3.obj);Debug.trace("Parser: BVar - BVAR("+$1.sval+") [IConst]  matched");}
	| LOCALBVAR { $$.obj = new LocalBVar($1.sval,currentProcess);Debug.trace("Parser: BVar - LOCALBVAR("+$1.sval+") matched");}
	| LOCALBAVAR  '[' IExpr ']' { $$.obj = new LocalBArray($1.sval,currentProcess,(IExpr) $3.obj);Debug.trace("Parser: BVar - LOCALBAVAR("+$1.sval+") '[' IExpr ']' matched");}
	;

IExpr:	IExpr '+' IExpr		{$$.obj = new Plus( (IExpr) $1.obj, (IExpr) $3.obj);Debug.trace("Parser: IExpr - IExpr + IExpr matched");}
	|IExpr '-' IExpr	{$$.obj = new Minus( (IExpr) $1.obj, (IExpr) $3.obj);Debug.trace("Parser: IExpr - IExpr - IExpr matched");}
	|IExpr '*' IExpr	{$$.obj = new Times( (IExpr) $1.obj, (IExpr) $3.obj);Debug.trace("Parser: IExpr - IExpr * IExpr matched");}
	|IExpr '/' IExpr	{$$.obj = new Div( (IExpr) $1.obj, (IExpr) $3.obj);Debug.trace("Parser: IExpr - IExpr / IExpr matched");}
	|IExpr '%' IExpr	{$$.obj = new Mod( (IExpr) $1.obj, (IExpr) $3.obj);Debug.trace("Parser: IExpr - IExpr % IExpr matched");}
	|MAX '(' IExpr ',' IExpr ')' {$$.obj = new Max( (IExpr)$3.obj, (IExpr)$5.obj); Debug.trace("Parser: IExpr - MAX ( IExpr, IExpr) matched");}
	|MIN '(' IExpr ',' IExpr ')' {$$.obj = new Min( (IExpr)$3.obj, (IExpr)$5.obj); Debug.trace("Parser: IExpr - MIN ( IExpr, IExpr) matched");}
	|'-' IExpr %prec NEG	{$$.obj = new Neg( (IExpr) $2.obj);Debug.trace("Parser: IExpr - -IExpr matched");}
	| '(' IExpr ')'		{$$.obj = $2.obj; Debug.trace("Parser: IExpr - (IExpr) matched");}
	| CONST			{ $$.obj = new IConstExpr(new VarIConst($1.sval)); Debug.trace("Parser: IExpr - CONST (" + $1.sval + ") matched");}
	| NUM			{ $$.obj = new IConstExpr(new NumIConst($1.ival)); Debug.trace("Parser: IExpr -  NUM (" + $1.ival + ") matched");}
	| IVar			{$$.obj = $1.obj;Debug.trace("Parser: IExpr - IVar matched");}
	;

IVar:	IVAR			{ $$.obj = new IVar($1.sval);Debug.trace("Parser: IVar - IVAR("+$1.sval+") matched");}
	| IAVAR '[' IExpr ']'	{ $$.obj = new IArray($1.sval, (IExpr) $3.obj);Debug.trace("Parser: IVar - IVAR("+$1.sval+") [IConst]  matched");}
	| LOCALIVAR { $$.obj = new LocalIVar($1.sval,currentProcess);Debug.trace("Parser: IVar - LOCALIVAR("+$1.sval+") matched");}
	| LOCALIAVAR  '[' IExpr ']' { $$.obj = new LocalIArray($1.sval,currentProcess,(IExpr) $3.obj);Debug.trace("Parser: IVar - LOCALIAVAR("+$1.sval+") '[' IExpr ']' matched");}
	;

Assertion:	ASSERT ALWAYS BExpr IN PROGRAM	
		{ 
			compiler.asserts.add(new BoolAssertion(compiler.progs.get($5.sval),(BExpr)$3.obj));
			Debug.trace("Parser: Assertion - ASSERT ALWAYS BExpr IN PROGRAM(" + $5.sval + ") matched");
		}
		|ASSERT NEVER BExpr IN PROGRAM	
		{ 
			compiler.asserts.add(new BoolAssertion(compiler.progs.get($5.sval),new Not((BExpr)$3.obj)));
			Debug.trace("Parser: Assertion - ASSERT NEVER BExpr IN PROGRAM(" + $5.sval + ") matched");
		}
		|ASSERT NOSIGNAL '{' SignalList '}' IN PROGRAM
		{
			compiler.asserts.add(new SignalAssertion(compiler.progs.get($7.sval), (Vector<String>)$4.obj));
			Debug.trace("Parser: Assertion - ASSERT NOSIGNAL { SignalList } IN PROGRAM(" + $7.sval + ")"); 
		}
		|ASSERT CSPASSERT IN PROGRAM
		{
			compiler.asserts.add(new CSPAssertion($2.sval,compiler.progs.get($4.sval)));
			Debug.trace("Parser: Assertion - ASSERT CSPASSERT IN PROGRAM(" + $4.sval + ")");
		}
		| ASSERT Proc SEQREFINEMENT Proc
		{
			compiler.asserts.add(new SeqRefinementAssertion((CompilableProc)$2.obj,(CompilableProc)$4.obj));
			Debug.trace("Parser: Assertion - ASSERT Proc SEQREFINEMENT Proc (" + $4.sval + ")");
		
		}
		| ASSERT Proc GENREFINEMENT Proc
		{
			compiler.asserts.add(new GenRefinementAssertion((CompilableProc)$2.obj,(CompilableProc)$4.obj));
			Debug.trace("Parser: Assertion - ASSERT Proc GENREFINEMENT Proc (" + $4.sval + ")");
		
		}
		| ASSERT ParRefTerm PARREFINEMENT ParRefTerm
		{
			compiler.asserts.add(new ParRefinementAssertion((CmdStructProgram)$2.obj,(CmdStructProgram)$4.obj));
			Debug.trace("Parser: Assertion - ASSERT ParRefTerm PARREFINEMENT ParRefTerm (" + $4.sval + ")");
		
		}
;

ParRefTerm:
	Proc {$$.obj = new CmdStructProgram($1.obj.toString(), new CSLeaf( (CompilableProc)$1.obj)); Debug.trace("Parser: ParRefTerm - Proc matched");}
	| PROGRAM {$$.obj = compiler.progs.get($1.sval); Debug.trace("Parser: ParRefTerm - PROGRAM matched");}
;

SignalList: 	Signal			{Vector<String> vec = new Vector<String>(); vec.add($1.sval); $$.obj = vec; Debug.trace("Parser: SignalList - Signal(" + $1.sval + ") matched");}
		| Signal ',' SignalList	{((Vector<String>)$3.obj).add($1.sval); $$.obj = $3.obj; Debug.trace("Parser: SignalList - Signal(" + $1.sval + ") , SignalList matched");}
		;
		
Signal:		SIGNAME	{$$.sval = $1.sval; Debug.trace("Parser: Signal - SIGNAME(" + $1.sval + ") matched");}
		|ISIGNAME '.' NUM {$$.sval = $1.sval + "." + String.valueOf($3.ival); Debug.trace("Parser: Signal - ISIGNAME(" + $1.sval + ") . NUM(" + String.valueOf($3.ival));}
//		| OUTOFRANGE {$$.sval = "outofrange"; Debug.trace("Parser: Signal - OUTOFRANGE matched");}
		| ERROR {$$.sval = "error"; Debug.trace("Parser: Signal - ERROR matched");} 
		| VERROR {$$.sval = "verror"; Debug.trace("Parser: Signal - VERROR matched");} 
		;

RefDeclaration: 
	PARREADS '=' '{' VarNameList '}' {compiler.context.setParReads((Vector<String>)$4.obj); Debug.trace("Parser: RefDeclaration - PARREADS '{' VarNameList '}' matched");} 
	| PARWRITES '=' '{' VarNameList '}' {compiler.context.setParWrites((Vector<String>)$4.obj); Debug.trace("Parser: RefDeclaration - PARWRITES '{' VarNameList '}' matched");}  
	| SEQREADS '=' '{' VarNameList '}' {compiler.context.setSeqReads((Vector<String>)$4.obj); Debug.trace("Parser: RefDeclaration - SEQREADS '{' VarNameList '}' matched");}  
	| SEQWRITES '=' '{' VarNameList '}' {compiler.context.setSeqWrites((Vector<String>)$4.obj); Debug.trace("Parser: RefDeclaration - SEQWRITES '{' VarNameList '}' matched");} 
;

VarNameList:
	{$$.obj = new Vector<String>(); Debug.trace("Parser: VarNameList - Empty List matched");}
	| NonEmptyVarNameList {$$.obj = $1.obj; Debug.trace("Parser: VarNameList - Empty List matched");}
;

NonEmptyVarNameList:
	NonEmptyVarNameList ',' VarName {Vector<String> vec = (Vector<String>)$1.obj; vec.add($3.sval); $$.obj = vec; Debug.trace("Parser: NonEmptyVarNameList - NonEmptyVarNameList ',' VarName matched");}
	| VarName {Vector<String> vec = new Vector<String>(); vec.add($1.sval); $$.obj = vec; Debug.trace("Parser: NonEmptyVarNameList - VarName matched");}
;

VarName:
	IVAR {$$.sval = $1.sval; Debug.trace("Parser: VarName - IVAR matched");}
	|BVAR {$$.sval = $1.sval; Debug.trace("Parser: VarName - BVAR matched");}
	|IAVAR {$$.sval = $1.sval; Debug.trace("Parser: VarName - IAVAR matched");}
	|BAVAR {$$.sval = $1.sval; Debug.trace("Parser: VarName - BAVAR matched");}
	|IAVAR '[' NUM ']' {$$.sval = $1.sval + "[" + $3.ival + "]"; Debug.trace("Parser: VarName - IAVAR '[' NUM ']' matched");}
	|BAVAR '[' NUM ']' {$$.sval = $1.sval + "[" + $3.ival + "]"; Debug.trace("Parser: VarName - BAVAR matched");}
;


%%
  
  

  private Yylex lexer;

  private Compiler compiler;

  private String currentProcess;

  public Compiler getCompiler()
  {
	yyparse();
	compiler.generateLocalVarNames();
	return compiler;		
  }

  private int yylex (){
    int yyl_return = -1;
    try{
      yylval = new ParserVal(0);
      yyl_return = lexer.yylex();

    }
    catch(IOException e){yyerror(e.getMessage());}
    return yyl_return;
  }


  public void yyerror (String error) {
    throw new ParserException("Error Near " + lexer.location() + ": " + error);
  }


  public Parser(Reader r) {
    compiler = new Compiler();
    lexer = new Yylex(r, this, compiler.context);
  }

	void declareContentSize( Vector<String> names, int size)
	{
		for (String name : names)
		{
			compiler.context.setContentSize(name,size);
		}
	}
	void declareArrayContentSize( Vector<String> names, int size)
	{
		for (String name : names)
		{
			compiler.context.setArrayContentSize(name,size);
		}
	}

	void declareArrayIndexSize( Vector<String> names, int size)
	{
		for (String name : names)
		{
			compiler.context.setArrayIndexSize(name,size);
		}
	}
	void declareDirty( Vector<String> names)
	{
		
		for (String name : names)
		{
			compiler.context.declareDirty(name);
		}
	}