import org.junit.*;
import static org.junit.Assert.*;

import java.io.*;
import java.util.*;

public class IntegrationTests
{
	protected Vector<TableDataInterpreter.AssertionOutput> interpret(String input) throws FileNotFoundException, InterruptedException, IOException
	{
		final String svlfile = "integrationTestsSVLFile.svl";
		final String cspfile = "integrationTestsCSPFile.csp";
		
		//Create svl file
		(new PrintStream(svlfile)).print(input);
		
		Interpreter interpreter = new Interpreter(svlfile,cspfile,null,null);
		TableDataInterpreter tdi = new TableDataInterpreter(interpreter.compiler.context.globalConsts, interpreter.compiler.context.dirtyVars);
		
		interpreter.interpret(new MergeTraceConstructor(), tdi);
		
		return tdi.getResults();
	}
	
	protected void checkResult(Vector<TableDataInterpreter.AssertionOutput> results, int index, String assertion, boolean expected)
	{
		TableDataInterpreter.AssertionOutput ao = results.elementAt(index);
		assertEquals("Checking wrong assertion - fix test!", assertion, ao.assertion);
		assertEquals("Assertion " + assertion + " returned unexpected result", expected, ao.result);
	}
	
	@Test public void checkExtAtomicWhenFalse() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkExtAtomicWhenFalse()");
		String code =
		"%%MinI = 0\n" +
		"%%MaxI = 1\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {MinI..MaxI}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"%%ext_atomic=false\n" +
		"int x,y\n" +
		"ParReads={x,y}\n" +
		"PP() = iter \n" +
		"{\n" +
		"    x  := 1;        \n" +
		"    x  := 0;                  \n" +
		"    y  := 1;                   \n" +
		"    y  := 0\n" +
		"}                   \n" +
		"QQ() = iter \n" +
		"{\n" +
		"    x  := 1;\n" +
		"    y  := 1;\n" +
		"    x  := 0;\n" +
		"    y  := 0\n" +
		"}\n" +
		"assert PP() [P= QQ()\n" +
		"assert QQ() [P= PP()\n" +
		"assert PP() [G= QQ()\n" +
		"assert QQ() [G= PP()\n"
		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		for (TableDataInterpreter.AssertionOutput ao : results)
		{
			assertTrue(ao.result);
		}
		
	}

	@Test public void checkExtAtomicWhenTrue() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkExtAtomicWhenTrue()");
		String code =
		"%%MinI = 0\n" +
		"%%MaxI = 1\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {MinI..MaxI}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"%%ext_atomic=true\n" +
		"int x,y\n" +
		"ParReads={x,y}\n" +
		"PP() = iter \n" +
		"{\n" +
		"    x  := 1;        \n" +
		"    x  := 0;                  \n" +
		"    y  := 1;                   \n" +
		"    y  := 0\n" +
		"}                   \n" +
		"QQ() = iter \n" +
		"{\n" +
		"    x  := 1;\n" +
		"    y  := 1;\n" +
		"    x  := 0;\n" +
		"    y  := 0\n" +
		"}\n" +
		"assert PP() [P= QQ()\n" +
		"assert QQ() [P= PP()\n" +
		"assert PP() [G= QQ()\n" +
		"assert QQ() [G= PP()\n"
		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert PP() [P= QQ()", false);
		checkResult(results, 1, "assert QQ() [P= PP()", true);
		checkResult(results, 2, "assert PP() [G= QQ()", false);
		checkResult(results, 3, "assert QQ() [G= PP()", true);
		
	}
	
	@Test public void checkLampba2() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkLampba2()");
		String code =
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 6\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {1..N}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"\n" +
		"sig css : int, cse : int;\n" +
		"dirty int[] number;\n" +
		"dirty bool[] choosing;\n" +
		"const N = 3;\n" +
		"\n" +
		"PP(i) = iter\n" +
		"{\n" +
		"    int j,acc;\n" +
		"    choosing[i] := true;\n" +
		"    j := 1;\n" +
		"    acc := 0;\n" +
		"    while j <= N do\n" +
		"    {\n" +
		"		acc := max(acc, number[j]);\n" +
		"		j := j + 1;\n" +
		"    };\n" +
		"    number[i] := acc + 1;\n" +
		"    choosing[i] := false;\n" +
		"    j := 1;\n" +
		"    while (j <= N) do\n" +
		"    {\n" +
		"		while choosing[j] do skip;\n" +
		"		if (j < i) then\n" +
		"			while number[j] > 0 && number[i] >= number[j] do skip;\n" +
		"		if j > i then\n" +
		"			while number[j] > 0 && number[i] > number[j] do skip;\n" +
		"		j := j + 1;\n" +
		"	};\n" +
		"	sig(css.i);\n" +
		"	sig(cse.i);\n" +
		"	number[i] := 0\n" +
		"\n" +
		"}\n" +
		"\n" +
		"WideStruct = hierarchCompress < PP(1),PP(2),PP(3)>\n" +
		"PairStruct = hierarchCompress < <PP(1),PP(2)>,PP(3)>\n" +
		"BalancedStruct = hierarchCompress < <PP(1),PP(2)>, PP(3)>\n" +
		"\n" +
		"WideStructUncompressed = < PP(1),PP(2),PP(3)>\n" +
		"\n" +
		"%%SPEC = css?i -> cse!i -> SPEC\n" +
		"\n" +
		"assert %- STOP [T= WideStructUncompressed \\{|css,cse|} -% in WideStructUncompressed\n" +
		"assert %- SPEC [T= WideStructUncompressed \\{|error,verror|} -% in WideStructUncompressed\n" +
		"assert %- STOP [T= WideStruct \\{|css,cse|} -% in WideStruct\n" +
		"assert %- SPEC [T= WideStruct \\{|error,verror|} -% in WideStruct\n" +
		"assert %- SPEC [T= PairStruct \\{|error,verror|} -% in PairStruct\n" +
		"assert %- SPEC [T= BalancedStruct \\{|error,verror|} -% in BalancedStruct\n" +
		"assert %- NEVS(8,{|css,cse|}) [T= BalancedStruct -% in BalancedStruct\n" +
		"\n" +
		"%%NEVS(0,X) = STOP\n" +
		"%%NEVS(n,X) = [] x:X @ x -> NEVS(n-1,X)\n"

		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert  STOP [T= WideStructUncompressed \\{|css,cse|} ", false);
		checkResult(results, 1, "assert  SPEC [T= WideStructUncompressed \\{|error,verror|} ", true);
		checkResult(results, 2, "assert  STOP [T= WideStruct \\{|css,cse|} ", false);
		checkResult(results, 3, "assert  SPEC [T= WideStruct \\{|error,verror|} ", true);
		checkResult(results, 4, "assert  SPEC [T= PairStruct \\{|error,verror|} ", true);
		checkResult(results, 5, "assert  SPEC [T= BalancedStruct \\{|error,verror|} ", true);
		checkResult(results, 6, "assert  NEVS(8,{|css,cse|}) [T= BalancedStruct ", false);
	}

	@Test public void checkSimpson() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkSimpson()");
		String code =
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 4\n" +
		"%%InitB = false\n" +
		"%%InitI = 0\n" +
		"%%ditype = {0,1}\n" +
		"%%dctype = {0,1}\n" +
		"\n" +
		"dirty int[4] slot;\n" +
		"dirty int[] index;\n" +
		"\n" +
		"dirty int reading,latest;\n" +
		"\n" +
		"sig collision;\n" +
		"\n" +
		"Writer() = iter\n" +
		"{\n" +
		"\n" +
		"    int wpair, windex, inp, wval;\n" +
		"    wval := inp;\n" +
		"    wpair := 1 - reading;\n" +
		"    windex := 1 - index[wpair];\n" +
		"    slot[2*wpair + windex] := wval;\n" +
		"    index[wpair] := windex;\n" +
		"    latest := wpair;\n" +
		"}\n" +
		"\n" +
		"Reader() = iter\n" +
		"{\n" +
		"    int rpair, rindex, outp;\n" +
		"    rpair := latest;\n" +
		"    reading := rpair;\n" +
		"    rindex := index[rpair];\n" +
		"    outp := slot[2*rpair + rindex];\n" +
		"    if outp = 1 then sig(collision);\n" +
		"}\n" +
		"\n" +
		"WideStruct = hierarchCompress <Reader(),Writer()>\n" +
		"\n" +
		"assert %- STOP [T= WideStruct\\({|error,verror|}) -% in WideStruct\n" +
		"\n"
		
		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert  STOP [T= WideStruct\\({|error,verror|}) ", true);

	}
	
	
	@Test public void checkCompMax() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkCompMax()");
		String code =
		"%%MinI = 0\n" +
		"%%MaxI = 5\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {1..N}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"%%ext_atomic=false\n" +
		"\n" +
		"const N = 4;\n" +
		"\n" +
		"int %3 [] turn;\n" +
		"int j, temp;\n" +
		"\n" +
		"SeqWrites = {turn[3]}\n" +
		"SeqReads = {turn[3]}\n" +
		"ParReads = {turn}\n" +
		"ParWrites = {turn[1],turn[2],turn[4]}\n" +
		"\n" +
		"\n" +
		"MaxC(i) = \n" +
		"{\n" +
		"	//int j, temp;\n" +
		"	j := 1;\n" +
		"	while (j <= N) do\n" +
		"	{\n" +
		"		temp := turn[j];\n" +
		"		if temp >= turn[i] then\n" +
		"			turn[i] := temp;\n" +
		"		j := j + 1;\n" +
		"	};\n" +
		"	turn[i] := turn[i] + 1;\n" +
		"}\n" +
		"\n" +
		"MaxA(i) = \n" +
		"{\n" +
		"	//int j, temp;\n" +
		"	temp := 0;\n" +
		"	j := 1;\n" +
		"	while j <= N do\n" +
		"	{\n" +
		"		temp := max(turn[j],temp);\n" +
		"		j := j + 1;\n" +
		"	};\n" +
		"	turn[i] := temp + 1\n" +
		"}\n" +
		"\n" +
		"MaxB(i) =\n" +
		"{\n" +
		"	//int j;\n" +
		"	j := 1;\n" +
		"	while j <= N do\n" +
		"	{\n" +
		"		turn[i] := max(turn[j],turn[i]);\n" +
		"		j := j + 1\n" +
		"	};\n" +
		"	turn[i] := turn[i] + 1\n" +
		"\n" +
		"}\n" +
		"\n" +
		"MaxD(i) = \n" +
		"{\n" +
		"	//int j;\n" +
		"	j := 1;\n" +
		"	while j <= N do\n" +
		"	{\n" +
		"		if turn[i] < turn[j] then\n" +
		"			turn[i] := turn[j];\n" +
		"		j := j + 1\n" +
		"	};\n" +
		"	turn[i] := turn[i] + 1\n" +
		"\n" +
		"}\n" +
		"\n" +
		"assert MaxA(3) [G= MaxB(3)\n" +
		"assert MaxB(3) [G= MaxC(3)\n" +
		"assert MaxC(3) [G= MaxB(3)\n" +
		"assert MaxC(3) [G= MaxD(3)\n" +
		"\n" +
		"assert MaxD(3) [G= MaxB(3)\n" +
		"assert MaxB(3) [G= MaxA(3)\n" +
		"assert MaxD(3) [G= MaxA(3)\n" +
		"\n" +
		"assert MaxA(3) [S= MaxB(3)\n" +
		"assert MaxB(3) [S= MaxC(3)\n" +
		"assert MaxC(3) [S= MaxB(3)\n" +
		"assert MaxC(3) [S= MaxD(3)\n" +
		"\n" +
		"assert MaxD(3) [S= MaxB(3)\n" +
		"assert MaxB(3) [S= MaxA(3)\n" +
		"\n"

		
		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		

		checkResult(results, 0, "assert MaxA(3) [G= MaxB(3)", false);
		checkResult(results, 1, "assert MaxB(3) [G= MaxC(3)", true);
		checkResult(results, 2, "assert MaxC(3) [G= MaxB(3)", true);
		checkResult(results, 3, "assert MaxC(3) [G= MaxD(3)", false);


		checkResult(results, 4, "assert MaxD(3) [G= MaxB(3)", true);
		checkResult(results, 5, "assert MaxB(3) [G= MaxA(3)" , false);
		checkResult(results, 6, "assert MaxD(3) [G= MaxA(3)" , false);


		checkResult(results, 7, "assert MaxA(3) [S= MaxB(3)", true);
		checkResult(results, 8, "assert MaxB(3) [S= MaxC(3)", true);
		checkResult(results, 9, "assert MaxC(3) [S= MaxB(3)" , true);
		checkResult(results, 10, "assert MaxC(3) [S= MaxD(3)" , true);


		checkResult(results, 11, "assert MaxD(3) [S= MaxB(3)" , true);
		checkResult(results, 12, "assert MaxB(3) [S= MaxA(3)" , true);


	}
	
	@Test public void checkLocalVarTestCompilesWithoutError() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkLocalVarTestCompilesWithoutError()");
		String code =
		
		"%%ext_atomic=false \n" +
		"%%MinI = 0 \n" +
		"%%MaxI = 4 \n" +
		"%%InitB = false \n" +
		"%%InitI = 0 \n" +
		"%%ditype = {0,1} \n" +
		"%%dctype = {0,1} \n" +
		"int a = 2, b = 4; \n" +
		"P(i) = \n" +
		"{\n" +
		"	int c;\n" +
		"	int d;\n" +
		"	bool e;\n" +
		"   c := a;\n" +
		"	d := b;\n" +
		"	e := true\n" +
		"}\n" +
		"Q(i,j,k) = \n" +
		"{\n" +
		"	int c;\n" +
		"	int[] e;\n" +
		"	bool[] f,g;\n" +
		"	a := c;\n" +
		"	b := c;\n" +
		"	f[0] := g[e[c]];\n" +
		"}\n" +
		"Prog = <P(1),P(2),Q(1,2,3),Q(1,2,4),P(3)>\n"+
		"assert %- STOP [T= Prog -% in Prog\n" +
		"Prog2 = <P(1),Q(1,2,3),Q(6,6,6),P(6)>\n"+
		"assert %- STOP [T= Prog2 -% in Prog2\n"
		
		;
	
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert  STOP [T= Prog ", true);
		checkResult(results, 1, "assert  STOP [T= Prog2 ", true);

	}


@Test public void checkBakeryProofBug() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkLocalVarTestCompilesWithoutError()");
		String code =
		"\n" +
		"\n" +
		"\n" +
		"\n" +
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 9\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {2,4}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"\n" +
		"sig css : int, cse : int;\n" +
		"\n" +
		"bool randb;\n" +
		"int[] number;\n" +
		"int randi;\n" +
		"int[] temp;\n" +
		"bool[10] hits;\n" +
		"int[10] omap;\n" +
		"bool[] choosing;\n" +
		"const N = 5;\n" +
		"\n" +
		"ParWrites = {randi,randb}\n" +
		"\n" +
		"PP(i) = iter\n" +
		"{\n" +
		"    int j;\n" +
		"    choosing[i] := true;\n" +
		"    j := 1;\n" +
		"    while j <= N do\n" +
		"      {if j%2=0  then \n" +
		"                   {if i!=j then atomic temp[i] := number[j];\n" +
		"                                 atomic number[i] := max(temp[i],number[j])}\n" +
		"                else\n" +
		"                while randb do\n" +
		"                   {atomic number[i] := max(number[i],randi)};\n" +
		"       j := j+1};\n" +
		"    atomic {if number[i] < 8 then\n" +
		"               if randb then number[i] := number[i] + 1\n" +
		"                        else number[i] := number[i] + 2\n" +
		"               else number[i] := number[i]+1};\n" +
		"    choosing[i] := false;\n" +
		"    j := 1;\n" +
		"    while (j <= N) do\n" +
		"    {if j%2=1 then\n" +
		"      while randb do\n" +
		"	{	while randb do skip;\n" +
		"		if (j < i) then\n" +
		"                        {\n" +
		"			while randi > 0 && number[i]>= randi do skip};\n" +
		"		if (i < j) then\n" +
		"                        {\n" +
		"			while randi > 0 && number[i]> randi do skip};\n" +
		"	} \n" +
		"      else\n" +
		"	{     while choosing[j] do skip;\n" +
		"		if (j < i) then\n" +
		"                       {atomic temp[i] := number[j];\n" +
		"                        while (temp[i] > 0) && (number[i] >= temp[i]) do \n" +
		"                              {atomic temp[i] := number[j];\n" +
		"                              }};\n" +
		"		if (j > i) then\n" +
		"                       {atomic temp[i] := number[j];\n" +
		"			while (temp[i] > 0) && (number[i] > temp[i]) do \n" +
		"                               {atomic temp[i] := number[j];\n" +
		"                                }}};\n" +
		"      j := j+1};\n" +
		"   sig(css.i);\n" +
		"   sig(cse.i);\n" +
		"   atomic number[i] := 0\n" +
		"}\n" +
		"\n" +
		"Dummy() = skip with {randi,randb}\n" +
		"\n" +
		"\n" +
		"\n" +
		"%%SPEC = css?i -> cse!i -> SPEC\n" +
		"\n" +
		"\n" +
		"\n" +
		"LO() =\n" +
		"{\n" +
		"int kk,nextw;\n" +
		"hits[number[2]] := true;\n" +
		"hits[number[4]] := true;\n" +
		"hits[temp[2]] := true;\n" +
		"hits[temp[4]] := true;\n" +
		"nextw := 0;\n" +
		"kk := 0;\n" +
		"while kk<9 do\n" +
		"  {kk := kk+1;\n" +
		"   if hits[kk] then \n" +
		"     {\n" +
		"      nextw := nextw+2;\n" +
		"      omap[kk] := nextw;\n" +
		"      hits[kk] := false;\n" +
		"     }};\n" +
		"number[2] := omap[number[2]];\n" +
		"number[4] := omap[number[4]];\n" +
		"temp[2] := omap[temp[2]];\n" +
		"temp[4] := omap[temp[4]];\n" +
		"kk := 0;\n" +
		"while kk<9 do\n" +
		"   {kk := kk+1;\n" +
		"    omap[kk] := 0}\n" +
		"}\n" +
		"\n" +
		"\n" +
		"WideStruct = hierarchCompress < PP(2),PP(4),Dummy()>\n" +
		"WideStructOS = hierarchCompress < PP(2),PP(4),Dummy(),overseer LO()>\n" +
		"\n" +
		"assert %- STOP [T= WideStruct \\{|css,cse|} -% in WideStruct\n" +
		"assert %- SPEC [T= WideStruct \\{|error,verror|} -% in WideStruct\n" +
		"assert %- STOP [T= WideStructOS \\{|css,cse|} -% in WideStruct\n" +
		"assert %- SPEC [T= WideStructOS \\{|error,verror|} -% in WideStructOS\n" +
		"assert %- SPEC [T= WideStructOS -% in WideStructOS\n" +
		"\n" +
		"";
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert  STOP [T= WideStruct \\{|css,cse|} ", false);
		checkResult(results, 1, "assert  SPEC [T= WideStruct \\{|error,verror|} ", true);
		checkResult(results, 2, "assert  STOP [T= WideStructOS \\{|css,cse|} ", true);
		checkResult(results, 3, "assert  SPEC [T= WideStructOS \\{|error,verror|} ", false);
		checkResult(results, 4, "assert  SPEC [T= WideStructOS ", false);

	}


@Test public void checkBakeryProof4() throws FileNotFoundException, InterruptedException, IOException
	{
		// System.out.println("checkLocalVarTestCompilesWithoutError()");
		String code =
				"\n" +
		"\n" +
		"\n" +
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 9\n" +
		"%%InitB = false\n" +
		"%%InitI = MinI\n" +
		"%%ditype = {2,4}\n" +
		"%%dctype = {MinI..MaxI}\n" +
		"\n" +
		"sig css : int, cse : int;\n" +
		"\n" +
		"bool randb;\n" +
		"int[] number;\n" +
		"int randi;\n" +
		"int[] temp;\n" +
		"bool[10] hits;\n" +
		"int[10] omap;\n" +
		"bool[] choosing;\n" +
		"const N = 5;\n" +
		"\n" +
		"ParWrites = {randi,randb}\n" +
		"\n" +
		"PP(i) = iter\n" +
		"{\n" +
		"    int j;\n" +
		"    bool b;\n" +
		"    choosing[i] := true;\n" +
		"    j := 1;\n" +
		"    while j <= N do\n" +
		"      {if j%2=0  then \n" +
		"                   {if i!=j then atomic temp[i] := number[j];\n" +
		"                                 atomic number[i] := max(temp[i],number[j])}\n" +
		"                else\n" +
		"                while randb do\n" +
		"                   {atomic number[i] := max(number[i],randi)};\n" +
		"       j := j+1};\n" +
		"    atomic {if number[i] < 8 then\n" +
		"               if randb then number[i] := number[i] + 1\n" +
		"                        else number[i] := number[i] + 2\n" +
		"               else number[i] := number[i]+1};\n" +
		"    choosing[i] := false;\n" +
		"    j := 1;\n" +
		"    while (j <= N) do\n" +
		"    {if j%2=1 then\n" +
		"      while randb do\n" +
		"	{	while randb do skip;\n" +
		"		if (j < i) then\n" +
		"                        {atomic b := randi > 0 && number[i]>= randi;\n" +
		"			while b do atomic b := randi > 0 && number[i]>= randi};\n" +
		"		if (i < j) then\n" +
		"                        {atomic b := randi > 0 && number[i]>randi;\n" +
		"			while b do atomic b := randi > 0 && number[i]>randi};\n" +
		"	} \n" +
		"      else\n" +
		"	{     while choosing[j] do skip;\n" +
		"		if (j < i) then\n" +
		"                       {atomic temp[i] := number[j];\n" +
		"                        atomic {b := (temp[i] > 0) && (number[i] >= temp[i])};\n" +
		"                        while b do \n" +
		"                              {atomic temp[i] := number[j];\n" +
		"                              atomic b := temp[i] > 0 && number[i] >= temp[i]}};\n" +
		"		if (j > i) then\n" +
		"                       {atomic temp[i] := number[j];\n" +
		"                        atomic b := temp[i] > 0 && number[i] > temp[i];\n" +
		"			while b do \n" +
		"                               {atomic temp[i] := number[j];\n" +
		"                               atomic b := temp[i] > 0 && number[i] > temp[i]};\n" +
		"                                }};\n" +
		"      j := j+1};\n" +
		"   sig(css.i);\n" +
		"   sig(cse.i);\n" +
		"   atomic number[i] := 0\n" +
		"}\n" +
		"\n" +
		"Dummy() = skip with {randi,randb}\n" +
		"\n" +
		"\n" +
		"\n" +
		"%%SPEC = css?i -> cse!i -> SPEC\n" +
		"\n" +
		"\n" +
		"\n" +
		"LO() =\n" +
		"{\n" +
		"int kk,nextw;\n" +
		"hits[number[2]] := true;\n" +
		"hits[number[4]] := true;\n" +
		"hits[temp[2]] := true;\n" +
		"hits[temp[4]] := true;\n" +
		"nextw := 0;\n" +
		"kk := 0;\n" +
		"while kk<9 do\n" +
		"  {kk := kk+1;\n" +
		"   if hits[kk] then \n" +
		"     {\n" +
		"      nextw := nextw+2;\n" +
		"      omap[kk] := nextw;\n" +
		"      hits[kk] := false;\n" +
		"     }};\n" +
		"number[2] := omap[number[2]];\n" +
		"number[4] := omap[number[4]];\n" +
		"temp[2] := omap[temp[2]];\n" +
		"temp[4] := omap[temp[4]];\n" +
		"kk := 0;\n" +
		"while kk<9 do\n" +
		"   {kk := kk+1;\n" +
		"    omap[kk] := 0}\n" +
		"}\n" +
		"\n" +
		"\n" +
		"WideStruct = hierarchCompress < PP(2),PP(4),Dummy()>\n" +
		"WideStructOS = hierarchCompress < PP(2),PP(4),Dummy(),overseer LO()>\n" +
		"\n" +
		"assert %- STOP [T= WideStruct \\{|css,cse|} -% in WideStruct\n" +
		"assert %- SPEC [T= WideStruct \\{|error,verror|} -% in WideStruct\n" +
		"assert %- STOP [T= WideStructOS \\{|css,cse|} -% in WideStruct\n" +
		"assert %- SPEC [T= WideStructOS \\{|error,verror|} -% in WideStructOS\n" +
		"assert %- SPEC [T= WideStructOS -% in WideStructOS\n" +
		"\n" +
		"";
		Vector<TableDataInterpreter.AssertionOutput> results = interpret(code);
		
		checkResult(results, 0, "assert  STOP [T= WideStruct \\{|css,cse|} ", false);
		checkResult(results, 1, "assert  SPEC [T= WideStruct \\{|error,verror|} ", true);
		checkResult(results, 2, "assert  STOP [T= WideStructOS \\{|css,cse|} ", true);
		checkResult(results, 3, "assert  SPEC [T= WideStructOS \\{|error,verror|} ", true);
		checkResult(results, 4, "assert  SPEC [T= WideStructOS ", true);

	}

}