import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

public class TableDataInterpreterTest
{
	

	protected String concatEvents(Vector<TableDataInterpreter.Entry> entries)
	{
		String result = "";
		for (TableDataInterpreter.Entry e : entries)
			result += e.text + "\n";
		return result;
	}
	
	
	protected String eval(String name, int val)
	{
		return name + " is " + val + "\n";
	}	
	protected String evalDirty(String name, int val)
	{
		return name + " is (dirtily) " + val + "\n";
	}
	protected String writeStart(String name, int val)
	{
		return name + " starts being assigned " + val + "\n";
	}
	protected String write(String name, int val)
	{
		return name + " assigned " + val + "\n";
	}
	protected String eval(String name, boolean val)
	{
		return name + " is " + val + "\n";
	}	
	protected String evalDirty(String name, boolean val)
	{
		return name + " is (dirtily) " + val + "\n";
	}
	protected String writeStart(String name, boolean val)
	{
		return name + " starts being assigned " + val + "\n";
	}
	protected String write(String name, boolean val)
	{
		return name + " assigned " + val + "\n";
	}

	@Test
	public void singleDirtyVarReadsAndWrites()
	{
		TableDataInterpreter tdi = new TableDataInterpreter(new TreeMap<String,Integer>(), 
			new Vector<String>(Arrays.asList(new String [] {"x"}))); 
	
		tdi.startAssertion(new DummyAssert(), false);
		
		tdi.iVarEval("x",1,0);
		tdi.iVarWrite("x",2,0);
		tdi.iVarEval("x",1,0);
		tdi.iVarWrite("x",2,0);
		tdi.iVarEval("x",2,0);
		
		String expected = eval("x",1) + writeStart("x",2) + evalDirty("x",1) + write("x",2) + eval("x",2);
		
		tdi.endTrace();
		String actual = concatEvents(tdi.getResults().get(0).counterexample);
	
		assertEquals(expected, actual);
	}

	
	
	@Test
	public void twoDirtyVarOneNormalReadsAndWrites()
	{
		TableDataInterpreter tdi = new TableDataInterpreter(new TreeMap<String,Integer>(), 
			new Vector<String>(Arrays.asList(new String [] {"x","y"}))); 
	
		tdi.startAssertion(new DummyAssert(), false);
		
		tdi.iVarEval("x",1,0);
		tdi.iVarWrite("x",2,0);
		tdi.iVarEval("x",1,0);
		tdi.bVarEval("y",true,0);
		tdi.iVarEval("z",1,0);
		tdi.bVarWrite("y",false,0);
		tdi.iVarEval("z",1,0);
		tdi.bVarEval("y",true,0);
		tdi.iVarWrite("z",2,0);
		tdi.bVarWrite("y",false,0);
		tdi.iVarEval("z",2,0);
		tdi.bVarEval("y",false,0);
		tdi.iVarEval("x",1,0);
		tdi.iVarWrite("x",2,0);
		tdi.iVarEval("x",2,0);
		
		String expected = eval("x",1) + writeStart("x",2) + evalDirty("x",1) + eval("y",true) + eval("z",1) + writeStart("y",false) + eval("z",1) + evalDirty("y",true) + write("z",2) + write("y",false) + eval("z",2) + eval("y",false) + evalDirty("x",1) + write("x",2) + eval("x",2);
		
		tdi.endTrace();
		String actual = concatEvents(tdi.getResults().get(0).counterexample);
	
		assertEquals(expected, actual);
	}	
	
	@Test
	public void dirtyArrayReadsAndWrites()
	{
		TableDataInterpreter tdi = new TableDataInterpreter(new TreeMap<String,Integer>(), 
			new Vector<String>(Arrays.asList(new String [] {"a"}))); 
	
		tdi.startAssertion(new DummyAssert(), false);
		
		tdi.iVarEval("a[1]",1,0);
		tdi.iVarWrite("a[1]",1,0);
		tdi.iVarEval("a[1]",1,0);
		tdi.iVarEval("a[2]",1,0);
		tdi.iVarWrite("a[2]",1,0);
		tdi.iVarEval("a[1]",1,0);
		tdi.iVarEval("a[2]",1,0);
		tdi.iVarEval("a[3]",1,0);
		tdi.iVarWrite("a[2]",1,0);
		tdi.iVarEval("a[1]",1,0);
		tdi.iVarEval("a[2]",1,0);
		tdi.iVarWrite("a[1]",1,0);
		tdi.iVarEval("a[1]",1,0);

		
		String expected = eval("a[1]",1) + writeStart("a[1]",1) + evalDirty("a[1]", 1) + eval("a[2]",1) + writeStart("a[2]",1) +  evalDirty("a[1]", 1) + evalDirty("a[2]", 1) +  eval("a[3]",1) + write("a[2]",1) + evalDirty("a[1]", 1) +  eval("a[2]",1) +  write("a[1]",1) + eval("a[1]",1);
		
		tdi.endTrace();
		String actual = concatEvents(tdi.getResults().get(0).counterexample);
	
		assertEquals(expected, actual);
	}	

}

class DummyAssert implements Assertion
{
	public int numProcs(){return 0;}

	public String process(int i){return "";}

	public String ownerOfSig(String sig,Map<String,Integer> consts){return "";}
	public int indexOfOwnerOfSig(String sig, Map<String,Integer> consts){return 0;}
	public boolean usesSig(String sig, Map<String, Integer> consts){return false;}
	public String genCode(Context con){return "";}
	public Iterator<CompilableProc> iterator(){return (new Vector<CompilableProc>()).iterator();}
	
// 	public boolean usesMinusOne() {return false;}
}