import org.junit.*;
import static org.junit.Assert.*;

import java.io.*;
import java.util.*;

public class CmdStructTest
{
	private Compiler compiler1;
	private Process p,q,r,s;
	
	
	@Before
	public void setup()
	{
		Vector<String> pArgs = new Vector<String>();
		pArgs.add("pArg1");
		p = new Process("P",pArgs,new Skip(), new VariableAssignment());
		
		Vector<String> qArgs = new Vector<String>();
		qArgs.add("qArg1");
		q = new Process("Q",qArgs,new Seq(new Skip(), new Skip()), new VariableAssignment());
		
		Vector<String> rArgs = new Vector<String>();
		r = new Process("R",rArgs,new Iter(new Skip()), new VariableAssignment());
		
		Vector<String> sArgs = new Vector<String>();
		sArgs.add("sArg1");
		sArgs.add("sArg2");
		s = new Process("S",sArgs,new While(new Eq(new IConstExpr(new VarIConst("sArg1")),new IConstExpr(new VarIConst("sArg2"))),new Skip()), new VariableAssignment());
		
		compiler1 = new Compiler();
		
		compiler1.procs.put("P",p);
		compiler1.procs.put("Q",q);
		compiler1.procs.put("R",r);
		compiler1.procs.put("S",s);
	
	}
	
	@Test public void preRGeneratesR() 
	{
		
		CmdStruct actual = (new PreCSLeaf("R", new Vector<ProcessParameter>())).generate(compiler1, new TreeMap<String,Integer>());
		CmdStruct expected = new CSLeaf(new InitializedProc(r,new Vector<Integer>()));
		assertEquals(expected,actual);
	
	}
	
	@Test public void preOverseerRGeneratesR() 
	{
		
		CmdStruct actual = (new PreCSOverseer("R", new Vector<ProcessParameter>())).generate(compiler1, new TreeMap<String,Integer>());
		CmdStruct expected = new CSOverseer(new InitializedProc(r,new Vector<Integer>()));
		assertEquals(expected,actual);
	
	}
	
	@Test public void prePOnePlusXXEq2GeneratesP3() 
	{
		Vector<ProcessParameter> args = new Vector<ProcessParameter>();
		args.add(new AddParam(new NumParam(1), new NameParam("x")));
		
		TreeMap<String,Integer> nameMap = new TreeMap<String,Integer>();
		nameMap.put("x",2);
		
		Vector<Integer>expectedArgs = new Vector<Integer>();
		expectedArgs.add(3);
		
		CmdStruct actual = (new PreCSLeaf("P", args)).generate(compiler1, nameMap);
		CmdStruct expected = new CSLeaf(new InitializedProc(p,expectedArgs));
		assertEquals(expected,actual);
	
	}
	
	
	@Test public void prePTwoTimesYQ3MinusYGenerates() 
	{
		Vector<ProcessParameter> args1 = new Vector<ProcessParameter>();
		args1.add(new MultParam(new NumParam(2), new NameParam("y")));
		
		Vector<ProcessParameter> args2 = new Vector<ProcessParameter>();
		args2.add(new SubParam(new NameParam("y"), new NumParam(3) ));
		
		TreeMap<String,Integer> nameMap = new TreeMap<String,Integer>();
		nameMap.put("y",3);
		
		Vector<Integer>expectedArgs1 = new Vector<Integer>();
		expectedArgs1.add(6);
		Vector<Integer>expectedArgs2 = new Vector<Integer>();
		expectedArgs2.add(0);
		
		
		Vector<PreCmdStruct> preNodes = new Vector<PreCmdStruct>();
		preNodes.add(new PreCSLeaf("P", args1));
		preNodes.add(new PreCSLeaf("Q", args2));
		
		CmdStruct actual = (new PreCSNode(preNodes)).generate(compiler1, nameMap);
		
		Vector<CmdStruct>nodes = new Vector<CmdStruct>();
		nodes.add(new CSLeaf(new InitializedProc(p,expectedArgs1)));
		nodes.add(new CSLeaf(new InitializedProc(q,expectedArgs2)));
		
		CmdStruct expected = new CSNode(nodes);
		assertEquals(expected,actual);
	
	}
	
	
	@Test public void prePMinusIQIMod2IFrom1To2Generates() 
	{
		Vector<ProcessParameter> args1 = new Vector<ProcessParameter>();
		args1.add(new NegParam(new NameParam("i")));
		
		Vector<ProcessParameter> args2 = new Vector<ProcessParameter>();
		args2.add(new ModParam(new NameParam("i"), new NumParam(2) ));
		
		TreeMap<String,Integer> nameMap = new TreeMap<String,Integer>();
		
		
		
		Vector<PreCmdStruct> preNodes = new Vector<PreCmdStruct>();
		preNodes.add(new PreCSLeaf("P", args1));
		preNodes.add(new PreCSLeaf("Q", args2));
		
		CmdStruct actual = (new PreCSParamaterisedNode(preNodes,"i", new NumParam(1), new NumParam(2))).generate(compiler1, nameMap);
				
		Vector<Integer>expectedArgs1 = new Vector<Integer>();
		expectedArgs1.add(-1);
		Vector<Integer>expectedArgs2 = new Vector<Integer>();
		expectedArgs2.add(1);
		Vector<Integer>expectedArgs3 = new Vector<Integer>();
		expectedArgs3.add(-2);
		Vector<Integer>expectedArgs4 = new Vector<Integer>();
		expectedArgs4.add(0);
		
		Vector<CmdStruct>nodes = new Vector<CmdStruct>();
		nodes.add(new CSLeaf(new InitializedProc(p,expectedArgs1)));
		nodes.add(new CSLeaf(new InitializedProc(q,expectedArgs2)));
		nodes.add(new CSLeaf(new InitializedProc(p,expectedArgs3)));
		nodes.add(new CSLeaf(new InitializedProc(q,expectedArgs4)));
		
		CmdStruct expected = new CSNode(nodes);
		assertEquals(expected,actual);
	
	}
	
	
	@Test public void preSJJPlus1Q2JFrom1To2TimesNGenerates() 
	{
		Vector<ProcessParameter> args1 = new Vector<ProcessParameter>();
		args1.add((new NameParam("j")));
		args1.add(new AddParam(new NameParam("j"), new NumParam(1)));
		
		Vector<ProcessParameter> args2 = new Vector<ProcessParameter>();
		args2.add(new NumParam(2) );
		
		TreeMap<String,Integer> nameMap = new TreeMap<String,Integer>();
		
		compiler1.context.globalConsts.put("N",2);
		
		
		Vector<PreCmdStruct> preNodes1 = new Vector<PreCmdStruct>();
		preNodes1.add(new PreCSLeaf("S", args1));
		preNodes1.add(new PreCSLeaf("Q", args2));
		
		
		Vector<PreCmdStruct> preNodes2 = new Vector<PreCmdStruct>();
		preNodes2.add(new PreCSNode(preNodes1));
		
		
		CmdStruct actual = (new PreCSParamaterisedNode(preNodes2,"j", new NumParam(1), new MultParam(new ConstParam("N"), new NumParam(2)))).generate(compiler1, nameMap);
				
				
		Vector<Integer>expectedArgs1a = new Vector<Integer>();
		expectedArgs1a.add(1);
		expectedArgs1a.add(2);
		Vector<Integer>expectedArgs1b = new Vector<Integer>();
		expectedArgs1b.add(2);
				
		Vector<Integer>expectedArgs2a = new Vector<Integer>();
		expectedArgs2a.add(2);
		expectedArgs2a.add(3);
		Vector<Integer>expectedArgs2b = new Vector<Integer>();
		expectedArgs2b.add(2);
				
		Vector<Integer>expectedArgs3a = new Vector<Integer>();
		expectedArgs3a.add(3);
		expectedArgs3a.add(4);
		Vector<Integer>expectedArgs3b = new Vector<Integer>();
		expectedArgs3b.add(2);
				
		Vector<Integer>expectedArgs4a = new Vector<Integer>();
		expectedArgs4a.add(4);
		expectedArgs4a.add(5);
		Vector<Integer>expectedArgs4b = new Vector<Integer>();
		expectedArgs4b.add(2);
		
		Vector<CmdStruct>nodes1 = new Vector<CmdStruct>();
		nodes1.add(new CSLeaf(new InitializedProc(s,expectedArgs1a)));
		nodes1.add(new CSLeaf(new InitializedProc(q,expectedArgs1b)));
		
		Vector<CmdStruct>nodes2 = new Vector<CmdStruct>();
		nodes2.add(new CSLeaf(new InitializedProc(s,expectedArgs2a)));
		nodes2.add(new CSLeaf(new InitializedProc(q,expectedArgs2b)));
		
		Vector<CmdStruct>nodes3 = new Vector<CmdStruct>();
		nodes3.add(new CSLeaf(new InitializedProc(s,expectedArgs3a)));
		nodes3.add(new CSLeaf(new InitializedProc(q,expectedArgs3b)));
		
		Vector<CmdStruct>nodes4 = new Vector<CmdStruct>();
		nodes4.add(new CSLeaf(new InitializedProc(s,expectedArgs4a)));
		nodes4.add(new CSLeaf(new InitializedProc(q,expectedArgs4b)));
		
		Vector<CmdStruct>outerNodes = new Vector<CmdStruct>();
		outerNodes.add(new CSNode(nodes1));
		outerNodes.add(new CSNode(nodes2));
		outerNodes.add(new CSNode(nodes3));
		outerNodes.add(new CSNode(nodes4));
		
		CmdStruct expected = new CSNode(outerNodes);
		assertEquals(expected,actual);
	
	}
}