import org.junit.*;
import static org.junit.Assert.*;
import static org.junit.matchers.JUnitMatchers.*;


import java.io.*;
import java.util.*;

public class CompilerTest
{

	private Compiler compiler1;
	private String context1;
	private String pi;
	private String qijk;
	//private String prog;
	private String[] asserts1;

	private Compiler compiler2;
	private String context2;
	private String[] asserts2;

	@Before
	public void setup()
	{
		Parser parser1 = new Parser(new StringReader(
		"int a = 2, b = 4; \n" +
		"P(i) = \n" +
		"{\n" +
		"	int c;\n" +
		"	int d;\n" +
		"	bool e;\n" +
		"   c := a;\n" +
		"	d := b;\n" +
		"	e := true\n" +
		"}\n" +
		"Q(i,j,k) = \n" +
		"{\n" +
		"	int c;\n" +
		"	int[] e;\n" +
		"	bool[] f,g;\n" +
		"	a := c;\n" +
		"	b := c;\n" +
		"	f[0] := g[e[c]];\n" +
		"}\n" +
		"Prog = <P(1),P(2),Q(1,2,3),Q(1,2,4),P(3)>\n"+
		"assert %- STOP [T= Prog -% in Prog\n" +
		"Prog2 = <P(1),Q(1,2,3),Q(6,6,6),P(6)>\n"+
		"assert %- STOP [T= Prog2 -% in Prog2\n"
		
		));
		compiler1 = parser1.getCompiler();
		context1 = compiler1.context.genCode();
		pi = compiler1.procs.get("P").genCode(compiler1.context);
		qijk = compiler1.procs.get("Q").genCode(compiler1.context);
		//prog = compiler1.progs.get("Prog").genCode(compiler1.context);
		asserts1 = new String [] {compiler1.asserts.elementAt(0).genCode(compiler1.context),compiler1.asserts.elementAt(1).genCode(compiler1.context)};
		
		
		Parser parser2 = new Parser(new StringReader(
		"int a,b;\n" +
		"bool c,d;\n" +
		"int[3] e,f;\n" +
		"int[] i;\n" +
		"bool[4] g,h;\n" +
		"ParReads = {a,e[1]};\n" +
		"ParWrites = {c,g[2]};\n" +
		"SeqReads = {f,h};\n" +
		"SeqWrites = {h,d,i};\n" +
		"P(j) ={ a := j;}\n" +
		"Q() ={ b := 2;}\n" +
		"Prog = < P(1),P(2) >\n" +
		"assert P(1) [S= Q()\n" +
		"assert P(2) [G= P(3)\n" +
		"assert Prog [P= Q()"
		));
		compiler2 = parser2.getCompiler();
		context2 = compiler2.context.genCode();
		asserts2 = new String [] { compiler2.asserts.elementAt(0).genCode(compiler2.context), compiler2.asserts.elementAt(1).genCode(compiler2.context), compiler2.asserts.elementAt(2).genCode(compiler2.context)};
		
		
		
	}
	
	@Test public void P1HasCorrectLocalIVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("c","P",new Vector<Integer>(Arrays.asList(new Integer [] { 1}))) , 
		 new LocalVarUsage("d","P",new Vector<Integer>(Arrays.asList(new Integer [] { 1})))
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.firstElement()).p.getLocalIVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void P2HasCorrectLocalIVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("c","P",new Vector<Integer>(Arrays.asList(new Integer [] { 2}))) , 
		 new LocalVarUsage("d","P",new Vector<Integer>(Arrays.asList(new Integer [] { 2})))
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(1)).p.getLocalIVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void P3HasCorrectLocalIVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("c","P",new Vector<Integer>(Arrays.asList(new Integer [] { 3}))) , 
		 new LocalVarUsage("d","P",new Vector<Integer>(Arrays.asList(new Integer [] { 3})))
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(4)).p.getLocalIVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void Q123HasCorrectLocalIVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("c","Q",new Vector<Integer>(Arrays.asList(new Integer [] { 1,2,3}))) 
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(2)).p.getLocalIVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void Q124HasCorrectLocalIVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("c","Q",new Vector<Integer>(Arrays.asList(new Integer [] { 1,2,4}))) 
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(3)).p.getLocalIVars();
		
		assertEquals(expected,actual);
	}
	
	@Test public void P1HasCorrectLocalBVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("e","P",new Vector<Integer>(Arrays.asList(new Integer [] { 1}))) }  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.firstElement()).p.getLocalBVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void P1HasCorrectLocalIAVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>();
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.firstElement()).p.getLocalIAVars();
		
		assertEquals(expected,actual);
	}		
	@Test public void P1HasCorrectLocalBAVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>();
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.firstElement()).p.getLocalBAVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void Q123HasCorrectLocalBVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>();
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(2)).p.getLocalBVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void Q123HasCorrectLocalIAVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("e","Q",new Vector<Integer>(Arrays.asList(new Integer [] { 1,2,3}))) 
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(2)).p.getLocalIAVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void Q123HasCorrectLocalBAVars() 
	{
		Vector<LocalVarUsage>  expected = new Vector<LocalVarUsage>(Arrays.asList(new LocalVarUsage [] 
		{new LocalVarUsage("f","Q",new Vector<Integer>(Arrays.asList(new Integer [] { 1,2,3}))),
		new LocalVarUsage("g","Q",new Vector<Integer>(Arrays.asList(new Integer [] { 1,2,3})))
		}  ) );
		Vector<LocalVarUsage> actual = ((CSLeaf)((CSNode)((CmdStructProgram)((CSPAssertion)compiler1.asserts.firstElement()).prog).CS).seq.elementAt(2)).p.getLocalBAVars();
		
		assertEquals(expected,actual);
	}	
	@Test public void context1HasCs1stEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_c(1) = IV.3"));
	}
	@Test public void context1HasCs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_c(2) = IV.5"));
	}
	@Test public void context1HasCs3rdEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_c(3) = IV.9"));
	}	
	@Test public void context1HasDs1stEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_d(1) = IV.4"));
	}
	@Test public void context1HasDs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_d(2) = IV.6"));
	}
	@Test public void context1HasDs3rdEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_d(3) = IV.10"));
	}	
	@Test public void context1HasDs4thEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_P_d(6) = IV.13"));
	}
		
	@Test public void context1HasQsCs1stEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_Q_c(1,2,3) = IV.7"));
	}		
	@Test public void context1HasQsCs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_Q_c(1,2,4) = IV.8"));
	}		
	@Test public void context1HasQsCs3rdEncoding() 
	{
		assertThat(context1,containsString("LocalIVarMapping_Q_c(6,6,6) = IV.11"));
	}

	@Test public void context1HasEs1stEncoding() 
	{
		assertThat(context1,containsString("LocalBVarMapping_P_e(1) = BV.1"));
	}		
	@Test public void context1HasEs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalBVarMapping_P_e(2) = BV.2"));
	}
	@Test public void context1HasQsEs1stEncoding() 
	{
		assertThat(context1,containsString("LocalIAVarMapping_Q_e(1,2,3) = IA.1"));
	}
	@Test public void context1HasQsEs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalIAVarMapping_Q_e(1,2,4) = IA.2"));
	}
	@Test public void context1HasFs1stEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_f(1,2,3) = BA.1"));
	}
	@Test public void context1HasFs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_f(1,2,4) = BA.3"));
	}
	@Test public void context1HasFs3rdEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_f(6,6,6) = BA.5"));
	}
	@Test public void context1HasGs1stEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_g(1,2,3) = BA.2"));
	}
	@Test public void context1HasGs2ndEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_g(1,2,4) = BA.4"));
	}
	@Test public void context1HasGs3rdEncoding() 
	{
		assertThat(context1,containsString("LocalBAVarMapping_Q_g(6,6,6) = BA.6"));
	}
	
	@Test public void pHas1stLocalAssignment() 
	{
		assertThat(pi,containsString("Iassign.(IVar.LocalIVarMapping_P_c(i),IVar.IV.1)"));
	}	
	@Test public void pHas2ndLocalAssignment() 
	{
		assertThat(pi,containsString("Iassign.(IVar.LocalIVarMapping_P_d(i),IVar.IV.2)"));
	}	
	@Test public void pHas3rdLocalAssignment() 
	{
		assertThat(pi,containsString("Bassign.(BVar.LocalBVarMapping_P_e(i),True)"));
	}	
	@Test public void qHas1stLocalAssignment() 
	{
		assertThat(qijk,containsString("Iassign.(IVar.IV.1,IVar.LocalIVarMapping_Q_c(i,j,k))"));
	}
	@Test public void qHas2ndLocalAssignment() 
	{
		assertThat(qijk,containsString("Iassign.(IVar.IV.2,IVar.LocalIVarMapping_Q_c(i,j,k))"));
	}
	@Test public void qHas3rdLocalAssignment() 
	{
		assertThat(qijk,containsString("Bassign.(BArc.(LocalBAVarMapping_Q_f(i,j,k),Const.0),BArc.(LocalBAVarMapping_Q_g(i,j,k),IArc.(LocalIAVarMapping_Q_e(i,j,k),IVar.LocalIVarMapping_Q_c(i,j,k))))"));
	}
	
	
	@Test public void pHasLocalVariables() 
	{
		assertThat(pi,containsString("(Union({{IVar.LocalIVarMapping_P_c(i),IVar.LocalIVarMapping_P_d(i)}}),Union({{BVar.LocalBVarMapping_P_e(i)}}))"));
	}
	@Test public void qHasLocalVariables() 
	{
		assertThat(qijk,containsString("(Union({{IVar.LocalIVarMapping_Q_c(i,j,k)},{IVar.LocalIAVarMapping_Q_e(i,j,k).v | v <- itype(LocalIAVarMapping_Q_e(i,j,k))}}),Union({{},{BVar.LocalBAVarMapping_Q_f(i,j,k).v | v <- itype(LocalBAVarMapping_Q_f(i,j,k))},{BVar.LocalBAVarMapping_Q_g(i,j,k).v | v <- itype(LocalBAVarMapping_Q_g(i,j,k))}}))"));
	}
	@Test public void assertsCorrect() 
	{
		assertEquals(new String[] {"assert  STOP [T= Prog ","assert  STOP [T= Prog2 "},asserts1);
	}
	
	@Test public void context1HasCorrectIVNums()
	{
	    assertThat(context1,containsString("ivnums = {1..13}"));
	}
	@Test public void context1HasCorrectBVNums()
	{
	    assertThat(context1,containsString("bvnums = {1..4}"));
	}
	@Test public void context1HasCorrectIAVNums()
	{
	    assertThat(context1,containsString("ianums = {1..3}"));
	}
	@Test public void context1HasCorrectBANums()
	{
	    assertThat(context1,containsString("banums = {1..6}"));
	}

// "int a,b;\n" +
// 		"bool c,d;\n" +
// 		"int[3] e,f;\n" +
// 		"int[] i;\n" +
// 		"bool[4] g,h;\n" +
// 		"ParReads = {a,e[1]};\n" +
// 		"ParWrites = {c,g[2]};\n" +
// 		"SeqReads = {f,h};\n" +
// 		"SeqWrites = {h,d,i};\n" +

	@Test public void parReadsCorrect()
	{
		assertThat(context2,containsString("ParReads = Union({{IV.1},{IA.1.1}})"));
	}
	@Test public void parWritesCorrect()
	{
		assertThat(context2,containsString("ParWrites = Union({{BV.1},{BA.1.2}})"));
	}
	@Test public void seqReadsCorrect()
	{
		assertThat(context2,containsString("SeqReads = Union({{IA.2.v | v <- itype(IA.2)},{BA.2.v | v <- itype(BA.2)}}"));
	}
	@Test public void seqWritesCorrect()
	{
		assertThat(context2,containsString("SeqWrites = Union({{BA.2.v | v <- itype(BA.2)},{BV.2},{IA.3.v | v <- itype(IA.3)}}"));
	}
	@Test public void seqAssertCorrect()
	{
		assertEquals("assert compress(SeqRefTest(fst(P(1)))) [T= SeqRefTest(fst(Q()))",asserts2[0]);
	
	}
	@Test public void genAssertCorrect()
	{
		assertEquals("assert compress(GenRefTest(fst(P(2)))) [T= GenRefTest(fst(P(3)))",asserts2[1]);
	
	}
	@Test public void parAssertCorrect()
	{
		assertEquals("assert ParRefTest(CSNode.<CSLeaf.P(1),CSLeaf.P(2)>) [T= ParRefTest(CSLeaf.Q())",asserts2[2]);
	}
	@Test public void parAssertToStringCorrect()
	{
		assertEquals("assert Prog [P= Q()", compiler2.asserts.elementAt(2).toString());
	}
	
	@Test public void overseerNonNull()
	{
		Parser parser3 = new Parser(new StringReader(
		
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 4\n" +
		"%%InitB = false\n" +
		"%%InitI = 0\n" +
		"%%ditype = {0,1}\n" +
		"%%dctype = {0,1}\n" +
		"Writer() = skip\n" +
		"Reader() = skip\n" +
		"WideStruct =  <Reader(), overseer Writer()>\n" +
		"assert %- STOP [T= WideStruct\\({|error,verror|}) -% in WideStruct\n"
		
		));
		Compiler compiler3 = parser3.getCompiler();
	
	
		Iterator<CompilableProc> i = compiler3.asserts.elementAt(0).iterator();
		assertNotNull(i.next());
		assertNotNull(i.next());
	}		
	
	@Test public void overseerExampleWithoutOverseerNonNull()
	{
		Parser parser3 = new Parser(new StringReader(
		
		"%%ext_atomic=false\n" +
		"%%MinI = 0\n" +
		"%%MaxI = 4\n" +
		"%%InitB = false\n" +
		"%%InitI = 0\n" +
		"%%ditype = {0,1}\n" +
		"%%dctype = {0,1}\n" +
		"Writer() = skip\n" +
		"Reader() = skip\n" +
		"WideStruct =  <Reader(), Writer()>\n" +
		"assert %- STOP [T= WideStruct\\({|error,verror|}) -% in WideStruct\n"
		
		));
		Compiler compiler3 = parser3.getCompiler();
	
	
		Iterator<CompilableProc> i = compiler3.asserts.elementAt(0).iterator();
		assertNotNull(i.next());
		assertNotNull(i.next());
	}	



}