import org.junit.*;
import static org.junit.Assert.*;

import java.io.*;
import java.util.*;

public class ParserTest
{

	private Compiler compiler1;

	@Before
	public void setup()
	{
		Parser parser = new Parser(new StringReader(
		"int a, b = 3;" + "\n" +
		"int %4 c= 2, d;" + "\n" +
		"int %3 []  e = < 1 , 2 ,3>, f;" + "\n" +
		"int [4] g, h= <1 ,2 ,3 ,4>; " + "\n" +
		"int % 5 [2] i = <1,2>, j; " + "\n" +
		"bool k, l = true; " + "\n" +
		"bool [] m = <true, false, true>, n; " + "\n" +
		"bool [7] o,p = <true,true,true,false,true,true,false>; " + "\n" +
		"dirty int q = 5; " + "\n" +
		"dirty int [4] r, s, t; " + "\n" +
		"dirty bool u = true;  " + "\n" +
		"dirty bool [ 2] v = < true, false> , w = <false,false>; " + "\n" +
		"const N = 2;" + "\n" +
		"P(iParam) = {a := iParam}" + "\n" +
		"Q() = {skip}" + "\n" +
		"R() = {atomic{skip}}" + "\n" +
		"Prog = < < P(jParam + 1), Q()> , R() | jParam from 1 to N >" + "\n" 
		
		));
		compiler1 = parser.getCompiler();
	}

	
	@Test public void compiler1HasBsEncoding() 
	{assertNotNull(compiler1.context.getIEncoding("b")); }

	@Test public void compiler1binits() 
	{assertTrue(compiler1.context.iinits.get("b")==3);}
	
	@Test public void compiler1aInitsFalse() 
	{assertFalse(compiler1.context.iinits.containsKey("a"));}
	
	@Test public void compiler1cinits() 
	{assertTrue(compiler1.context.iinits.get("c")==2);}
	
	@Test public void compiler1csize() 
	{assertTrue(compiler1.context.contentSizes.get("c")==4);}
	
	@Test public void compiler1aNosize() 
	{assertFalse(compiler1.context.contentSizes.containsKey("a"));}
	
	@Test public void compiler1psize() 
	{assertTrue(compiler1.context.arrayIndexSizes.get("p")==7);}
	
	@Test public void compiler1pinits() 
	{assertEquals(new Vector<Boolean>(Arrays.asList(new Boolean [] {true , true, true, false, true, true, false}  ) ), compiler1.context.bainits.get("p"));}
	
	@Test public void compiler1iarraysize() 
	{assertTrue(compiler1.context.arrayIndexSizes.get("i")==2);}
	
	@Test public void compiler1icontentsize() 
	{assertTrue(compiler1.context.arrayContentSizes.get("i")==5);}

	@Test public void complexVariableAssignmentsParseCorrectly()
	{
		Parser parser = new Parser(new StringReader(
		"int i1, i2;" + "\n" +
		"bool b1, b2;" + "\n" +
		"int[] ia1, ia2;" + "\n" +
		"bool[] ba1, ba2;" + "\n" +
		"p(i) = skip with {i2,ia1[i*i],ia2,b1,ba2[2*i+1],ba1}" + "\n"
		
		
		));
		
		String actual = parser.getCompiler().procs.get("p").genCode(parser.getCompiler().context);
		String expected = "p(i) = (Skip\n, (Union({{IVar.IV.2,IVar.IA.1.(i*i)},{IVar.IA.2.v | v <- itype(IA.2)}}),Union({{BVar.BV.1,BVar.BA.2.((2*i)+1)},{BVar.BA.1.v | v <- itype(BA.1)}})))";
		assertEquals(expected,actual);
		
		
	}
	
	@Test public void localVarsWithVariableAssignmentsParseCorrectly()
	{
		Parser parser = new Parser(new StringReader(
		"int i1, i2;" + "\n" +
		"bool b1, b2;" + "\n" +
		"int[] ia1, ia2;" + "\n" +
		"bool[] ba1, ba2;" + "\n" +
		"p(i) = {int li;bool lb; int[] lia; bool[] lba; skip} with {i1,b1,ia1,ba1}" + "\n"
		
		
		));
		
		String actual = parser.getCompiler().procs.get("p").genCode(parser.getCompiler().context);
		String expected = "p(i) = (Skip\n, (Union({{IVar.LocalIVarMapping_p_li(i),IVar.IV.1},{IVar.LocalIAVarMapping_p_lia(i).v | v <- itype(LocalIAVarMapping_p_lia(i))},{IVar.IA.1.v | v <- itype(IA.1)}}),Union({{BVar.LocalBVarMapping_p_lb(i),BVar.BV.1},{BVar.LocalBAVarMapping_p_lba(i).v | v <- itype(LocalBAVarMapping_p_lba(i))},{BVar.BA.1.v | v <- itype(BA.1)}})))";
		assertEquals(expected,actual);
		
		
	}
	
	@Test public void paramaterisedProgramParsesCorrectly()
	{
	
// 		"const N = 2;" + "\n" +
// 		"P(iParam) = {a := iParam}" + "\n" +
// 		"Q() = {skip}" + "\n" +
// 		"R() = {atomic{skip}}" + "\n" +
// 		"Prog = < < P(jParam + 1), Q()> , R() | jParam from 1 to N >" + "\n" 
		
		Vector<String> pArgs = new Vector<String>();
		pArgs.add("iParam");
		Process p = new Process("P",pArgs,new IAssign(new IVar("a"), new IConstExpr(new VarIConst("iParam"))), new VariableAssignment());
		
		Vector<String> qArgs = new Vector<String>();
		Process q = new Process("Q",qArgs,new Skip(), new VariableAssignment());
		
		Vector<String> rArgs = new Vector<String>();
		Process r = new Process("R",qArgs,new Atomic(new Skip()), new VariableAssignment());
		
		Vector<Integer> emptyArgs = new Vector<Integer>();
		
		Vector<Integer>pArgs1 = new Vector<Integer>();
		Vector<Integer>pArgs2 = new Vector<Integer>();
		pArgs1.add(2);
		pArgs2.add(3);
		
		Vector<CmdStruct>nodes1 = new Vector<CmdStruct>();
		nodes1.add(new CSLeaf(new InitializedProc(p,pArgs1)));
		nodes1.add(new CSLeaf(new InitializedProc(q,emptyArgs)));
		
		Vector<CmdStruct>nodes2 = new Vector<CmdStruct>();
		nodes2.add(new CSLeaf(new InitializedProc(p,pArgs2)));
		nodes2.add(new CSLeaf(new InitializedProc(q,emptyArgs)));
		
		Vector<CmdStruct> outerNodes = new Vector<CmdStruct>();
		outerNodes.add(new CSNode(nodes1));
		outerNodes.add(new CSLeaf(new InitializedProc(r,emptyArgs)));
		outerNodes.add(new CSNode(nodes2));
		outerNodes.add(new CSLeaf(new InitializedProc(r,emptyArgs)));
		
		CmdStructProgram prog = new CmdStructProgram("Prog",new CSNode(outerNodes));
		
		
		assertEquals(prog, compiler1.progs.get("Prog"));
	
	
	}
	
	@Test public void callingSimpleProc()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p(i) = a := i" + "\n" +
		"q() = p(2)" + "\n"
		));
		Compiler compiler = parser.getCompiler();
		
		Cmd actual = compiler.procs.get("q").prog;
		
		Cmd expected = new IAssign(new IVar("a"), new IConstExpr(new NumIConst(2)));
		
		assertEquals(expected,actual);
	}
	
	@Test public void callingSimplestProc()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p() = skip" + "\n" +
		"q() = p()" + "\n"
		));
		Compiler compiler = parser.getCompiler();
		
		Cmd actual = compiler.procs.get("q").prog;
		
		Cmd expected = new Skip();
		
		assertEquals(expected,actual);
	}
	
	@Test public void callingSimpleProcWithCompoundConst()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p(i) = a := i" + "\n" +
		"q(j) = p(j + 3)" + "\n"
		));
		Compiler compiler = parser.getCompiler();
		
		Cmd actual = compiler.procs.get("q").prog;
		
		Cmd expected = new IAssign(new IVar("a"), new IConstExpr(new AddConst(new VarIConst ("j"),new NumIConst(3))));
		
		assertEquals(expected,actual);
	}
	
	@Test public void callingProcWithMultipleParameters()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p(i,j) = {if i > 0 then a := j; a := i + j}" + "\n" +
		"q(k) = if k = 0 then p(1,3) else p(k,4)" + "\n"
		));
		Compiler compiler = parser.getCompiler();
		
		Cmd actual = compiler.procs.get("q").prog;
		
		Cmd expected = 
			new If( 
				new Eq(new IConstExpr(new VarIConst("k")), new IConstExpr(new NumIConst(0))),
				new Seq(
					new If(
						new Gt(new IConstExpr(new NumIConst(1)), new IConstExpr(new NumIConst(0))),
						new IAssign(new IVar("a"), new IConstExpr(new NumIConst(3))),
						new Skip()),
					new IAssign(new IVar("a"), new Plus( new IConstExpr(new NumIConst (1)), new IConstExpr(new NumIConst(3))))
				),
				new Seq(
					new If(
						new Gt(new IConstExpr(new VarIConst("k")), new IConstExpr(new NumIConst(0))),
						new IAssign(new IVar("a"), new IConstExpr(new NumIConst(4))),
						new Skip()),
					new IAssign(new IVar("a"), new Plus(new IConstExpr(new VarIConst ("k")),new IConstExpr(new NumIConst(4))))
				)
			);
		assertEquals(expected,actual);
	}
	
	
	@Test(expected=ParserException.class)
	public void callingSimpleProcWithNonConstFails()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p(i) = a := i" + "\n" +
		"q(j) = p(a)" + "\n"
		));
		Compiler compiler = parser.getCompiler();
	}
	
	@Test public void localsInheritedFromProc()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p() = {int li; bool lb; int [] lia; bool [] lba; skip;}" + "\n" +
		"q() = {skip ; p() ; skip}" + "\n"
		));
		Compiler compiler = parser.getCompiler();
		
		Vector<String> lis = new Vector<String>();
		Vector<String> lbs = new Vector<String>();
		Vector<String> lias = new Vector<String>();
		Vector<String> lbas = new Vector<String>();
		
		lis.add("li");
		lbs.add("lb");
		lias.add("lia");
		lbas.add("lba");
		
		Process q = compiler.procs.get("q");
		
		assertEquals(lis,q.getLocalIVars());
		assertEquals(lbs,q.getLocalBVars());
		assertEquals(lias,q.getLocalIAVars());
		assertEquals(lbas,q.getLocalBAVars());
	}
	
	@Test(expected=ParserException.class)
	public void nestedScopingCausedByMacroFails()
	{
		Parser parser = new Parser(new StringReader(
		"int a;" + "\n" +
		"p() = int li; skip" + "\n" +
		"q() = int li; p()" + "\n"
		));
		Compiler compiler = parser.getCompiler();
	}
	
	
	
}