import org.junit.*;
import static org.junit.Assert.*;
import static org.junit.matchers.JUnitMatchers.*;

import java.util.*;

public class ProcessTest
{

	Context context1;

	@Before
	public void setup()
	{
		context1 = new Context();
		
		context1.is.add("i1");
		context1.is.add("i2");
		
		context1.ias.add("ia1");
		context1.ias.add("ia2");
		
		context1.bs.add("b1");
		context1.bs.add("b2");
		
		context1.bas.add("ba1");
		context1.bas.add("ba2");
		
	}


	@Test public void isAndBsGenerateOk()
	{
		String name = "p";
		Vector<String> consts = new Vector<String>();
		consts.add("i");
		Cmd prog = new Skip();
		VariableAssignment va = new VariableAssignment();
		
		/////////////////
		
		va.is.add("i1");
		va.is.add("i2");
		
		va.bs.add("b1");
		va.bs.add("b2");
		
		String expected = "p(i) = (Skip\n, (Union({{IVar.IV.1,IVar.IV.2}}),Union({{BVar.BV.1,BVar.BV.2}})))";
		
		/////////////////
		
		Process proc = new Process(name,consts,prog,va);
		Vector<String> empty = new Vector<String>();
		proc.addLocalIVars(empty);
		proc.addLocalBVars(empty);
		proc.addLocalIAVars(empty);
		proc.addLocalBAVars(empty);
		String actual = proc.genCode(context1);
		assertEquals(expected,actual);
		
	}
	
	@Test public void iasAndBasGenerateOk()
	{
		String name = "p";
		Vector<String> consts = new Vector<String>();
		consts.add("i");
		Cmd prog = new Skip();
		VariableAssignment va = new VariableAssignment();
		
		/////////////////
		
		va.is.add("i1");
		
		va.ias.add("ia1");
		
		va.bs.add("b1");
		
		va.bas.add("ba1");
		
		String expected = "p(i) = (Skip\n, (Union({{IVar.IV.1},{IVar.IA.1.v | v <- itype(IA.1)}}),Union({{BVar.BV.1},{BVar.BA.1.v | v <- itype(BA.1)}})))";
		
		/////////////////
		
		Process proc = new Process(name,consts,prog,va);
		Vector<String> empty = new Vector<String>();
		proc.addLocalIVars(empty);
		proc.addLocalBVars(empty);
		proc.addLocalIAVars(empty);
		proc.addLocalBAVars(empty);
		String actual = proc.genCode(context1);
		assertEquals(expected,actual);
		
	}	
	
	
	@Test public void iasAndBasWithAccessesGenerateOk()
	{
		String name = "p";
		Vector<String> consts = new Vector<String>();
		consts.add("i");
		Cmd prog = new Skip();
		VariableAssignment va = new VariableAssignment();
		
		/////////////////
		
		va.is.add("i1");
		
		va.iaAccesses.add(new Access("ia1",new VarIConst("i")));
		
		va.bs.add("b1");
		
		va.baAccesses.add(new Access("ba2",new NumIConst(0)));
		
		String expected = "p(i) = (Skip\n, (Union({{IVar.IV.1,IVar.IA.1.i}}),Union({{BVar.BV.1,BVar.BA.2.0}})))";
		
		/////////////////
		
		Process proc = new Process(name,consts,prog,va);
		Vector<String> empty = new Vector<String>();
		proc.addLocalIVars(empty);
		proc.addLocalBVars(empty);
		proc.addLocalIAVars(empty);
		proc.addLocalBAVars(empty);
		String actual = proc.genCode(context1);
		assertEquals(expected,actual);
		
	}	
	
	@Test public void iasAndBasWithMoreComplexAccessesGenerateOk()
	{
		String name = "p";
		Vector<String> consts = new Vector<String>();
		consts.add("i");
		Cmd prog = new Skip();
		VariableAssignment va = new VariableAssignment();
		
		/////////////////
		
		va.is.add("i2");
		
		va.iaAccesses.add(new Access("ia1",new MultConst(new VarIConst("i"),new VarIConst("i") )));
		
		va.ias.add("ia2");
		
		va.bs.add("b1");
		
		va.baAccesses.add(new Access("ba2",new AddConst(new MultConst(new NumIConst(2),new VarIConst("i")),new NumIConst(1))));
		
		va.bas.add("ba1");
		
		String expected = "p(i) = (Skip\n, (Union({{IVar.IV.2,IVar.IA.1.(i*i)},{IVar.IA.2.v | v <- itype(IA.2)}}),Union({{BVar.BV.1,BVar.BA.2.((2*i)+1)},{BVar.BA.1.v | v <- itype(BA.1)}})))";
		
		/////////////////
		
		Process proc = new Process(name,consts,prog,va);
		Vector<String> empty = new Vector<String>();
		proc.addLocalIVars(empty);
		proc.addLocalBVars(empty);
		proc.addLocalIAVars(empty);
		proc.addLocalBAVars(empty);
		String actual = proc.genCode(context1);
		assertEquals(expected,actual);
		
	}
	
	@Test public void iasAndBasAndLocalVariablesGenerateOk()
	{
		String name = "p";
		Vector<String> consts = new Vector<String>();
		consts.add("i");
		Cmd prog = new Skip();
		VariableAssignment va = new VariableAssignment();

		va.is.add("i1");
		
		va.ias.add("ia1");
		
		va.bs.add("b1");
		
		va.bas.add("ba1");
		
		Vector<String> livs = new Vector<String>();
		Vector<String> lbvs = new Vector<String>();
		Vector<String> liavs = new Vector<String>();
		Vector<String> lbavs = new Vector<String>();
		livs.add("li");
		lbvs.add("lb");
		liavs.add("lia");
		lbavs.add("lba");
		
		Process proc = new Process(name,consts,prog,va);
		proc.addLocalIVars(livs);
		proc.addLocalBVars(lbvs);
		proc.addLocalIAVars(liavs);
		proc.addLocalBAVars(lbavs);
		
		String expected = "p(i) = (Skip\n, (Union({{IVar.LocalIVarMapping_p_li(i),IVar.IV.1},{IVar.LocalIAVarMapping_p_lia(i).v | v <- itype(LocalIAVarMapping_p_lia(i))},{IVar.IA.1.v | v <- itype(IA.1)}}),Union({{BVar.LocalBVarMapping_p_lb(i),BVar.BV.1},{BVar.LocalBAVarMapping_p_lba(i).v | v <- itype(LocalBAVarMapping_p_lba(i))},{BVar.BA.1.v | v <- itype(BA.1)}})))";

		String actual = proc.genCode(context1);
		assertEquals(expected,actual);
		
	}	
}				